/** This program reads a "list" format file, 



 */


import java.io.*;
import java.util.*;

public class Mkcs {

    /** If true, include external evidence into the output file with 
     */
    static boolean writeFv = Boolean.getBoolean("fv");

    /** Stemmer modes */
    static class Stemmer {
	static final int NONE=0, EN=1, FR=2;
	
	static String describe() {
	    if (mode==NONE) return "None";
	    else if (mode==EN) return "English";
	    else if (mode==FR) return "French";
	    else return "Unknown";
	}

	/** Get the mode based on the -Dstem=... option 
	 */
	static int getMode() {
	    final String flag = "stem";
	    if (System.getProperty(flag)== null) return NONE;
	    String m = System.getProperty(flag).toLowerCase();
	    if (m.equals("none")) return NONE;
	    else if (m.equals("en")) return EN;
	    else if (m.equals("fr")) return FR;
	    else {
		String msg = "Unknown stemmer mode: " + m;
		throw new IllegalArgumentException(msg);
	    }
	}

	private static String trySuffix( Hashtable t, String x, String suff) {
	    if (x.endsWith(suff)) {
		String base = x.substring(0, x.length() - suff.length());
		if (t.get(base)!=null) return base;
	    } 
	    return null;
	}

	private static String trySuffix( Hashtable t, String x, String suff,
					 String baseEnd) {
	    String base = trySuffix(t,x,suff);
	    return (base != null && base.endsWith(baseEnd))? base: null;
	}
	/** Finds the "base form" for word x in table t.
	    For example, if x is "persons", and t contains
	    "person", return person. Otherwise, return null
	*/
	static String findBase(Hashtable t, String x) {
	    String base=null;
	    if (mode==EN) {
		base = trySuffix(t, x, "s");
		if (base!=null) return base;
	    } else if (mode==FR) {
		base = trySuffix(t, x, "s");
		if (base!=null) return base;
		base = trySuffix(t, x, "e");
		if (base!=null) return base;
		base = trySuffix(t, x, "es");
		if (base!=null) return base;

		base = trySuffix(t, x, "les", "l");
		if (base!=null) return base;

		base = trySuffix(t, x, "le", "l");
		if (base!=null) return base;

		base = trySuffix(t, x, "nes", "n");
		if (base!=null) return base;

		base = trySuffix(t, x, "ne", "n");
		if (base!=null) return base;

		base = trySuffix(t, x, "x", "eau");
		if (base!=null) return base;

	    }
	    return base;
	}

	/** What stemmer to apply, if any */
	static int mode = Stemmer.getMode();

    };


    /** A (paragraph, freq) pair */
    static class PosF {
	int para;
	int freq;	
	PosF(int para, int freq) {
	    this.para = para;
	    this.freq = freq;
	}
	PosF(int para) {
	    this(para,0);
	}
	/** Adds the frequence for x to this term */
	void merge(PosF x) {
	    if (x.para != para) throw new IllegalArgumentException("Paragraph mismatch");
	    freq += x.freq;
	}
	public String toString() {
	    return "" + para + " " + (double)freq;
	}
    }

    /** All pertinent information about one term */
    static class TermData {
	String term;
	/** Vector of PosF instances, describing frequencies
	    of this term in all paragraphs */
	Vector v=new Vector();
	TermData(String term) {
	    this.term = term;
	}	
	void addPair(PosF x) {
	    v.addElement(x);
	}
	/** Increments the term frequency for the specified paragraph
	    by the specified increment */
	void addFreq(int para, int incf) {
	    if (v.size() == 0) {
		v.addElement( new PosF(para, incf));
	    } else {
		PosF lastP = (PosF)v.elementAt( v.size()-1);
		if (lastP.para != para) {
		    v.addElement( new PosF(para, incf));
		} else {
		    lastP.freq += incf;
		}
	    }
	}
	/** Computes co-occurrence between this term and the other
	    one.
	    @param x The other term
	    @return Tanimoto coefficient  Nc / (N1+N2-Nc), where
	    Nc is the number of paragraphs where both terms occur, N1 and N2
	    are the number of paragraphs where N1 or N2, respectively,
	    occurs. Thus (N1+N2-Nc) is the number of paragraphs where at
	    least one of the terms occurs.
	    
	*/
	double cooc( TermData x) {
	    int i1=0, i2=0;
	    int common=0;
	    while(i1<v.size() && i2<x.v.size()) {
		int p1 = ((PosF)v.elementAt(i1)).para;
		int p2 = ((PosF)x.v.elementAt(i2)).para;
		if (p1 < p2) {
		    i1 ++;
		} else if (p1 > p2) {
		    i2 ++;
		} else if (p1==p2){
		    common ++;
		    i1 ++;
		    i2 ++;
		}
	    }
	    return (double)common/ (double)(v.size()+x.v.size()-common);

	}

	int df() {
	    return v.size();
	}
	int sumtf() {
	    int cnt=0;

	    for(int i=0; i<v.size(); i ++) {
		cnt += ((PosF)v.elementAt(i)).freq;
	    }

	    return cnt;
	}

	/** Adds the feature for term x into the vector for this term.
	    This method can be used when we have found that term x
	    is equivalent to this term, and want to remove it
	    from the corpus.
	*/
	    
	void merge(TermData x) {
	    int i1=0, i2=0;
	    Vector s=new Vector();
	    while(i1<v.size() && i2<x.v.size()) {
		PosF e1 = (PosF)v.elementAt(i1);
		PosF e2 = (PosF)x.v.elementAt(i2);
		int p1 = e1.para;
		int p2 = e2.para;
		if (p1 < p2) {		    
		    s.addElement(e1);
		    i1 ++;
		} else if (p1 > p2) {
		    s.addElement(e2);
		    i2++;
		} else if (p1==p2) {
		    e1.merge(e2);
		    s.addElement(e1);
		    i1 ++;
		    i2 ++;
		}
	    }
	    // Copy leftovers
	    while(i1<v.size()) {
		s.addElement(v.elementAt(i1++));
	    }
	    while(i2<x.v.size()) {
		s.addElement(x.v.elementAt(i2++));
	    }
	    v = s;
	}

    }


    static void error( String msg) {
	System.out.println(msg);
	System.exit(1);
    }

    /** Reads the input file. Each line should have this format:
	<pre>
	Para Position Term
	</pre>
	where Para and Position are the 0-based paragraph index
	and word-in-paragraph index.

	@return A hashtable that maps a term (String) to a TermData instance
	that describes the occurences of this term in all paragraphs.
    */
    static Hashtable readFile(String fileName) throws Exception {
	Hashtable t = new Hashtable();

	File f = new File(fileName);

	FileReader fr=null;
	try {
	    fr = new FileReader(fileName);
	} catch(FileNotFoundException e) {
	    error("Cannot read file: " + fileName);
	}
	BufferedReader reader = new BufferedReader(fr);
	String line;
	int section = -1; // current section # read

	int lineCnt =0;
	while ( ( line = reader.readLine() ) != null ) {
	    lineCnt ++;
	    if ( line.startsWith( "#" )) continue;
	    //// System.out.println("[sec="+section+"] input: "+line);

	    StreamTokenizer st = new StreamTokenizer(new StringReader(line)); 
						     

	    if (st.nextToken() == StreamTokenizer.TT_EOF) continue;
	    else if (st.ttype != StreamTokenizer.TT_NUMBER) {		
		throw new Exception("Number expected in line "+lineCnt);
	    }
	    int para = (int) st.nval;

	    if (st.nextToken() != StreamTokenizer.TT_NUMBER) {
		throw new Exception("Number expected in line "+lineCnt);
	    }
	    int pos = (int) st.nval;

	    if (st.nextToken() != StreamTokenizer.TT_WORD) {
		throw new Exception("Word expected in line "+lineCnt);
	    }    
	    String term = st.sval;
	    TermData d = (TermData)t.get(term);
	    if (d==null) {
		d = new TermData(term);
		t.put(term,d);
	    }
	    d.addFreq(para,1);
	}
	System.out.println("# " + lineCnt + " lines read\n");
	System.out.println("# " + t.size() + " distinct terms\n");
	return t;
    }


    /** Removes all suffixes (.*) from a string, and replaces "-" with
	"_".
	Extracts "foo" from "foo.ext1.ext2..."
     */
    static String getBaseName(String f) {
	int pos = f.indexOf(".");
	String x = (pos < 0) ?  f : f.substring(0,pos);
	return x.replace('-', '_');
	   
    }

    /** Returns a sorted array of keys of a hash table */
    static String[] listKeys( Hashtable t) {
	String termList [] = new String[t.size()];
	int i=0;
	for (Enumeration e = t.keys() ; e.hasMoreElements() ;) {
	    termList[i++] = (String)e.nextElement();
	}
	Arrays.sort(termList);	
	return termList;
    }

    /** Applies the stemmer to table t. Scans table t, and
	removes all terms that the stemmer believes are
	variant forms of some other "base" term. Adds the
	frequency vectors of the removed terms to those
	of the remaining base terms.
     */
    static void applyStemmer(Hashtable t) {
	System.out.println("# Stemmer: " + Stemmer.describe());
	if (Stemmer.mode==Stemmer.NONE) return;

	String termList [] =listKeys(t);
	for(int i=0; i<termList.length; i++) {	    
	    String x=termList[i];

	    String base=null;
	    if ((base=Stemmer.findBase(t, x)) != null) {
		System.out.println("# Stem: " +x+ " --> " + base);
		((TermData)t.get(base)).merge( (TermData)t.get(x));
		t.remove(x);
	    } 
	}
    }

    static public void main(String argv[]) throws Exception {
	if (argv.length != 1) {
	    error( "Usage: java [-Dfv=true] [-Dstem=en|fr] Mkcs input file");
	}

	String fileName = argv[0];

	System.out.println("# Source file: " + fileName);

	Hashtable t=	readFile( fileName);
	applyStemmer(t);
	String termList [] =listKeys(t);

	final String sep = "\n\n#----------------------------------------------------------------";
	String csName = getBaseName(fileName);

	System.out.println(sep);
	System.out.println("ConceptSystem: "+ csName);

	System.out.println(sep);
	System.out.println("ConceptCategories:");
	System.out.println("#	Using default category for all nodes.");

	System.out.println(sep);
	System.out.println("RelationModels:");

	final String COOC = "Cooccurrence";

	System.out.println(COOC + "    Undirected");
	System.out.println("#	1 relation models.");

	System.out.println(sep);
	System.out.print("Concepts:");
	if (writeFv) {
	    System.out.print(" fv");
	}
	System.out.println();
	System.out.println("# Using terms as concepts:");
	for(int i=0; i<termList.length; i++) {	    
	    System.out.print(termList[i]);
	    TermData d = (TermData)t.get(termList[i]);
	    if (writeFv) {
		// Print the term frequency data for this term in all docs
		// where it occurs
		for(int k=0; k<d.v.size(); k++) {
		    System.out.print(" " + d.v.elementAt(k));
		}
	    }   
	    System.out.println();
	    System.out.println("# DF=" + d.df() + " " + d.sumtf() + " " +termList[i]);  
	}
	System.out.println("#	" + termList.length + " concepts.");
	
	System.out.println(sep);
	System.out.println("Relations:");

	int relCnt=0;
	for(int i=0; i<termList.length; i++) {
	    //for(int j=0; j<termList.length; j++) {		
	    for(int j=0; j<i; j++) {		
		double cooc = 
		    ((TermData)t.get(termList[i])).cooc( (TermData) t.get(termList[j]));
		if (cooc > 0) {
		    System.out.println(COOC + " " + termList[i] + " " +
				       termList[j] + " " + cooc);
		    relCnt++;
		}
	    } 
	}

	System.out.println("#	" + relCnt + " relations.");


    }



}
