///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// ABSURDIST II
// Ying Feng, Vladimir Menkov
// Computer Science, Indiana University
// Feb-Aug 2003
//
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////



package absurdist;
    
import java.io.*;
import java.text.*;
import java.util.*;

import common.*;  // Classes used both by ABSURDIST and by other our packages
import driver.*;
import gui.*;
import bnb.BranchAndBound;



/**
   Class Absurdist provides the ABSURDIST II graph matching algorithms
   to compute the correspondence between two concept systems, and
   finds the optimal mappings between them. 

   The behavior of this class is affected by a number of parameters,
   which can be supplied as Java System Properties on the command 
   line (java -Dname=value ...). 

   For example, running 

   java -DresetL=false -DL=0.01 -DmaxLoop=5000 driver.ApplicDriver

   will run the program with fixed L equal to 0.01, interrupting
   the loop at the 5000-th iteration, if the convergence has
   not been achieved.

   java -DresetL=true  driver.ApplicDriver

   will adjust the learning rate at every step to 1/2 of the maximum
   safe rate. This is supposed to achieve fastest convergence.
   
   One can find all the parameters in the code by searching for
   'prop.getOption'.
*/
public class Absurdist
{
    /** An empty property list -- just an interface to system properties */
    public static ParseConfig prop = new ParseConfig();

    // status of system mapping
    /** Concept systems not mapped. */
    public static final int UNMAPPED = -1;
    /** Concept systems being mapped. */
    public static final int MAPPING = 0;
    /** Concept systems mapped. */
    public static final int MAPPED = 1;

    /** maximum noise percentage level */
    public static final int maxCCNoise = 10;
    public static final int maxCINoise = 10;
    public static final int maxRCNoise = 10;
    public static final int maxRINoise = 10;

    // mapping parameters
    /** Intitial value for correspondence matrix. */
    public static double cInit = 0.5;
    /** Learning rate (this may be adjusted if resetL=true). */
    public static double lRate = prop.getOptionDouble("L", 0.01);
    /** Maximum number of iterations. */
    public static int maxLoop = prop.getOption("maxLoop", 3000);   
    /** Sampling interval for plots. */
    public static int intLoop = 10;

    // mapping thresholds
    /** Threshold for iteration convergence. */
    public static double cvgThresh = prop.getOptionDouble("eps",1E-10);
    /** Threshold for valid mapping. */
    public static double mapThresh = 0;

    // controllers for mapping methods options

    /** Various ways of setting Beta and Chi. */
    static class CoefMode {
	/** Use the values passed in via -Dbeta=... -Dchi=... */
	final static int FIXED_GIVEN = 0;
	/** Fixed values set by Ying Feng's early 2003 code */
	final static int FIXED_YF2003 = 1;
	/** Fixed values: beta as given, chi=beta*(n-2). This is near
	    the chi/beta threshold above which the convergence to *a*
	    permutation is almost ensured. This seems to give a better
	    permutation at convergence.  (VM, June 2004) */
	final static int FIXED_N_2 = 2;
	/**  Fixed beta. Recompute chi at every step, maintaining
        sum_{a,x}(N(a,x))=0. This heuristic was used for October 2003
	experiments. */
	final static int RESET_BALANCE = 3;
	/** Verify it's a legal mode */
	void validate(int mode) {
	    if (mode <  FIXED_GIVEN || mode > RESET_BALANCE) 
		throw new AssertionError("Invalid coefficient setting mode = " + mode);
	}
    };

    /** This parameter controls how chi and beta are initially set,
	and whether they are reset during the iterative process. */
    public static int coefMode=prop.getOption("coefMode",CoefMode.FIXED_N_2);

    /** If true, we recompute the learning rate at each step,
	for fastest convergence. (For compatibility with the 2003
	code, set the default to false!)  */
    public static boolean resetL = prop.getOption("resetL",true);

    /** The ceiling for adjustable learning rate. If 0 or negative,
	it is ignored */
    public static double maxL = prop.getOptionDouble("maxL",0);

    /** Damping modes */ 
    static private class Damping {
	/** Soft damping as per Goldstone and Rogosky (2002) */
	final static int GR2002=0;
	/** "Hard" damping */
	final static int HARD=1;
    };

    /** Damping mode. Should be 0 (= G & R 2002)  or 1 (= Hard damping) */
    //public static int dampingMode=prop.getOption("dampingMode",Damping.HARD);
    public static int dampingMode=prop.getOption("dampingMode",Damping.GR2002);	

    /** If true, use the "fast" (N^2 D^2) method for 
	general (labeled, weighted) graphs. */
    public static boolean useFastGeneral = true;

    /** If true, use the exponential version of activator and inhibitor
        to compute netwoek input. */
    public static boolean useExponential = false;

    /** If true, 1-1 mapping will be enforced. */
    public static boolean enforce1to1map = true;

    /** If true, we'll find the optimal permutation (this may be very
        expensive!) and compare the quality of our permutation to it */
    public static boolean adjBest = prop.getOption( "adjBest", false);
     
    /** This will be set to true in the constructor if both 
	concept systems being matched are "binary", that is have no other
	relationship weights than 1 and 0. If this is the case, we can
	use faster algorithms for computing inhibition, excitation, etc. */
    public boolean systemsAreSimple = false;

    // controllers for PhyloTree mapping (a special application)
    /** If true, use anchors overlap for external similarity */
    public boolean useAnchors = prop.getOption( "useAnchors", false );
    
    /**
       If true, the alphaShare parameter is used to set alpha via
       beta. This is false by default.
    */
    public boolean useAlphaShare = prop.getOption( "useAlphaShare", true);

    /** This flag determines the alpha/(alpha+beta) ratio to be used in
	adjustCoeff(); it only is used if useAlphaShare is true, and is ignored
	otherwise. If it is 1.0, then alpha will be set to 1.0 and
	beta to 0.

	@see #adjustCoeff	
    */
    public static double alphaShare = prop.getOptionDouble("alphaShare", 0.5);

    // coefficients for computing network input.
    /** Coefficients for external similarity. */
    public double alpha = 0;
    /** Coefficients for activation. */
    public double beta = prop.getOptionDouble("beta", 1.0);
    /** Coefficients for inhibition. */
    public double chi = prop.getOptionDouble("chi", 1.0);
    
    /** Source system for mapping. */
    public ConceptSystem systemO; 
    /** Target system for mapping. */
    public ConceptSystem systemN; 

    /** graph-style interface for systemO if available. */
    CSGI sysO; 
    /** graph-style interface for systemN, if available. */
    CSGI sysN; 

    /** Actual number of steps being recorded during network iteration. */
    public int nStep = 0;
    /** Status of the system. */
    public int status = UNMAPPED;
    
    // matrices storing intermediate mapping results
    /** External similarity E[row][col]. */
    public double[][] exSim;
    /** Internal similarity, also called activator, R[step][row][col]. */
    public double[][][] inSim;
    /** Inhibitor I[step][row][col]. */
    public double[][][] inhib;
    /** Network input N[step][row][col]. */
    public double[][][] netIn;
    /** Correspondence matrix C[step][row][col]. */
    public double[][][] correspond;

    // mapping results
    /** Mappings from systemO to systemN, recording indices of matched concepts. */
    public int[] mapON;
    /** Mappings from systemN to systemO, recording indices of matched concepts. */
    public int[] mapNO;


    /** Used for reporting. An instance of this class can be used
	for an individual runs, or for accumulated data from multiple runs.*/
    static public class Results {
	// mapping evaluations
	/** Total number of mismapped concepts. */
	public double misCon = 0;
	/** Sum of mismapped relations (i.e. link difference). */
	public double misRel = 0;
	/** Sum of mismapped relations adjusted by subtracting the
	    sum of mismapped relations for the identity mapping */
	public double misRelAdjId = 0;
	/** Same thing, but adjusted by the misrel for the optimal mapping */
	public double misRelAdjOp = 0;
	/** Mapping accuracy based on internal similarity. */
	public double accuracy = 0;
	/** How many iterations have actually been performed by
	    the end of the iterative process */
	public double iterCnt = 0;
	/** Experiment repetition count, for statistics */
	public int rpt = 0;
	/** How many runs resulted in an identity permutation */
	public int perfectCnt=0;
	/** Squares of the averaged values; maintained so that we can
	    compute standard deviation */
	public double misCon2=0, misRel2=0, misRelAdjId2=0, misRelAdjOp2=0, accuracy2=0, iterCnt2=0;
	/** Standard deviations */
	public double misConDev=0, misRelDev=0, misRelAdjIdDev=0, misRelAdjOpDev=0, accuracyDev=0, iterCntDev=0;
	boolean averaged = false;
	/** Adds the results of one more experiment. Used for statistics */
	public void add(Results r) {
	    rpt ++;
	    misCon   += r.misCon;
	    misRel   += r.misRel;
	    misRelAdjId += r.misRelAdjId;
	    misRelAdjOp += r.misRelAdjOp;
	    accuracy += r.accuracy;
	    iterCnt  += r.iterCnt;

	    if (r.misCon==0) perfectCnt++;

	    misCon2   += r.misCon * r.misCon;
	    misRel2   += r.misRel * r.misRel;
	    misRelAdjId2 += r.misRelAdjId * r.misRelAdjId;
	    misRelAdjOp2 += r.misRelAdjOp * r.misRelAdjOp;
	    accuracy2 += r.accuracy * r.accuracy;
	    iterCnt2  += r.iterCnt * r.iterCnt;
	}

	/** Computes averages and standard deviations */
	public void averages() {
	    if (averaged) throw new AssertionError("Can't average again!");
	    averaged = true;
	    misCon  /=rpt;
	    misRel  /=rpt;
	    misRelAdjId  /=rpt;
	    misRelAdjOp  /=rpt;
	    accuracy/=rpt;
	    iterCnt /=rpt;
	    misConDev   = Math.sqrt( misCon2/rpt - misCon*misCon);
	    misRelDev   = Math.sqrt( misRel2/rpt - misRel*misRel);
	    misRelAdjIdDev   = Math.sqrt( misRelAdjId2/rpt - 
					  misRelAdjId*misRelAdjId);
	    misRelAdjOpDev   = Math.sqrt( misRelAdjOp2/rpt - 
					  misRelAdjOp*misRelAdjOp);
	    accuracyDev = Math.sqrt( accuracy2/rpt - accuracy*accuracy);
	    iterCntDev  = Math.sqrt( iterCnt2/rpt - iterCnt*iterCnt);
	}
    };

    public Results results=new Results();

    /** Stores link similarity information, to avoid recomputation.
	Key is a LinkPair; value is a Double. We store sim info
	for link pairs (link1, link2) where both link1 and link2 are
	non-empty links; we also store info for (EMPTY, link2) and
	(link1, EMPTY), for all non-empty link1 and link2.
    */
    private LinkSimCache linkSimCache = prop.getOption("compactCache",false) ?
	(LinkSimCache)new LinkSimCacheArray() : (LinkSimCache)new LinkSimCacheHash();


    /** Creates an empty absurdist system. */
    public Absurdist()
    {
    }

    /** Creates an absurdist system containing the source and target concept systems 
        and random external similarities. */
    public Absurdist( ConceptSystem systemO, ConceptSystem systemN )
    {
	this.systemO = systemO;
	this.systemN = systemN;
    }


    /** Sets the source concepts system to be mapped from.*/
    public void setSystemO( ConceptSystem systemO )
    {
	this.systemO = systemO;
	status = UNMAPPED;
    }

    /** Sets the target concepts system to be mapped to. */
    public void setSystemN( ConceptSystem systemN )
    {
	this.systemN = systemN;
	status = UNMAPPED;
    }

    /** Sets the two concepts systems for mapping. */
    public void setSystems( ConceptSystem systemO, ConceptSystem systemN )
    {
	this.systemO = systemO;
	this.systemN = systemN;
	status = UNMAPPED;
    }

    /** Returns true if the two systems have not yet been created or mapped; otherwise false.
	In this status, all data, i.e. matrices, and nStep are invalid, either null or obsolete. */
    public boolean isUnmapped()
    {
	if ( status == UNMAPPED )
	    return true;
	else
	    return false;
    }

    /** Returns true if the two systems are being mapped; otherwise false.
	In this status, all data, i.e. matrices, and nStep are already initialized. */
    public boolean isMapping()
    {
	if ( status == MAPPING )
	    return true;
	else
	    return false;
    }

    /** Returns true if the two systems are already mapped; otherwise false. */
    public boolean isMapped()
    {
	if ( status == MAPPED )
	    return true;
	else
	    return false;
    }

    /** Sets the status for Absurdist. */
    public void setStatus( int status )
    {
	this.status = status;
    }

    /** Resets the status for Absurdist to UNMAPPED. */
    public void resetStatus()
    {
	status = UNMAPPED;
    }

    /*
    // Returns the actual number steps recorded in correspondence matrix.
    public int getStepNum()
    {
	return nLoop / intLoop;
    }
    */


    /** Turn off the iteration history recording so that less memory
	will be used.  This is used if you don't need to keep the
	intermediate results during the iteration. This is important
	for saving memory, and should always be done when you don't
	use the GUI and don't need the "timeseries" graphs.
    */
    public static void turnOffHistory()
    {
	intLoop = maxLoop; // then only one step, i.e. the last step, will be recorded
    }

    /** Computes value ranges of the specified matrix.
	@param a a rectangualr matrix
	@return (minE, maxE) 
    */
    private double[] elementRange(double a[][]) {
	double r[] = new double[2];
	r[0] = r[1] = a[0][0];
	for ( int io=0; io<a.length;  io++ ) {
	    for ( int in=0; in<a[io].length; in++ ) {
		double e= a[io][in];
		if (e< r[0]) r[0] = e;
		if (e> r[1]) r[1] = e;
	    }
	}
	return r;
    }
    
    /** Computes a safe learning rate for the current value of the
	chi and beta parameters. This is supposed to be in the
	middle of the interval [0; maxL]. Here maxL = min(maxL1,
	maxL2); maxL1 is the maximum value of learning rate that
	ensures that the elements of the correspondence matrix stay in
	the [0;1] range; maxL2 is the maximum value of learning rate
	that guarantees convergence.  
    */
    void setSafeL() {
	int no = systemO.getConceptNum();
	int nn = systemN.getConceptNum();
        int n = Math.max( no, nn );

	// Ensuring convergence
	double maxL2 = 2.0/( beta * (n-1) + 2 * chi);
	// Set L in the middle of the safe interval
	double mymax = maxL2;

	if (dampingMode == Damping.GR2002) {
	    // Ensuring that the elements of the correspondence matrix stay in
	    // the [0;1] range


	    // The range of E...
	    double rangeE[]= elementRange(exSim);//FIXME: could be precomputed!
	    // The range of N:
	    double minNet = alpha * Math.min(0, rangeE[0]) - chi;
	    double maxNet = alpha * Math.max(0, rangeE[1]) + beta*(n-1);
	    //System.out.println("Predicted: N in ["+minNet+" : " + maxNet+"]");
	    
	    double maxL1 = 1.0/( Math.max( -minNet, maxNet));
	    if (maxL1 < mymax) mymax = maxL1;
	} 

	lRate = 0.5 * mymax;
	// Apply ceiling, if applicable
	if (maxL >0 && lRate > maxL) lRate = maxL;
	System.out.println("Set L=" + lRate);
    }
 
    /** Prints out ABSURDIST parameters (if fixed), or rules for adjusting 
	them (if adjustable) */
    void describe() {

	String msg = "Max iter = " + maxLoop +", eps=" + cvgThresh;
	msg += "\nConstant beta=" + beta;
    
	if (coefMode == CoefMode.RESET_BALANCE) {	    
	    msg += "\nChi will be re-set at every step to keep sum(N)=0";
	} else {
	    msg += "\nConstant chi = " + chi;   
	}

	if (resetL) {
	    msg += "\nL will be set based on chi and beta; " + 
		(maxL>0? " ceiling=" + maxL : " no ceiling");
	} else {
	    msg += "\nFixed L = " + lRate;
	}

	if (dampingMode == Damping.GR2002) {
	    msg += "\nDamping: G&R 2002";
	} else if  (dampingMode == Damping.HARD) {
	    msg += "\nDamping: Hard";	    
	}
	//SimpleDriver.debugln(msg);
	System.out.println(msg);

    }

    /** Initializes external similarity matrix to all 0's. */
    public void initExSim()
    {
	int no = systemO.getConceptNum();
	int nn = systemN.getConceptNum();	
	for ( int io=0; io<no;  io++ ) {
	    for ( int in=0; in<nn; in++ ) {		       
		exSim[io][in] = 0;
	    }
	}
    }

    /** Initializes various data structures to be used for mapping. */
    protected void initData()
    {
	// reset number steps
	nStep = 0;

	// See if we can use an economical way to compute sums
	systemsAreSimple = systemO.isSimple() && systemN.isSimple();

	// prepare graph-style interface
	if ( useFastGeneral ) {
	    SimpleDriver.debugln("Absurdist: will use the O(N^2*D^2)/step algorithm.");
	    
	    // prepare a graph-style interface, to exploit sparseness
	    sysO = new CSGI( systemO );
	    sysN = new CSGI( systemN );

	    // Pre-compute the E^2 = N^2*D^2 link similarity values,
	    // for re-use on each iteration
	    cacheLinkSimData();
	} else {
	    SimpleDriver.debugln( "Absurdist: will use the O(N^4)/step algorithm." );
	}

	int no = systemO.getConceptNum();
	int nn = systemN.getConceptNum();
        int n = Math.min( no, nn );

	// compute external similarity. 
	exSim = new double[no][nn];	
	if (useAnchors && 
	    (systemO.allAnchorCnt() > 0 || systemN.allAnchorCnt() > 0)) {
	    SimpleDriver.debugln("Absurdist: Computing external similarity based on anchors ("+
				 systemO.allAnchorCnt() + " in systemO, " + 
				 systemN.allAnchorCnt() + " in systemN).");

	    // Computing external similarity based on anchor overlap
	    int denom = Math.min( no,nn)-1;	
	    for ( int io=0; io<no;  io++ ) {
		for ( int in=0; in<nn; in++ ) {
		    exSim[io][in] =
			(double)systemO.anchorOverlap( systemN, io, in)/
			(double)denom;
		}
	    }
	    alpha = 1;
	    System.out.println("Using alpha=" + alpha);
	} 
	else {
	    SimpleDriver.debugln( "Absurdist: computing external similarity based on concept features.");

	    // compute external similarity based on the semantic distance
	    // between concepts in the two systems
	    for ( int io=0; io<no;  io++ ) {
		for ( int in=0; in<nn; in++ ) {
		    Concept co = systemO.getConcept(io);
		    Concept cn = systemN.getConcept(in);
		    exSim[io][in] = 1 - co.distance(cn);
		}
	    }
	}

	// allocate arrays right here, to avoid null pointers in graphics code
	SimpleDriver.debugln( "Absurdist: allocating big matrices ...");
        int maxStep = maxLoop / intLoop;
	if (maxStep < 1) maxStep = 1;
        inSim = new double[maxStep][no][nn];
        inhib = new double[maxStep][no][nn];
        netIn = new double[maxStep][no][nn];
	correspond = new double[maxStep][no][nn];	
	SimpleDriver.debugln( "Absurdist: big matrices allocated");
    }

    /** Initial setting of the beta and chi coefficients according to
        the node number and edge fullness of the graph. */
    protected void adjustCoeff()
    {
	int no = systemO.getConceptNum();
	int nn = systemN.getConceptNum();
        int n = Math.min( no, nn );

	if (coefMode == CoefMode.FIXED_YF2003) {
	    // Automatic initialization of chi and beta, if prescribed
	    beta = ( double )( 3*n - 1 ) / ( 2*n*n - 2 );
	    chi = 1 - beta;
	    if ( useExponential ) { 
		beta = Math.exp( beta );
		chi = Math.exp( chi );
		double c = ( beta + chi ) / ( 1 - alpha );
		beta /= c;
		chi /= c;
	    }
	}  else if (coefMode == CoefMode.FIXED_N_2){
	    // Keep the default beta, or the one passed via -Dbeta=...
	    // Set chi/beta=n-2, to encourage convergence
	    // to a permutation matrix
	    chi = beta * (n-2);
	}     

	if (useAlphaShare) {
	    // Special mode (not normally used) to set alpha via beta. 
	    // Was used for some experiments in the summer of 2003.
	    if (alphaShare==1.0) {
		alpha = 1.0;
		beta = 0.0;
	    } else if (0 <= alphaShare && alphaShare < 1) {
		alpha = beta * alphaShare / (1 - alphaShare);
	    } else {
		String msg="Illegal ratio a/(a+b)=" +alphaShare;
		throw new IllegalArgumentException(msg);
	    }
	}

	describe();
	if (resetL) setSafeL();	
    }
    
    /** Compares the directed relation bundles (io --> jo) and (in --> jn), returns the 
	difference, measured as the sum_{type} |Weight(io,jo,type) - Weight(in,jn,type) */
    private double relationBundleDiff( Vector rso, Vector rsn )
    {
	int ko, kn;
	Relation ro, rn;
	double dw = 0;
	if ( rso == null && rsn == null ) { } // don't do anything
	else if ( rsn == null ) {
	    for ( ko=0; ko < rso.size(); ko++ ) {
		ro = ( Relation ) rso.get(ko);
		dw += ro.weight;
	    }
	}
	else if ( rso == null ) {
	    for ( kn=0; kn < rsn.size(); kn++ ) {
		rn = ( Relation ) rsn.get(kn);
		dw += rn.weight;
	    }
	}
	else {
	    ko = 0;
	    kn = 0;
	    // scan both lists
	    while ( ko < rso.size() && kn < rsn.size() ) {
		ro = ( Relation ) rso.get(ko);	
		rn = ( Relation ) rsn.get(kn);
		if ( ro.compareTo( rn ) < 0 ) { // rn missing
		    dw += ro.weight;
		    ko++;
		}
		else if ( ro.compareTo( rn ) > 0 ) { // ro missing
		    dw += rn.weight;
		    kn++;	
		}	
		else { // both exist
		    dw += ro.difWeight( rn );
		    ko++;
		    kn++;
		}
	    }
	    // rsn finished
	    while ( ko < rso.size() ) { 
		ro = ( Relation ) rso.get(ko);	
		dw += ro.weight;
		ko++;
	    }
	    // rso finished
	    while ( kn < rsn.size() ) {
		rn = ( Relation ) rsn.get(kn);
		dw += rn.weight;
		kn++;	
	    }		
	} // end of if    }
	return dw;
    }

    /** Converts the "relation difference" (|Rel1 - Rel2|) to "relation similarity". */
    private double diffToSim(double dw) 
    {
	double sim=0;
	if ( dw < 1 ) {
	    sim = 1 - dw;
	} else {
	    sim = 0;
	} 
	return sim;
    }
    
    /** Computes the similarity between link (io, jo) in systemO and
        (in, jn) in systemN. */
    protected double calcLinkSim( int io, int jo, int in, int jn ) 
    {
	return calcLinkSim( new Link(io,jo), new Link(in,jn) );
    }

    /** Computes the similarity between the two links.  Each link may
	be a "real" link (a pair of numbers), or the EMPTY link. */
    public double calcLinkSim( Link lo, Link ln )
    {
	double dw = 0;
	Vector rso, rsn;

	// add relation weight differences in direction i->j
	rso = lo.equals(Link.EMPTY)? null :
	    (Vector)systemO.getRelations( lo.i, lo.j);
	rsn = ln.equals(Link.EMPTY)? null :
	    (Vector)systemN.getRelations( ln.i, ln.j);

	dw += relationBundleDiff( rso, rsn);

	// add relation weight differences in direction j->i
	rso = lo.equals(Link.EMPTY)? null :
	    (Vector)systemO.getRelations( lo.j, lo.i);
	rsn = ln.equals(Link.EMPTY)? null :
	    (Vector)systemN.getRelations( ln.j, ln.i);

	dw += relationBundleDiff( rso, rsn);

	dw /= 2 * systemO.getModelNum(); // the two systems have the same z# of models
	return diffToSim(dw);
    }

    /** Pre-computes all non-trivial link similarity values, and stores them
	in linkSimCache. This method is only used when useFastGeneral=true. */
    private void cacheLinkSimData() 
    {
	int no = systemO.getConceptNum();
	int nn = systemN.getConceptNum();
	int cnt=0;

	// Estimate table sizes
	int linkCntO = 0; 
	for(int a=0; a<no; a++) {
	    linkCntO += sysO.degree(a);
	}
	int linkCntN = 0; 
	for(int x=0; x<nn; x++) {
	    linkCntN += sysN.degree(x);
	}
	// 1 is added to account for EMPTY
	linkSimCache.setSizes(linkCntO+1, linkCntN+1); 
	

	for(int a=0; a<no; a++) {
	    int bList[] = sysO.getNei(a);
	    for(int j=0; j<bList.length; j++) {
		Link lo = new Link(a, bList[j]);		
		for(int x=0; x<nn; x++) {
		    int yList[] = sysN.getNei(x);
		    for(int k=0; k<yList.length; k++) {
			Link ln = new Link(x, yList[k]);
			linkSimCache.put( lo, ln, calcLinkSim(lo, ln));
			cnt ++;
			if ((cnt % 10000)==0) SimpleDriver.debugln("cached " + cnt+ "...");
		    }
		}		
		linkSimCache.put(lo, Link.EMPTY, calcLinkSim(lo, Link.EMPTY));
		cnt ++;
	    }
	}
			
	for(int x=0; x<nn; x++) {
	    int yList[] = sysN.getNei(x);
	    for(int k=0; k<yList.length; k++) {
		Link ln = new Link(x, yList[k]);
		linkSimCache.put( Link.EMPTY, ln, calcLinkSim(Link.EMPTY, ln));
		cnt ++;		
	    }
	}
	linkSimCache.put( Link.EMPTY, Link.EMPTY, calcLinkSim(Link.EMPTY, Link.EMPTY));
	cnt++;
	SimpleDriver.debugln( "Absurdist: cached " + cnt + " link similarity values." );
	SimpleDriver.debugln( "Absurdist: table size = " + linkSimCache.size() );
    }

    /** Access function for linkSimCache; retrieves the cached link
        similarity value. */
    double lookupLinkSim( Link lo, Link ln )     
    {
	if (useFastGeneral) {
	    double d = linkSimCache.get(lo,ln);
	    if (Double.isNaN(d)) {
		String msg = "lookupLinkSim("+lo+","+ln+") failed!\n"  +
		    "Table size = " + linkSimCache.size();
		SimpleDriver.debugln(msg);
		
		throw new AssertionError(msg);
	    } 
	    return d;
	    // linkSimCache.put(lo, ln, d = calcLinkSim(lo, ln));

	    /*
	    for (Enumeration e = linkSimCache.keys() ; e.hasMoreElements() ;) {
		LinkPair k= (LinkPair)e.nextElement();
		System.out.println("" + k + " ("+k.equals(key)+") ==> " + linkSimCache.get(k));
	    }
	    */
	} else {
	    return calcLinkSim(lo, ln);
	}
    }

    /** Computes the internal similarity matrix fast (in O(N^2*D^2)
	 time), for an unlabeled unweighted graph.

	<p>
	We use the following sparsity-exploiting transformation:
	<pre>
	R(q,x)*(n-1) = Sum_{r: r!=q} Sum_{y: y!=q} Sim(D(r,q), D(x,y)) =
	= C( Nei(q), Nei(x)) + C( All-q-Nei(q), All-x-Nei(x)) =
	= C(All,All) - C(All,x)  - C(q, All) + C(q,x) -
	- C(All,Nei(x)) - C(Nei(q),All) + C(q,Nei(x)) + C(Nei(q),x) +
	+ 2*C(Nei(q), Nei(x)).

	Here, C(set1, set2) = sum_{r in set1, y in set 2} C(r,y);
	All = set of all nodes of the 1st or 2nd graph, as appropriate;
	Nei(q) = set of all nodes of the 1st graph connected to q;
	Far(q) = A - (Nei(q) + {q});
	set1 - set2 = difference of set1 and set2.
	</pre> 
    */
    protected CorrespondenceVector fastInSim( CorrespondenceVector c )     
    {
	CorrespondenceVector inSim =  new CorrespondenceVector(c.n, c.m);
	double [] rsum = c.rowSum(), csum = c.colSum();
	int denom = Math.min( c.n, c.m)-1;
	
	for(int q=0; q<c.n; q++) {
	    for(int x=0; x<c.m; x++) {

		// 1. Sum(All,All) + c(q,x) -C(q,All) -C(All,x) 
		double total = c.allSum() + c.d[q][x] - rsum[q] - csum[x];

		// 2. +C(q,Nei(x)-C(All,Nei(x)) +C(Nei(q),x) -C(Nei(q),All)
		int yList[] = sysN.getNei(x);
		for(int j=0; j<yList.length; j++) {
		    int y = yList[j];
		    total +=  c.d[q][y] - csum[y];
		}
		int rList[] = sysO.getNei(q);
		for(int j=0; j<rList.length; j++) {
		    int r = rList[j];
		    total += c.d[r][x] - rsum[r];
		}

		// 3. +2*C(Nei(q), Nei(x)).
		double s=0;
		for(int j=0; j<rList.length; j++) {
		    int r = rList[j];
		    for(int k=0; k<yList.length; k++) {
			int y = yList[k];
			s +=  c.d[r][y];
		    }
		}
		total += 2 * s;

		inSim.d[q][x]=total / denom;
	    }
	}
	return inSim;
    }

    /** Fast similarity calculation for a general graph.
	
	<pre>

	Here, C(set1, set2) = sum_{r in set1, y in set 2} C(r,y);
	All = set of all nodes of the 1st or 2nd graph, as appropriate;
	Nei(q) = set of all nodes of the 1st graph connected to q;
	Far(q) = A - (Nei(q) + {q});
	set1 - set2 = difference of set1 and set2.



	(n-1)*R(a,x)
	= sum_{b in A-{a}} sum_{y in B-{x}} ( Sim( a,b; x,y) C(b,y) ) 
	= S11(a,x) + S12(a,x) + S21(a,x) + S22(a,x),    with 

	S11(a,x)
	= sum_{b in Nei(a)} sum_{y in Nei(x)} ( Sim(a,b; x,y) C(b,y) )

	S12(a,x)
	= sum_{b in Nei(a)} sum_{y in B-Nei'(x)} (Sim(a,b; x,y) C(b,y))
	= sum_{b in Nei(a) } (Sim( a,b; nolink) C(b, ALL)) -
	sum_{b in Nei(a) } (Sim( a,b; nolink) C(b, Nei(x))) -
	sum_{b in Nei(a) } (Sim( a,b; nolink) C(b,x))  
	= sum_{b in Nei(a)} (Sim(a,b; nolink) C(b,Far(x))

	S21(a,x) 
	= sum_{b in A-Nei'(a) } sum_{y in B } (Sim( a,b; x,y) C(b,y))=
	= sum_{y in Nei(B) } (Sim(nolink; x,y) C(ALL,x)) -
	sum_{y in Nei(B) } (Sim(nolink; x,y) C(Nei(a),x)) -
	sum_{y in Nei(B) } (Sim(nolink; x,y) C(a,y)) 
	= sum_{y in B } (Sim(nolink; x,y) C(Far(a),y))=

	S22(a,x) 
	=  sum_{b in A-Nei'(a) } sum_{y in B-Nei'(x)} (Sim( a,b; x,y) C(b,y))
	= Sim(nolink, nolink) * [
	C(ALL,ALL)     - C(Nei(a), ALL)   - C(a,ALL) 
	- C(ALL, Nei(x)) + C(Nei(a),Nei(x)) + C(a,Nei(x)) 
	- C(ALL, x)      + C(Nei(a),x)      + C(a,x) ]
	= Sim(nolink, nolink) C(Far(a), Far(x))

	</pre>
    */
    protected CorrespondenceVector fastInSimGeneral( CorrespondenceVector c )
    {
	CorrespondenceVector inSim =  new CorrespondenceVector(c.n, c.m);
	int denom = Math.min( c.n, c.m)-1;

	for(int a=0; a<c.n; a++) {
	    int anei[] = sysO.getNei(a);
	    int aneiPlus[] = sysO.getNeiPlus(a);

	    for(int x=0; x<c.m; x++) {
		double total=0;	
		int xnei[] = sysN.getNei(x);
		int xneiPlus[] = sysN.getNeiPlus(x);
		
		// S11= sum_{b in Nei(a)} sum_{y in Nei(x)} (Sim(a,b;x,y) C(b,y))
		for(int j=0; j<anei.length; j++) {
		    int b=anei[j];
		    Link lo = new Link(a,b);
		    for(int k=0; k<xnei.length; k++) {
			int y=xnei[k];
			Link ln = new Link(x,y);
			total += lookupLinkSim(lo, ln) * c.d[b][y];
		    }
		}
				
		// S12(a,x) = sum_{b in Nei(a)} (Sim(a,b; nolink) C(b,Far(x))
		for(int j=0; j<anei.length; j++) {
		    int b=anei[j];
		    Link lo = new Link(a,b);
		    total += lookupLinkSim(lo, Link.EMPTY) * c.sum_1_far(b,x,xneiPlus);
		}
	
		// S21(a,x) = sum_{y in B } (Sim(nolink; x,y) C(Far(a),y))
		for(int k=0; k<xnei.length; k++) {
		    int y=xnei[k];
		    Link ln = new Link(x,y);
		    total += lookupLinkSim(Link.EMPTY, ln) * c.sum_far_1(a,y,aneiPlus);
		}

		// S22(a,x) = Sim(nolink, nolink) C(Far(a), Far(x))
		total += lookupLinkSim(Link.EMPTY, Link.EMPTY) * c.sum_far_far(a,x, aneiPlus, xneiPlus);

		inSim.d[a][x] = total / denom;
		
		/*
		// Corectness check: comparing to the original method
		double other=0;
		for ( int jo=0; jo<c.n; jo++ ) {
		    if ( jo == a ) continue;
		    Link lo = new Link(a, jo);
		    for ( int jn=0; jn<c.m; jn++ ) {
			if ( jn == x )  continue;
			other +=lookupLinkSim(lo, new Link(x,jn))*c.d[jo][jn];
		    }			
		}
		if (Math.abs(other-total)>0.001 && errCnt<=20) {
		    String msg="E["+a+","+x+"]: Orig="+other +", fast="+total;
		    System.out.println(msg);
		    errCnt++;
		    if (errCnt==20) System.out.println(" ... etc....");
		}
		*/
	    }
	}

	return inSim;
    }
    
    /** For testing a different normalization method. */
    /*
    public double getRI_ratio( double inSim[][], double inhib[][] ) 
    {
	final double eps = 1e-6; 
	double a=sumElem(inSim), b=sumElem(inhib);
	// avoid division by zero
	return (Math.abs(b)<= eps) ? 1.0: a/b;
    }
    */

    /** A normalization method that takes both exSim and inSim into
	account. The goal is to have
	sum_{a,b}( (alpha*exSim + beta*inSim - chi*inhib)[a][b]) = 0.
	Learning rate may be recomputed too.
    */
    public void readjustChi(double exSim[][], double inSim[][], double inhib[][])
    {
	final double eps = 1e-6; 
	double sumEx =sumElem(exSim),  
	    sumInSim=sumElem(inSim), sumInhib=sumElem(inhib);
	double sum1 = alpha*sumEx + beta*sumInSim;
	// avoid division by zero
	chi = (Math.abs(sumInhib)<= eps) ? 1.0: sum1/sumInhib;
	System.out.println("Set chi=" + chi);
	if (resetL) setSafeL();
    }

    /** Computes the sum of the elements in an array. */
    public double sumElem( double a[][] ) 
    {
	double s=0;
	for(int i=0; i<a.length; i++) {
	    double b[] = a[i];
	    for(int j=0; j<b.length; j++) s += b[j];
	}
	return s;
    }

    /**
       <pre>
       inhib(q,x) = ( C(All-q,x) + C(q, All-x) ) /(n+m-2)
       </pre>
    */
    protected CorrespondenceVector fastInhib( CorrespondenceVector c )     
    {
	CorrespondenceVector inhib= new CorrespondenceVector(c.n, c.m);
	double [] rsum=c.rowSum(), csum=c.colSum();
	int denom = c.n+ c.m -2;
	for(int q=0; q<c.n; q++) {
	    for(int x=0; x<c.m; x++) {
		inhib.d[q][x]=(rsum[q] +csum[x] - 2* c.d[q][x])/denom;
	    }
	}
	return inhib;
    }


    /** Iterate the network to comppute correspondence between unit pairs. */
    protected void iterateNet()
    {
        SimpleDriver.debugln( "Absurdist: iterating through the network ..." );
	
	int no = systemO.getConceptNum();
	int nn = systemN.getConceptNum();
        int n = Math.min( no, nn );

	//// errCnt=0;
	/*
	   this.systemO = systemO;
	   this.systemN = systemN;
	   inSim = new double[maxStep][no][nn];
	   inhib = new double[maxStep][no][nn];
	   netIn = new double[maxStep][no][nn];
	   int maxStep = maxLoop / intLoop;
	   correspond = new double[maxStep][no][nn];	
	*/

	// initialize all correspondences to preset initial value
        //// double initCorr = 0.5; //// = 1 / ( double )n;
	for ( int io=0; io<no; io++ )
	    for ( int in=0; in<nn; in++ ) 
		correspond[0][io][in] = cInit;

	// iteratively update the corresppondence matrix
	double dc = 0; // 2-norm of the update to the correspondence matrix
	int step, next;
	int loop = 0;
	for ( loop=0; loop < maxLoop-1; loop++ ) {
	    step = loop / intLoop;
	    next = ( loop + 1 ) / intLoop;

	    if ( systemsAreSimple ) {
		// Compute inhibition and excitation in O(N^2 D^4),
		// exploiting the sparseness of the matrix, for 
		// a graph with unweighted unlabeled edges
		CorrespondenceVector c = new CorrespondenceVector(correspond[step]);
		inSim[step] = fastInSim(c).getMatrix();
		inhib[step] = fastInhib(c).getMatrix();
	    } else if (useFastGeneral) {
		// Use the fast method for a General labeled graph
		CorrespondenceVector c = new CorrespondenceVector(correspond[step]);
		inSim[step] = fastInSimGeneral(c).getMatrix();
		inhib[step] = fastInhib(c).getMatrix();		
	    } else {
		// compute excitation and inhibition using the general O(N^4) algorithm, 
		// which will work with arbitrary relations
		for ( int io=0; io<no; io++ ) {
		    for ( int in=0; in<nn; in++ ) {	
			inSim[step][io][in] = 0;
			inhib[step][io][in] = 0;
		    
			// internal similarity between systemO:concept[io] and systemN:concept[in]
			for ( int jo=0; jo<no; jo++ ) {
			    if ( jo == io ) continue;
			    Link lo = new Link(io, jo);
			    for ( int jn=0; jn<nn; jn++ ) {
				if ( jn == in ) continue;
				inSim[step][io][in] += 
				    calcLinkSim( lo, new Link(in, jn) ) *
				    correspond[step][jo][jn];
			    }			
			}
			inSim[step][io][in] += correspond[step][io][in];
			inSim[step][io][in] /= n - 1;

			// inhibition between systemO:concept[io] and systemN:concept[in]
			for ( int jo=0; jo<no; jo++ ) {
			    if ( jo != io ) {
				inhib[step][io][in] += correspond[step][jo][in];
			    }
			}
			 
			for ( int jn=0; jn<nn; jn++ ) {
			    if ( jn != in ) {
				inhib[step][io][in] += correspond[step][io][jn];
			    }
			}
			inhib[step][io][in] /= no + nn - 2;

		    } // end of in loop
		} // end of io loop
	    }

	    if (coefMode==CoefMode.RESET_BALANCE) {
		// Recompute chi, and possibly learning rate.
		// chi = beta * getRI_ratio(inSim[step], inhib[step]);
		readjustChi(exSim, inSim[step], inhib[step]);
	    }
	    // Test if it's a valid value
	    testNaN(chi, "chi");

            dc = 0;
	    // update net input
	    for ( int io=0; io<no; io++ ) {
		for ( int in=0; in<nn; in++ ) {			    
		    if ( useExponential )  { // exponential activator/inhibitor		    
			netIn[step][io][in] = 
			    alpha * exSim[io][in] + 
			    beta * Math.exp( inSim[step][io][in] ) - 
			    chi * Math.exp( inhib[step][io][in] ); 
		    }  else { // linear activator/inhibitor
			netIn[step][io][in] = 
			    alpha * exSim[io][in] + 
			    beta * inSim[step][io][in] - 
			    chi * inhib[step][io][in];
		    }
		}
	    }

	    //double rn[]= elementRange(netIn[step]);
	    //System.out.println("Actual---: N in ["+rn[0]+" : "+rn[1]+"]");

	    for ( int io=0; io<no; io++ ) {
		for ( int in=0; in<nn; in++ ) {	

		    double oc = correspond[step][io][in]; 
		    if (dampingMode == Damping.GR2002) {
			// as per Goldstone and Rogosky paper 2002
			if ( netIn[step][io][in] >= 0 ) {
			    correspond[next][io][in] = oc +
				netIn[step][io][in]*(1-oc)*lRate;
			    if ( correspond[next][io][in] > 1) 
				throw new AssertionError("C=" + correspond[next][io][in] + "> 1.0!");
								   
			}  else {
			    correspond[next][io][in] = oc +
				netIn[step][io][in]*oc*lRate;  
			    if ( correspond[next][io][in] < 0) 
				throw new AssertionError("C=" + correspond[next][io][in] + "< 0.0!");
			}
		    } else if  (dampingMode == Damping.HARD) {
			// Hard damping
			correspond[next][io][in] =oc+netIn[step][io][in]*lRate;
			if (correspond[next][io][in] < 0) {
			    correspond[next][io][in] = 0;
			} else if (correspond[next][io][in] > 1) {
			    correspond[next][io][in] = 1;
			}
		    } else {
			throw new AssertionError("Wrong damping mode");
		    }

		    dc += Math.abs( correspond[next][io][in] - oc );
		    
		    // output correspondence for this loop                    
		} // end of for ( int in=0; in<nn; in++ ) ...
	    } // end of for ( int io=0; io<no; io++ )

            // if converged, i.e. update to C is smaller than threshold, then stop
            dc /= n*n;
            if ( dc < cvgThresh ) {
                System.out.println( "Converged at loop " + loop );
                break;
            }
	    nStep = step + 1;   
            
	    /*
            // addjust correspondence so the sum is n
            double boost = n / sumCorr;
	    for ( int io=0; io<no; io++ ) 
		for ( int in=0; in<nn; in++ ) 	    
                    correspond[io][in] *= boost;        
	    // yield current computation thread to let graphics thread update the scene
	    if ( loop % intLoop == 0 )
	    Thread.yield(); 
	    */
	} // end of for ( int loop=1; loop<5000; loop++ ) ...
	System.out.println( loop + " iterations performed; |N|^2/n^2=" + dc);
	results.iterCnt = loop;
    }


    /** An auxiliary class used during the randm tie-breaking process
	in find2WayMap */
    static class AllBest 
    {
	/** The currently best value */ 
	double max;

	/** A pair of (row,column) indexes */
	static class BestPair {
	    int mo, mn;
	    BestPair(int io, int in) {
		this.mo = io;
		this.mn = in;
	    }
	}

	/** A vector of BestPair instances, listing all occurrences of
            the best value in the currently analyzed submatrix. */
	Vector allBest=new Vector(); 

	AllBest(double max) {
	    this.max = max;
	}

	/** Compare the value at the (io,in) position against the
	    current best value, and decide whether this is the new
	    best value, a tie with already stored values, or neither.
	*/
	void testValue( int io, int in, double val) {
	    if (allBest.size() ==0 || val > max ) {
		// The new max
		max = val;
		allBest.clear();
		allBest.addElement( new BestPair(io,in));
	    } else if ( val == max ) {
		// A tie
		allBest.addElement(new BestPair(io,in));
	    }  
	}

	/** Random tie-breaking.
	@return One of the elements of this vector, chosen at random;
	null if the vector is empty */
	BestPair pickOne() {
	    if (allBest.size() > 1) {
		int k = (int)(Math.random()*allBest.size());
		return (BestPair)allBest.elementAt(k);
	    } else if (allBest.size()==1) {
		return (BestPair)allBest.elementAt(0);
	    } else {
		return null;
	    }
	}
    }


    /** Finds the symmetrical two-way mapping using the N global strongest valid correspondences. 
	<p>
        In this case 1-1 mapping is enforced, and 
	if the two systems are of different size, there will be some concepts with no mappings.
    */
    protected void find2WayMap()
    {
        SimpleDriver.debugln( "Finding two-way mapping ..." );

        int no = systemO.getConceptNum();
        int nn = systemN.getConceptNum();
        int n = Math.min( no, nn );
        int io, in;
	mapON = new int[no];
	mapNO = new int[nn];
        for ( io=0; io<no; io++ ) 
            mapON[io] = -1;
        for ( in=0; in<nn; in++ ) 
            mapNO[in] = -1;
        
        //// int nStep = getStepNum();
        for ( int k=0; k<n; k++ ) {
	    // This vector will store BestPair entries where the
	    // current max is reached.  It will be used for tie-breaking
	    AllBest allBest=new AllBest( mapThresh);
            for ( io=0; io<no; io++ ) {
                if ( mapON[io] != -1 )  continue;
                for ( in=0; in<nn; in++ ) {
                    if ( mapNO[in] != -1 ) continue;
		    allBest.testValue(io, in,  correspond[nStep-1][io][in]);
                }
            }
	    // Break the tie randomly, if needed
	    AllBest.BestPair best = allBest.pickOne();

            if (best !=null ) {
                mapON[best.mo] = best.mn;
                mapNO[best.mn] = best.mo;
                /*
                mapON[mo] = systemN.getConcept(mn);
                mapNO[mn] = systemO.getConcept(mo);
                */
            }
            else break;
        }
    }	       
    
    /** Find the one-way O-N and N-O mapping using the strongest correspondence in each row/column.
	<p>
	In this case all concepts will be mapped even if the two systems are of different size, 
        and it's possible to have 1-n or m-1 mapping.
    */
    protected void find1WayMap()
    {
        SimpleDriver.debugln( "Finding one-way mapping ..." );

        int no = systemO.getConceptNum();
        int nn = systemN.getConceptNum();
        //// int nStep = getStepNum();

	mapON = new int[no];
	for ( int io=0; io<no; io++ ) {	    
            mapON[io] = -1;
	    AllBest  allBest=new AllBest( mapThresh);
	    for ( int in=0; in<nn; in++ ) {
		allBest.testValue(io, in, correspond[nStep-1][io][in]  );
	    }
	    // Break the tie randomly, if needed
	    AllBest.BestPair best = allBest.pickOne();
	    if (best != null) {
		mapON[io] = best.mn;
	    }
	}

	mapNO = new int[nn];
	for ( int in=0; in<no; in++ ) {	    
            mapNO[in] = -1;
	    AllBest  allBest=new AllBest( mapThresh);
	    for ( int io=0; io<no; io++ ) {
		allBest.testValue(io, in, correspond[nStep-1][io][in]  );
	    }

	    AllBest.BestPair best = allBest.pickOne();
	    if (best != null) {
		mapNO[in] = best.mo;
	    }
	}        
    }	       

    /** Evaluates the quality of the mapping in mapON[]. 
        <p>
        Note: mismapped concepts and relations only apply to mappings between
        a  concept system and its noisy copy, assuming that the noise is small
        enough so that the mapping should be identity, i.e. Ci -> Ci'.
    */
    protected void evalMap()
    {
	Measures.Quality q= Measures.evalMap(this, mapON, correspond[nStep-1]);
	results.misCon =q.misCon;
	results.misRel =q.misRel;
	results.accuracy =q.accuracy;

	// Now, compare with the indentity mapping;
	int identity[] = new int[ mapON.length];
	for(int i=0; i<identity.length; i++) {
	    identity[i] = i;
	}
	// To find "adjusted misRel", we compare this permutation's
	// misRel with its counterpart for the identity mapping.
	// It makes no sense to pass the correspondence matrix to
	// this function now and compute "accuracy", but we'll ignore
	// it anyway
	Measures.Quality idQ =  Measures.evalMap(this, identity, correspond[nStep-1]);
	results.misRelAdjId = results.misRel - idQ.misRel;
	System.out.println("Adjusting misRel (vs. idnt): " + results.misRel + " - " + idQ.misRel);

	// What about the BEST permutation?
	if (adjBest) {
	    BranchAndBound bb = new BranchAndBound(this);
	    int mapBest[] = bb.getMap();
	    System.out.println("Branch-and-bound has found an optimal permutation:");
	    for(int i=0; i<mapBest.length; i++) {
		System.out.println("" + i + " --> " + mapBest[i]);
	    }
		
	    Measures.Quality bestQ =  Measures.evalMap(this, mapBest, correspond[nStep-1]);
	    results.misRelAdjOp = results.misRel - bestQ.misRel;
	    System.out.println("Adjusting misRel (vs. best): " + results.misRel + " - " + bestQ.misRel);
	}

    }
    
    /** Maps orginal concept system to the one with added noise, 
	using exponenital/linear activator/inhibitor.
	systemO and systemN have to be set preceding to calling this method.
    */
    public void mapSystem()
    {
        initData();
        status = MAPPING;
        adjustCoeff();
        iterateNet();
	if ( enforce1to1map )
	    find2WayMap();
	else
	    find1WayMap();
	status = MAPPED;
        evalMap();
    }

    /** Trying to do mapSystem with varying chi... */
    /*
    public void mapSystemMultiLevel()
    {
        //// setSystem( systemO, systemN );
        // adjustCoeff();

	alpha = 0;
	beta = 1;
	double maxE = 1; // (FIXME!) I suppose...
	double chiMax = alpha * maxE  + beta * 
	for( chi=1; chi

        status = ITERATING;
        iterateNet();
        status = CONVERGED;
	if ( enforce1to1map )
	    find2WayMap();
	else
	    find1WayMap();
        evalMap();
	status = MAPPED;
    }
    */


    /** Creates a new instance of Absurdist initialized with the 2
	systems read from a file containing Java serialized objects. 
	Such a file may have been produced with saveSystems().
	
	For display purposes, both graph are to be fitted into a w-by-x
	panel, splitting it into 2 halves.
	
	@param w Width of the display area for the graphs.
	@param h Height of the display area for the graphs.
    *
    public static Absurdist readSystems( String fileName, int w, int h ) 
    {
	try {
	    SimpleDriver.debugln( "Absurdist: reading systems from file " + fileName );

	    FileInputStream fis = new FileInputStream( fileName );
	    ObjectInputStream ois = new ObjectInputStream( fis );
	
	    ConceptSystem systemO = ( ConceptSystem )ois.readObject();
	    ConceptSystem systemN = ( ConceptSystem )ois.readObject();

	    ConceptGraph graphO = new ConceptGraph( systemO, 0, 0, w, h/2 );
	    ConceptGraph graphN = new ConceptGraph( systemN, 0, h/2, w, h/2 );

	    ois.close();
	    return new Absurdist( graphO, graphN );
	} catch (Exception e) {
	    SimpleDriver.debugln( "Absurdist: serialization from file " + fileName +
				  " failed: " + e.getMessage() );
	    e.printStackTrace( System.err );
	    return null;
	}	    
    }
    */

    /** Read an Absurdist session from a file containing Java serialized objects. 
	Such a file may have been produced with writeSession().
    */
    public void readSession( String fileName ) throws Exception
    {
	SimpleDriver.debugln( "Absurdist: reading session from file " + fileName );
	FileInputStream fis = new FileInputStream( fileName );
	ObjectInputStream ois = new ObjectInputStream( fis );	
	systemO = ( ConceptSystem )ois.readObject();
	systemN = ( ConceptSystem )ois.readObject();
	ois.close();
    }
    
    /** Writes out the Absurdist session as Java serialized objects into a file. 
	This method can be used if you want to restore the data
	later from the file, using readSession(). 
    */
    public void writeSession( String fileName ) throws Exception
    {
	System.out.println( "Absurdist: writing systems to file " + fileName );
	FileOutputStream fos = new FileOutputStream( fileName );
	ObjectOutputStream oos = new ObjectOutputStream( fos );	    
	oos.writeObject( systemO );
	oos.writeObject( systemN );
	oos.close();
    }

    /** Writes the correspondence matrix to a text file. */
    public void writeCorrespond( String filename ) throws Exception
    {
	SimpleDriver.debugln( "CorFile path: " + filename );

	PrintStream out = new PrintStream( new FileOutputStream( filename ), true );
	out.println( "\nCorrespondence matrix for System " + systemO.name  + " --> System " + systemN.name );
	DecimalFormat df = new DecimalFormat( "0.0000" );
        int no = systemO.getConceptNum();
        int nn = systemN.getConceptNum();

        for ( int io=0; io<no; io++ ) {
            for ( int in=0; in<nn; in++ ) {
		double c = correspond[nStep-1][io][in];
		String cstr = df.format(c);
		if ( cstr.equals("?") )
		    cstr = "" + c; // Uh-oh...
		out.print( cstr + " " );
	    }
	    out.println();        
	}
	
	out.close();
    }
     
    /** Writes the mapping between systemO and systemN to a text file. */
    public void writeMap( String filename ) throws Exception
    {
	SimpleDriver.debugln( "CorFile path: " + filename );

	PrintStream out = new PrintStream( new FileOutputStream( filename ), true );

	// mapping evaluation
        out.println( "\nMapped System " + systemO.name + " --> System " + systemN.name );
	out.println( "Mismapped concepts number (only applies to identity mpg)  : " + results.misCon );
	out.println( "Mismapped relations (sum of link differences)             : " + results.misRel );
	out.println( "Mismapped relations (adjusted by identity mapping)        : " + results.misRelAdjId );
	out.println( "Mismapped relations (adjusted by optimal mapping)         : " + results.misRelAdjOp );
	out.println( "Mapping accuracy (based on link simlrt and correspondence): " + results.accuracy );
        out.println( "------------------------------------------------------------------" );

	// mapped concept pairs
        int no = systemO.getConceptNum();
	DecimalFormat df = new DecimalFormat( "0.0000" );
        for ( int io=0; io<no; io++ ) {
            int in = mapON[io];
            if ( in != -1 ) 
                out.println( "[" + io + "] " + systemO.getConcept(io).name + 
				    " --> [" + in + "] " + systemN.getConcept(in).name + 
				    " : " + df.format( correspond[nStep-1][io][in] ) );
            else 
		out.println( "[" + io + "] " + systemO.getConcept(io).name + " --> null" );
	}
    }     

   
    /** Prints the source system to the standard output. */
    public void printSystemO()
    {
        System.out.println( "\n" + systemO.toString() );
    }
    
    /** Prints the target system to the standard output. */
    public void printSystemN()
    {
        System.out.println( "\n" + systemN.toString() );
    }

    /** Prints the two systems to the standard output. */
    public void printSystems()
    {
	String strO = "", strN = "";
	if ( systemO instanceof ConceptSystem ) {
	    strO = systemO.toString();
	    strN = systemN.toString();
	}
	else {
	    strO = ( ( ConceptGraph )systemO ).toString();
	    strN = ( ( ConceptGraph )systemN ).toString();
	}

        System.out.println( "\n" + strO );
        System.out.println( "\n" + strN );
    }
    
    /** Prints the external similarity matrix to the standard output. */
    public void printExSim()
    {
	System.out.println( "\nExternal similarity matrix -----------" );
	DecimalFormat df = new DecimalFormat( "0.0000" );
        int no = systemO.getConceptNum();
        int nn = systemN.getConceptNum();

        for ( int io=0; io<no; io++ ) {
            for ( int in=0; in<nn; in++ ) {
		double e = exSim[io][in];
		String estr = df.format(e);
		if ( estr.equals("?") )
		    estr = "" + e; // Uh-oh...
                System.out.print( estr + " " );
	    }
	    System.out.println();
        }
    }

    /** Prints the correspondence matrix to the standard output. */
    public void printCorrespond()
    {
	System.out.println( "\nCorrespondence matrix -----------" );
	DecimalFormat df = new DecimalFormat( "0.0000" );
        int no = systemO.getConceptNum();
        int nn = systemN.getConceptNum();

        for ( int io=0; io<no; io++ ) {
            for ( int in=0; in<nn; in++ ) {
		double c = correspond[nStep-1][io][in];
		String cstr = df.format(c);
		if ( cstr.equals("?") )
		    cstr = "" + c; // Uh-oh...
                System.out.print( cstr + " " );
	    }
	    System.out.println();
        }
    }
    
    /** Prints the mapping between systemO and systemN to the standard output. */
    public void printMap()
    {
	// mapping evaluation
        System.out.println( "\nMapped System " + systemO.name + " --> System " + systemN.name );
	System.out.println( "Mismapped concepts number (only applies to identity mpg)  : " + results.misCon );
	System.out.println( "Mismapped relations (sum of link differences)             : " + results.misRel );
	System.out.println( "Mismapped relations (adjusted by identity mapping)        : " + results.misRelAdjId );
	System.out.println( "Mismapped relations (adjusted by optimal mapping)         : " + results.misRelAdjOp );



	System.out.println( "Mapping accuracy (based on link simlrt and correspondence): " + results.accuracy );
        System.out.println( "------------------------------------------------------------------" );	

	// mapped concept pairs
        int no = systemO.getConceptNum();
	DecimalFormat df = new DecimalFormat( "0.0000" );
        for ( int io=0; io<no; io++ ) {
            int in = mapON[io];
            if ( in != -1 ) 
                System.out.println( "[" + io + "] " + systemO.getConcept(io).name + 
				    " --> [" + in + "] " + systemN.getConcept(in).name + 
				    " : " + df.format( correspond[nStep-1][io][in] ) );
            else 
		System.out.println( "[" + io + "] " + systemO.getConcept(io).name + " --> null" );
	}
    }     

    /** Prints the source and target concept systems,
	the correspondence matrix, and the final mappings. */
    public void print() throws Exception
    {
	printSystems();
	printExSim();
	printCorrespond();
	printMap();
    }    


    /** Checks if the value contains NaN, and aborts the program if
	so.  You can use this method if you suspect that some values
	are corrputed.  If the value is a NaN, an error message is
	printed and the program exits.

	@param x The value to check
	@param name The "name" of the variable we are checking, to print
	in an error message.
    */
    void testNaN(double x, String name) {
	if (Double.isNaN(x)) {
	    System.err.println(name + " is NaN");
	    System.exit(1);
	}
    }

}
