///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// ABSURDIST II
// Ying Feng
// Computer Science, Indiana University
// Feb-Aug 2003
//
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////



package absurdist;

import java.text.*;
import java.util.*;
import java.io.Serializable;
 
import driver.*;



/** Class Concept
    specifies the data structure for a concept, which contains its name and some features.
*/
public class Concept implements Serializable
{
    /** coefficient for category distance */
    public static final double DISC = 0.3;

    /** coefficient for rank distance */
    public static final double DISR = 0.3;

    /** coefficient for dimension value distance */
    public static final double DISD = 0.4;

    /** maximum rank */
    public static final int RANK = 10;

    /**  Unique identifier of this concept.
     */
    public String name = "";
    /**  Nominal attribute, can be used to restrict mapping range.
     */
    public String category = "";
    /**  Quantitative attribute in the range of [1, RANK]; can be used to decide mapping priority.
     */
    public int rank = 0;
    /**  2D normalized values; can be extended to n-ary attribute vector.
     */
    public double d1 = 0, d2 = 0;
    //// Node node = null;	      // corresponding node
    
    /** Creates an empty concept. 
     */
    public Concept()
    {
    }

    /** Creates a concept with specified index and category and default features. 
     */
    public Concept( int i, String category )
    {
	this.name = "C" + i;
        this.category = category;
        rank = RANK;
	d1 = 0;
	d2 = 0;
	/*
        d1 = 0.2 + 0.6 * Math.random();
        d2 = 0.2 + 0.6 * Math.random();
	*/
    }

    /** Creates a concept with specified index and random features if indicated.
     */
    public Concept( int i, Vector categories, boolean ranked, boolean valued )
    {
	this.name = "C" + i;
	int catIdx = ( int )( categories.size() * Math.random() );
	category = ( String )categories.get( catIdx );
	if ( ranked )
	    rank = ( int )( RANK * Math.random() );
	else
	    rank = RANK;
	if ( valued ) { // optional restriction to keep away from the range border
	    d1 = 0.1 + 0.9 * Math.random();
	    d2 = 0.1 + 0.9 * Math.random();
	}
	else {
	    d1 = 0;
	    d2 = 0;
	}
	//// this.node = node;
    }

    /** Creates a concept with full features.
     */
    public Concept( String name, String category, int rank, double d1, double d2 )
    {
	this.name = name;
        this.category = category;
        this.rank = rank;
        this.d1 = d1;
        this.d2 = d2;
        //// this.node = node;
    }

    /** Creates a copy of a concept.
     */
    public Concept( Concept concept )
    {
	name = concept.name + "'";
        category = concept.category;
        rank = concept.rank;
        d1 = concept.d1;
        d2 = concept.d2;
    }

    /** Creates a copy of a concept with some noise added to the feature values.
     */
    public Concept( Concept concept, double noise )
    {
	this.name = concept.name + "'";
        category = concept.category;
        rank = concept.rank;
	d1 = concept.d1;
	d2 = concept.d2;
	if ( noise == 0 ) return;

        d1 += noise * ( 2 * Math.random() - 1 );
        d2 += noise * ( 2 * Math.random() - 1 );
	// keep within range and optionally away from the border
	if ( d1 < 0.1 )
	    d1 = 0.1;
	if ( d1 > 0.9 )
	    d1 = 0.9;
	if ( d2 < 0.1 )
	    d2 = 0.1;
	if ( d2 > 0.9 )
	    d2 = 0.9;
    }

    /** Adds some noise to this concept's feature values.
     */
    public void addNoise( double noise )
    {
	if ( noise == 0 ) return;

        d1 += noise * ( 2 * Math.random() - 1 );
        d2 += noise * ( 2 * Math.random() - 1 );
	// keep within range and optionally away from the border
	if ( d1 < 0.1 )
	    d1 = 0.1;
	if ( d1 > 0.9 )
	    d1 = 0.9;
	if ( d2 < 0.1 )
	    d2 = 0.1;
	if ( d2 > 0.9 )
	    d2 = 0.9;
    }

    /*
    // Encode this node with a visual graph node inside the frame.
    public encode( Rectangle frame )
    {
        Color color = Color.red();
        int size = Node.D + rank * Node.D / 4;
        int x = frame.getX() + d1 * frame.getWdith();
        int y = frame.getY() + d2 * frame.getHeight();
        node = new Node( this, color, size, x, y );                
    }
    */
    
    /** Checks whether this concept is ranked, i.e. has a non-default rank.
     */
    public boolean isRanked()
    {
	return ( rank != 0 && rank != RANK );
    }

    /** Checks whether this concept is valued, i.e. has non-zero feature values.
     */
    public boolean isValued()
    {
	return ( d1 != 0 || d2 != 0 );
    }

    /** Computes the normalized semantic distance between this concept and the specified one, 
	based on their category, rank, and feature values.
     */
    public double distance( Concept concept )
    {
	double dd1 = d1 - concept.d1;
	double dd2 = d2 - concept.d2;
	double disD = Math.sqrt( dd1 * dd1 + dd2 * dd2 ) / Math.sqrt(2);

	double disC = category.equals( concept.category ) ? 0 : 1;
	double disR = Math.abs( rank - concept.rank ) / (double)RANK;

	double dis = DISC * disC + DISR * disR + DISD * disD;
	return dis;
    }

    /** Returns a string representation of this concept.
     */
    public String toString()
    {
	DecimalFormat df = new DecimalFormat( "0.0000" );
	String str = "Concept " + name + ":\t" + category + "\t" + rank + "\t" +
	    "( " + df.format(d1) + ", " + df.format(d2) + " )";
	return str;
    }
}
