///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// ABSURDIST II
// Ying Feng
// Computer Science, Indiana University
// Feb-Aug 2003
//
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////



package absurdist;

import java.util.*;
import java.io.*;

import driver.*;
import graph.*;



/** Class ConceptSystem
    is the data model for a concept system specifying the semantic network between the concepts.     
    It contains all concept categoies and relation models used in the system,
    as well as all its concepts and relations.
    Note: for directed graph, if there's a relation from concept i to concept j,
    then this relation is recorded in the relation hashtable with key "i-j", 
    and inside the relation, the two concepts go in the order of first i then j;
    for undirected graph, a relation between concept i and j is converted into 
    two directed relation from i to j and from j to i.
*/
public class ConceptSystem implements Serializable
{
    // Direction modes for relation models.
    /** Direction mode for undirected relation models. */
    public final static int UNDIRECTED = 0;
    /** Direction mode for directed relation models. */
    public final static int DIRECTED = 1;
    /** Direction mode for a mixture of directed and undirected relation models. */
    public final static int MIXED = 2;

    /**  Identifier of the system.  */
    public String name = "";
    /**  All concept categories involved in the system. */
    public Vector categories = new Vector();
    /**  All relation models used in the system. */
    public Vector models = new Vector();
    /**  All concepts in the system. */
    public Vector concepts = new Vector();

    /** All relation instances in the system.
	Each hashtable slot cooresponds to a concept pair and is keyed by the combaination of their indices.
	Each slot stores a list that contains various labeled relations existing between the concept pair.
	Keys of the table are obtained by getKey().
	Note: An undirected relation is stored in both slots that correspond to its concept pair
	in each direction correspondingly. */
    public Hashtable relations = new Hashtable(); 

    /** This is a vector that is either null (if no anchors are in use),
	or is of the same length as "concepts". The i-th
	element is either a null, or  a Vector of Strings; each
	component of that is the name of a leaf node that had been
	attached to the i-th internal node (concept) prior to the
	graph transformation done by splitAnchors.pl. This information
	comes from the "Anchors" section of the input text file, and
	is non-empty only in the case of "tree mapping" (July 29, 2003) 
    */
    public Vector anchors = null;


    /** Creates an empty concept system. */
    public ConceptSystem()
    {
    }

    /** Creates a concept systems with specified contents. */
    public ConceptSystem( String name, Vector categories, Vector models, Vector concepts, Hashtable relations )
    {
        this.name = name;
	this.categories = categories;
	this.models = models;
	this.concepts = concepts;
	this.relations = relations;
    }

    /** Creates a concept system from a graph instance. */
    public ConceptSystem( Graph graph )
    {
	int i, j, k;
	int n = graph.getNodeNum();
        this.name = "RandomSystem";        

	// generate default concept category and relation model
	categories.add( "Default" );
	models.add( new RelationModel( "Default", false ) );

	// generate default concepts
	for ( i=0; i<n; i++ )
	    concepts.add( new Concept( i, getCategory(0) ) );
  
	// convert the graph links into system relations
	RelationModel model = getModel(0);
	for ( i=0; i<n; i++ ) {
	    for ( k=0; k < graph.getNodeDegree(i); k++ ) {
		j = graph.getNodeNeigh( i, k );
		Relation relation =  new Relation( model, ( Concept )concepts.get(i), ( Concept )concepts.get(j) );
		putRelation( i, j, relation );
		// graph stores links in both neightbour lists, so does system in both slots in the hashtable
	    }
	}	    	    
    }

    /**
       Creates a concept system represented as a random simple graph, 
       with default concepts and default relations.
       @param n Number of concepts in the system.
       @param fullness Percentage of relations among the total possible relation links in the system.
       The value of fullness ranges between [0,1].
    */
    public ConceptSystem( int n, double fullness ) throws Exception
    {
	int i, j, k;
        this.name = "RandomSystem";        

	// generate default concept category and relation model
	categories.add( "Default" );
	models.add( new RelationModel( "Default", false ) );

	// generate default concepts
	for ( i=0; i<n; i++ )
	    concepts.add( new Concept( i, getCategory(0) ) );

	// genreate asymmetric graph
	double d = (n-1) * fullness;
	Graph graph = GenGraphs.genAsymGraph( n, d );	    

	// convert the graph links into system relations
	RelationModel model = getModel(0);
	for ( i=0; i<n; i++ ) {
	    for ( k=0; k < graph.getNodeDegree(i); k++ ) {
		j = graph.getNodeNeigh( i, k );
		Relation relation =  new Relation( model, ( Concept )concepts.get(i), ( Concept )concepts.get(j) );
		putRelation( i, j, relation );
		// graph stores links in both neightbour lists, so does system in both slots in the hashtable
	    }
	}  
    }

    /**
       Creates a concept system represented as a random graph, 
       categorized/randked/valued, directed/weighted/labeled as indicated, 
       with concepts of random categories and ranks, and randomly weighted/labeled relations.  
       @param n Number of concepts in the system.
       @param nCat Number of concept categories in the system.
       @param ranked Whether the concepts are rankded.
       @param valued Whether the concepts have feature values.
       @param fullness Percentage of relations among the total possible relation links in the system.
       @param nLabel Number of different relation labels in the system.
       @param directMode Whether the system has directed, undirected, or both types of relations.
       @param weighted Whether the relations are weighted.
    */
    public ConceptSystem( int n, int nCat, boolean ranked, boolean valued,
			  double fullness, int nLabel, int directMode, boolean weighted ) throws Exception
    {
	int i, j, k;
	this.name = "RandomSystem";

	// generate concept categories
	for ( i=0; i < nCat; i++ ) {
	    String category = "Cat" + i;
	    categories.add( category );
	}

	// generate relation models: even numbered as undirected, odd numbered as directed
	boolean directed = false;
	if ( directMode == DIRECTED )
	    directed = true;
	for ( i=0; i < nLabel; i++ ) {
	    if ( directMode == MIXED )
		directed = ( i%2 != 0 );
	    RelationModel model = new RelationModel( "R"+i, directed );
	    models.add( model );
	}

	// randomly generate concepts
	for ( i=0; i<n; i++ ) 
	    concepts.add( new Concept( i, categories, ranked, valued ) );

	// for unlabeled, undirected, unweighted simple system, convert from asymmetric graph
	if ( nLabel == 1 && directMode == UNDIRECTED && !weighted ) {
	    // genreate asymmetric graph
	    double d = (n-1) * fullness;
	    Graph graph = GenGraphs.genAsymGraph( n, d );	    
	    
	    // convert the graph links into system relations
	    RelationModel model = getModel(0);
	    for ( i=0; i<n; i++ ) {
		for ( k=0; k < graph.getNodeDegree(i); k++ ) {
		    j = graph.getNodeNeigh( i, k );
		    Relation relation =  new Relation( model, ( Concept )concepts.get(i), ( Concept )concepts.get(j) );
		    putRelation( i, j, relation );
		    // graph stores links in both neightbour lists, so does system in both slots in the hashtable
		}
	    }  

	    return;
	}

	// randomly generate relations, with a possibility of fullness		    
	Random random = new Random();
	for ( i=0; i<n; i++ ) {
	    for ( j=i+1; j<n; j++ ) {
		Vector rellist1 = new Vector();
		Vector rellist2 = new Vector();
		for ( k=0; k < getModelNum(); k++ ) {	
		    Relation relation = null;
		    if ( random.nextDouble() <= fullness ) { // hit an out-edge		
			relation = new Relation( getModel(k), getConcept(i), getConcept(j), weighted );
			rellist1.add( relation );
		    }		
		    if ( !getModel(k).directed ) {
			if ( relation != null ) // add its symmetric relation to the slots for the other direction
			    rellist2.add( relation.symRelation() ); 
		    }
		    else if ( random.nextDouble() <= fullness ) { // hit an in-edge
			relation = new Relation( getModel(k), getConcept(j), getConcept(i), weighted );
			rellist2.add( relation );
		    }
		}
		if ( rellist1.size() > 0 )
		    putRelations( i, j, rellist1 );
		if ( rellist2.size() > 0 )
		    putRelations( j, i, rellist2 );
	    }
	}
    }

    /** Creates a copy of a concept system. */
    public ConceptSystem( ConceptSystem system )
    {
	int n = system.getConceptNum();
        name = system.name + "'";
	categories = system.categories;
	models = system.models;

	// copy all concepts
	for ( int i=0; i<n; i++ ) 
	    concepts.add( new Concept( system.getConcept(i) ) );	

	// copy all relations
        Enumeration enum = system.relations.keys();        
        while ( enum.hasMoreElements() ) {
	    String key = ( String )enum.nextElement();
	    Vector syslist = ( Vector )system.relations.get( key );
	    Vector rellist = new Vector();
	    for ( int i=0; i < syslist.size(); i++ ) {		
		Relation sysrel = ( Relation )syslist.get(i);
		int idx1 = system.getConceptIndex( sysrel.getConcept1() );
		int idx2 = system.getConceptIndex( sysrel.getConcept2() );
		Relation relation = new Relation( sysrel, getConcept( idx1 ), getConcept( idx2 ) );   
		rellist.add( relation );
	    }
	    relations.put( key, rellist );                    
        }
    }

    /**
       Creates another concept system by adding some noise to an existing system. 
       All noise values are in the range of [0, 1]). 

       This is how it works:
       <ul>
       <li> Existing edges may be deleted or inserted with the noiseRC probability
       <li> Weight of all existing edges changes; the weight change is within noiseRI (that is,
       if noiseRI=0, there is no weight change)
       </ul>

       @param noiseCC Concept noise coverage, i.e. percentage of concepts that are affected
       @param noiseCI Concept noise intensity, i.e. amount of changes to concept quantitative values (d1 & d2) 
       @param noiseRC Relation noise coverage, i.e. percentage of relations that are changed and even inverted
       @param noiseRI Relation noise intensity, i.e. amount of change to relation weight
    */
    public ConceptSystem( ConceptSystem system, double noiseCC, double noiseCI, double noiseRC, double noiseRI ) 
    {
	int i, j, k;
	int n = system.getConceptNum();
        name = system.name + "'";
	categories = system.categories;
	models = system.models;

	// randomly choose concepts to change their quantitative values
	boolean valued = system.isValued();
	for ( i=0; i<n; i++ ) {
	    Concept syscon = system.getConcept(i);
	    Concept concept = null;
	    if ( valued && Math.random() <= noiseCC )
		concept = new Concept( syscon, noiseCI );
	    else 
		concept = new Concept( syscon );
	    concepts.add( concept );			
	}

	// randomly add/delete edges and change their weights
	boolean weighted = system.isWeighted();
	for ( i=0; i<n; i++ ) {
	    for ( j=i+1; j<n; j++ ) {
		Vector syslist1 = system.getRelations( i, j );
		Vector syslist2 = system.getRelations( j, i );
		Vector rellist1 = new Vector();
		Vector rellist2 = new Vector(); 
 
		for ( k=0; k < getModelNum(); k++ ) {	
		    Relation sysrel = getRelation( syslist1, getModel(k).label );
		    Relation relation  = 
			system.noisyReplace( sysrel, getModel(k), getConcept(i), getConcept(j), 
					     weighted, noiseRC, noiseRI );		    
		    if ( relation != null ) 
			rellist1.add( relation );

		    // for undirected model, add a symmetric relation to the slots for the other direction
		    if ( !getModel(k).directed ) { 
			if ( relation != null ) 
			    rellist2.add( relation.symRelation() ); 	 
		    } else {
			// for directed model repeat the same procedure in the other direction
			sysrel = getRelation( syslist2, getModel(k).label );
			relation = 
			    system.noisyReplace( sysrel, getModel(k), getConcept(j), getConcept(i), 
						 weighted, noiseRC, noiseRI );
			if ( relation != null )
			    rellist2.add( relation );
		    }		
		} // end of for ( k= ... )

		if ( rellist1.size() > 0 )
		    putRelations( i, j, rellist1 );
		if ( rellist2.size() > 0 )
		    putRelations( j, i, rellist2 );
	    }  // end of for ( j= ... )
	} // of for ( i= ... )
    }
        
    /* // Old version; replaced Oct 22, 2003
    public ConceptSystem( ConceptSystem system, double noiseCC, double noiseCI, double noiseRC, double noiseRI ) 
    {
	int i, j, k;
	int n = system.getConceptNum();
        name = system.name + "'";
	categories = system.categories;
	models = system.models;

	// randomly choose concepts to change their quantitative values
	boolean valued = system.isValued();
	for ( i=0; i<n; i++ ) {
	    Concept syscon = system.getConcept(i);
	    Concept concept = null;
	    if ( valued && Math.random() <= noiseCC )
		concept = new Concept( syscon, noiseCI );
	    else 
		concept = new Concept( syscon );
	    concepts.add( concept );			
	}

	// randomly add/delete edges and change their weights
	boolean weighted = system.isWeighted();
	Random random = new Random();
	for ( i=0; i<n; i++ ) {
	    for ( j=i+1; j<n; j++ ) {
		Vector syslist1 = system.getRelations( i, j );
		Vector syslist2 = system.getRelations( j, i );
		Vector rellist1 = new Vector();
		Vector rellist2 = new Vector(); 
 
		for ( k=0; k < getModelNum(); k++ ) {	
		    Relation relation = null;
		    Relation sysrel = getRelation( syslist1, getModel(k).label );
		    boolean rHit = sysrel != null;                  // relation exists	     
		    boolean nHit = random.nextDouble() <= noiseRC;  // hit a relation noise
		    boolean nnHit = random.nextDouble() <= noiseRC; // hit a further relation noise for inverting
		    boolean neHit = nHit && ( nnHit || !weighted ); // hit a relation existance invertion noise
		    boolean nwHit = nHit && !nnHit && weighted ;    // hit a relation weight change noise
		    if ( neHit && !rHit ) { // add a new relation 
			relation = new Relation( getModel(k), getConcept(i), getConcept(j), weighted );
			rellist1.add( relation );
		    }		
		    else if ( nwHit ) { // add noise when copy existing or new relation
			if ( !rHit ) // create a dummy relation with 0 weight to limit new relation weight below noise
			    sysrel = new Relation( getModel(k), getConcept(i), getConcept(j), 0 );
			relation = new Relation( sysrel, getConcept(i), getConcept(j), noiseRI );	  
			rellist1.add( relation );
		    }		
		    else if ( !nHit && rHit ) { // copy existing relation
			relation = new Relation( sysrel, getConcept(i), getConcept(j) );	  
			rellist1.add( relation );
		    }

		    // for undirected model, add the generated relation to the slots for the other direction
		    if ( !getModel(k).directed ) { 
			if ( relation != null )
			    rellist2.add( relation ); 	 
			continue; // no need to check the other direction
		    }
		    
		    // for directed model repeat the same procedure in the other direction
		    sysrel = getRelation( syslist2, getModel(k).label );
		    rHit = sysrel != null;                  // relation exists	     
		    nHit = random.nextDouble() <= noiseRC;  // hit a relation noise
		    nnHit = random.nextDouble() <= noiseRC; // hit a further relation noise for inverting
		    neHit = nHit && ( nnHit || !weighted ); // hit a relation existance invertion noise
		    nwHit = nHit && !nnHit && weighted ;    // hit a relation weight change noise
		    if ( neHit && !rHit ) { // add a new relation 
			relation = new Relation( getModel(k), getConcept(j), getConcept(i), weighted );
			rellist2.add( relation );
		    }		
		    else if ( nwHit ) { // add noise when copy existing or new relation
			if ( !rHit ) // create a dummy relation with 0 weight to limit new relation weight under noise
			    sysrel = new Relation( getModel(k), getConcept(j), getConcept(i), 0 );
			relation = new Relation( sysrel, getConcept(j), getConcept(i), noiseRI );	  
			rellist2.add( relation );
		    }		
		    else if ( !nHit && rHit ) { // copy existing relation
			relation = new Relation( sysrel, getConcept(j), getConcept(i) );	  
			rellist2.add( relation );
		    }
		} // end of for ( k= ... )

		if ( rellist1.size() > 0 )
		    putRelations( i, j, rellist1 );
		if ( rellist2.size() > 0 )
		    putRelations( j, i, rellist2 );
	    }  // end of for ( j= ... )
	} // of for ( i= ... )
    }
    */


    /** Takes an existing or a non-existent (null) relation in this conceptSystem, 
	and returns a relation that is to be its replacement in a graph with noise
	to relation weights or topology.
	@param sysrel An existing relation (or null, if one existed between
	              the two concepts in question) from the original system
	@param noiseRC probability of dropping/creation 
	@param noiseRI the max change of weight for weighted relations
	@return A new relation to appear between the two concepts in the noisy system. 
	        May be null if the new system should not have a relation here 
    */ 
    private Relation noisyReplace( Relation sysrel, RelationModel model, Concept c1, Concept c2, 
				   boolean weighted, double noiseRC, double noiseRI )  {
	Relation relation = null;
	boolean rHit = (sysrel != null);          // relation exists	     
	boolean nHit = Math.random() <= noiseRC;  // edge will be dropped or created
	
	if ( rHit) {
	    // Existing relations:
	    if (nHit) {
		// Drop relation (= do not copy it)
	    } else {
		// Copy the relation, with some weight change (if weighted)
		relation = weighted?
		    new Relation( sysrel, c1, c2, noiseRI ):
		    new Relation( sysrel, c1, c2);	  
	    }
	} else {
	    // No relation existed: 
	    if (nHit) {  // add a new relation 
		relation = new Relation( model, c1, c2, weighted );
	    }
	}		
	return relation;
    }

    /** Takes an existing relation in this conceptSystem, and returns a relation that 
	is to be its replacement in a graph with noise.
	@param sysrel An existing relation (or null, if one existed between
	              the two concepts in question) from the original system
	@param noiseRC probability of dropping/creation 
	@param noiseRI the max change of weight for weighted relations
	@return A new relation to appear between the two concepts in the noisy system. 
	        May be null if the new system should not have a relation here 
    *
    private Relation noisyRewire( Relation sysrel, RelationModel model, Concept c1, Concept c2, 
				  boolean weighted, double noiseRC, double noiseRI )  {
	Relation relation = null;
	boolean rHit = (sysrel != null);          // relation exists	     
	boolean nHit = Math.random() <= noiseRC;  // edge will be dropped or created
	
	if ( rHit) {
	    // Existing relations:
	    if (nHit) {
		// Drop relation (= do not copy it)
	    } else {
		// Copy the relation, with some weight change (if weighted)
		relation = weighted?
		    new Relation( sysrel, c1, c2, noiseRI ):
		    new Relation( sysrel, c1, c2);	  
	    }
	} else {
	    // No relation existed: 
	    if (nHit) {  // add a new relation 
		relation = new Relation( model, c1, c2, weighted );
	    }
	}		
	return relation;
    }
    */

    /** Adds some noise to this system, inluding noise to concept attributes and
	relation weights or topology, according to noise coverage and intensity level.
	All noise are in the range of [0, 1].
	@param noiseCC Concept noise coverage, i.e. percentage of concepts that are affected
	@param noiseCI Concept noise intensity, i.e. amount of changes to concept quantitative values (d1 & d2) 
	@param noiseRC Relation noise coverage, i.e. percentage of relations that are changed and even inverted
	@param noiseRI Relation noise intensity, i.e. amount of change to relation weight
    */
    public void addNoise( double noiseCC, double noiseCI, double noiseRC, double noiseRI ) 
    {
	int i, j, k;
	int n = getConceptNum();

	// randomly choose concepts to change their quantitative values
	boolean valued = isValued();
	for ( i=0; i<n; i++ ) {
	    if ( valued && Math.random() <= noiseCC )
		getConcept(i).addNoise( noiseCI );			
	}

	// randomly add/delete edges and change their weights
	boolean weighted = isWeighted();
	Random random = new Random();
	for ( i=0; i<n; i++ ) {
	    for ( j=i+1; j<n; j++ ) {
		Vector syslist1 = getRelations( i, j );		
		Vector syslist2 = getRelations( j, i );
		if ( syslist1 == null )
		    syslist1 = new Vector();
		if ( syslist2 == null )
		    syslist2 = new Vector();
		Vector rellist1 = new Vector();
		Vector rellist2 = new Vector();

		for ( k=0; k < getModelNum(); k++ ) {	
		    Relation sysrel = getRelation( syslist1, getModel(k).label );
		    Relation relation  = noisyReplace( sysrel, getModel(k), getConcept(i), getConcept(j), 
						       weighted, noiseRC, noiseRI );		    
		    if ( relation != null ) {
			rellist1.add( relation );
		    }

		    // for undirected model, add the generated relation to the slots for the other direction
		    if ( !getModel(k).directed ) { 
			if ( relation != null ) 			
			    rellist2.add( relation.symRelation() ); 	 
		    } else {
			// for directed model repeat the same procedure in the other direction
			sysrel = getRelation( syslist2, getModel(k).label );
			relation  = noisyReplace( sysrel, getModel(k), getConcept(j), getConcept(i), 
						  weighted, noiseRC, noiseRI );
			if ( relation != null ) 
			    rellist2.add( relation );
		    }		
		} // end of for ( k= ... )

		// Replace the existing relations bundle with the new one
		removeRelations(i,j);
		if ( rellist1.size() > 0 )
		    putRelations( i, j, rellist1 );

		removeRelations(j,i);
		if ( rellist2.size() > 0 )
		    putRelations( j, i, rellist2 );

	    }  // end of for ( j= ... )
	} // of for ( i= ... )
    }

    /** Adds some noise to this system, inluding changing some relation weights 
	to random values and rewire some relations.
	@param noiseWT percentage of relations whose weights will be changed to random values 
	@param noiseRW percentage of relations that will be rewired on one end
    */
    public void addNoise( double noiseWT, double noiseRW ) 
    {
	int n = getConceptNum();
	Enumeration enum = relations.elements();

	// for all existing relations, randomly change their weights and
	// switch one of the two ends to a new end
	// note that because of the sequential way of rewiring, 
	// rewired relations may be rewired again
	while ( enum.hasMoreElements() ) {
	    Vector rellist = ( Vector )enum.nextElement();

	    for ( int k=0; k < rellist.size(); k++ ) { // loop through all existing relation labels
		Relation relation = ( Relation )rellist.get(k);
		boolean wtHit = Math.random() <= noiseWT;  // weight will be changed
		boolean rwHit = Math.random() <= noiseRW;  // relation will be rewired	

		if ( wtHit )
		    relation.addNoise();

		if ( rwHit ) {
		    String label = relation.getLabel();
		    boolean directed = relation.isDirected();
		    int i = getConceptIndex( relation.getConcept1() );
		    int j = getConceptIndex( relation.getConcept2() );
		    boolean rwout = Math.random() <= 0.5; // 50% chance to rewire outgoing end, 50% otherwise
		    boolean found = false; // found a new end
		    Relation newrel = null;

		    // to find a random new end between which and the unchanged end 
		    // there was no existing relation, we choose a random point to 
		    // start searching until such an end is found
		    int start = (int)( Math.random() * n ); // index of the start searching point 
		    int rwidx = 0; // index of the rewired end found
		    if ( rwout ) {
			for ( int count=0; count<n; count++ ) { // search through all n points
			    rwidx = ( start + count ) % n; // choose random end among all concepts
			    found = ( rwidx != i ) && ( rwidx != j ) && !containsRelation( i, rwidx, label );
			    if ( found )
				break;
			}
			if ( found ) {			    
			    newrel = new Relation( relation, relation.getConcept1(), getConcept( rwidx ) );
			    removeRelation( i, j, relation ); // remove the original relation
			    putRelation( i, rwidx, newrel ); // add new relation
			    /*
			    SimpleDriver.debugln( "----- Rewired outgoing relation ----- " );
			    SimpleDriver.debugln( "remove ( " + i + ", " + j + " ): " + relation );
			    SimpleDriver.debugln( "add ( " + i + ", " + rwidx + " ): " + newrel );
			    */
			    if ( !directed ) {
				removeRelation( j, i, relation.getLabel() );
				putRelation( rwidx, i, newrel.symRelation() ); 
				/*
				SimpleDriver.debugln( "remove sym( " + j + ", " + i + " ): " );
				SimpleDriver.debugln( "add sym( " + rwidx + ", " + i + " ): " ); 
				*/
			    }
			}
		    }
		    else {
			for ( int count=0; count<n; count++ ) { // search through all n points
			    rwidx = ( start + count ) % n; // choose random end among all concepts
			    found = ( rwidx != j ) && ( rwidx != i ) && !containsRelation( rwidx, j, label );
			    if ( found )
				break;
			}
			if ( found ) {			    
			    newrel = new Relation( relation, getConcept( rwidx ), relation.getConcept2() );
			    removeRelation( i, j, relation ); // remove the original relation
			    putRelation( rwidx, j, newrel ); // add new relation
			    /*
			    SimpleDriver.debugln( "----- Rewired incoming relation ----- " );
			    SimpleDriver.debugln( "remove ( " + i + ", " + j + " ): " + relation );
			    SimpleDriver.debugln( "add ( " + rwidx + ", " + j + " ): " + newrel );
			    */
			    if ( !directed ) {
				removeRelation( j, i, relation.getLabel() );
				putRelation( j, rwidx, newrel.symRelation() );
				/*
				SimpleDriver.debugln( "remove sym( " + j + ", " + i + " ): " );
				SimpleDriver.debugln( "add sym( " + j + ", " + rwidx + " ): " );
				*/
			    }
			}
		    }
		}
	    }
	}
    }


    /** Gets the total number of concept categories in the system. */    
    public int getCategoryNum()
    {
	return categories.size();
    }

    /** Gets the total number of relation models in the system. */
    public int getModelNum()
    {
	return models.size();
    }

    /** Gets the total number of concepts in the system. */
    public int getConceptNum()
    {
	return concepts.size();
    }

    /** Gets the total number of relation lists in the system. */
    public int getRelListNum()
    {
	return relations.size();
    }        
    
    /** Gets the index of a concept category in the system. */
    public int getCategoryIndex( String category )
    {
	return categories.indexOf( category );
    }

    /** Gets the index of a relation model in the system. */
    public int getModelIndex( RelationModel model )
    {
	return models.indexOf( model );
    }

    /** Gets the index of a relation model with the specified label.
	@return -1 if not found.
    */
    public int getModelIndex( String label )
    {
	for ( int i=0; i < getModelNum(); i++ ) {
	    if ( label.equals( getModel(i).label ) )
		return i;
	}
	return -1;
    }

    /** Gets the index of a concept in the system. */
    public int getConceptIndex( Concept concept )
    {	
	return concepts.indexOf( concept );
    }

    /** Gets the index of a concept with the specified name.
	@return -1 if not found.
    */
    public int getConceptIndex( String name )
    {
	for ( int i=0; i < getConceptNum(); i++ ) {
	    if ( name.equals( getConcept(i).name ) )
		return i;
	}
	return -1;
    }

    /** Gets the ith concept category. */
    public String getCategory( int i )
    {
	return ( String )categories.get(i);
    }

    /** Gets the ith relation model. */
    public RelationModel getModel( int i )
    {
	return ( RelationModel )models.get(i);
    }

    /** Gets the relation model with the specified label. 
	@return Null if not found.
    */
    public RelationModel getModel( String label )
    {	
	int index = getModelIndex( label );
	if ( index != -1 )
	    return getModel( index );
	else
	    return null;
    }

    /** Gets the ith concept. */
    public Concept getConcept( int i )
    {
	return ( Concept )concepts.get(i);
    }

    /** Gets the concept with the specified name
    	@return Null if not found.
    */
    public Concept getConcept( String name )
    {	
	int index = getConceptIndex( name );
	if ( index != -1 )
	    return getConcept( index );
	else
	    return null;
    }

    /** Adds a concept category. */
    public void addCategory( String category )
    {
	categories.add( category );
    }

    /** Adds a relation model. */
    public void addModel( RelationModel model )
    {
	models.add( model );
    }

    /** Adds a concept. */
    public void addConcept( Concept concept )
    {
	concepts.add( concept );
    }


    /** Gets the key for the relations between concept i and j. */
    public static String getKey( int i, int j )
    {	
	String key = i + "-" + j;
	return key;
    }

    /** Gets the relation list between concept i and concept j. */
    public Vector getRelations( int i, int j )
    {
	return ( Vector )relations.get( getKey( i, j ) );
    }

    /** Gets the relation of the specified label in a relation list. */
    public static Relation getRelation( Vector rellist, String label )
    {
	if ( rellist == null )
	    return null;
	for ( int k=0; k < rellist.size(); k++ ) {
	    Relation relation = ( Relation )rellist.get(k);
	    if ( relation.getLabel().equals( label ) )
		return relation;
	}
	return null;
    }
  
    /** Gets the relation of the specified label between concept i and concept j. */
    public Relation getRelation( int i, int j, String label )
    {
	Vector rellist = ( Vector )relations.get( getKey( i, j ) );
	return getRelation( rellist, label );   
    }
  
    /** Returns true if the relation of the specified label between concept i and concept j exists. */
    public boolean containsRelation( int i, int j, String label )
    {
	return ( getRelation( i, j, label ) != null );   
    }
    
    /** Puts the relation list between concept i and concept j. */
    public void putRelations( int i, int j, Vector rellist )
    {
	if ( rellist == null )
	    return;
	relations.put( getKey( i, j ), rellist );
    }

    /** Adds the specified relation to a relation list. */
    public static void putRelation( Vector rellist, Relation relation )
    {
	if ( rellist == null )
	    return;
	for ( int k=0; k < rellist.size(); k++ ) {
	    Relation currel = ( Relation )rellist.get(k);
	    if ( relation.compareTo( currel ) <= 0 ) {
		rellist.add( k, relation ); // insert in middle
		return;
	    }
	}
	rellist.add( relation ); // add to end
    }

    /** Adds the specified relation to the relation list between concept i and concept j. */
    public void putRelation( int i, int j, Relation relation )
    {
	Vector rellist = ( Vector )relations.get( getKey( i, j ) );
	if ( rellist == null ) { // create new list
	    rellist = new Vector();
	    rellist.add( relation );
	    relations.put( getKey( i, j ), rellist );
	}
	else { // insert into a sorted list
	    putRelation( rellist, relation );
	}
    }

    /** Removes the relation list between concept i and concept j. */
    public void removeRelations( int i, int j )
    {
	relations.remove( getKey(i, j ) );		
    }

    /** Removes the relation of the specified label in a relation list. */
    public static void removeRelation( Vector rellist, String label )
    {
	if ( rellist == null )
	    return;
	for ( int k=0; k < rellist.size(); k++ ) {
	    Relation relation = ( Relation )rellist.get(k);
	    if ( relation.getLabel().equals( label ) ) {
		rellist.remove(k);
		break;
	    }
	}
   }

    /** Removes the specified relation in a relation list. */
    public static void removeRelation( Vector rellist, Relation relation )
    {
	if ( rellist == null )
	    return;
	rellist.remove( relation );
    }

    /** Removes the relation of the specified label from the relation list between concept i and concept j;
	if this causes the list to be empty, remove the list from the system. */
    public void removeRelation( int i, int j, String label )
    {
	Vector rellist = ( Vector )relations.get( getKey( i, j ) );
	removeRelation( rellist, label );
	if ( rellist.size() == 0 )
	    removeRelations( i, j );
     }
      
    /** Removes the specified relation from the relation list between concept i and concept j;
	if this causes the list to be empty, remove the list from the system. */
    public void removeRelation( int i, int j, Relation relation )
    {
	Vector rellist = ( Vector )relations.get( getKey( i, j ) );
	removeRelation( rellist, relation );
 	if ( rellist.size() == 0 )
	    removeRelations( i, j );
   }

    
    /** Checks whether this concept system is categorized, i.e. contains multiple concept categories. */
    public boolean isCategorized()
    {
	return ( getCategoryNum() > 1 );
    }

    /** Checks whether any of the concept in this concept system is ranked. */
    public boolean isRanked()
    {
	for ( int i=0; i < getConceptNum(); i++ )
	    if ( getConcept(i).isRanked() )
		return true;
	return false;
    }

    /** Checks whether any of the concept in this concept system is valued. */
    public boolean isValued()
    {
	for ( int i=0; i < getConceptNum(); i++ )
	    if ( getConcept(i).isValued() )
		return true;
	return false;
    }

    /** Checks whether this concept system contains only simple, i.e. 
	uncategorized(mono-categorized)/unranked/unvalued, concepts. */
    public boolean simpleConcepts()
    {
	return ( !isCategorized() && !isRanked() && !isValued() );
    }

    /** Checks whether this concept system is labeled, i.e. contains multiple relation labels/models. */
    public boolean isLabeled()
    {
	return ( getModelNum() > 1 );
    }

    /** Checks whether this concept system contains any directed relation models. */
    public boolean isDirected()
    {
	for ( int i=0; i < getModelNum(); i++ ) {
	    if ( getModel(i).directed )
		return true;
	}
	return false;
    }

    /** Checks whether any of the relation in this concept system is weighted. */
    public boolean isWeighted()
    {
        Enumeration enum = relations.elements();
        while ( enum.hasMoreElements() ) {
            Vector rellist = ( Vector )enum.nextElement();
	    for ( int i=0; i < rellist.size(); i++ ) {
		Relation relation = ( Relation )rellist.get(i);
		if ( relation.isWeighted() )
		    return true;
	    }
	}
	return false;
    }

    /** Checks whether this concept system contains only simple, i.e.
	unlabeled(monolabled)/undirected/unweighted relations. */
    public boolean simpleRelations()
    {
	return ( !isLabeled() && !isDirected() && !isWeighted() );
    }

    /** Checks whether this is a simple system, i.e. contains only simple concepts and relations. */
    public boolean isSimple()
    {
	return ( simpleConcepts() && simpleRelations() );
    }


    /** Adds another "anchor" (information about formerly attached
	leaf) to the inner node. This method should only be used
	after all concepts have been initialized */
    public void addAnchor(int innerIdx, String leafName) {
	if (anchors==null) {
	    int n=getConceptNum();
	    anchors=new Vector(); 
	    for(int i=0; i< n; i++) {
		anchors.addElement( new Vector(0));
	    }
	}

	if (anchors.size() != getConceptNum())
	    throw new IllegalArgumentException("Wrong anchors.size()");

	if (innerIdx >= anchors.size()) throw new IllegalArgumentException("We don't have node no. " +  innerIdx + " here...");
	Vector v = (Vector)anchors.elementAt( innerIdx);
	v.addElement(leafName);
    }
	
    /** How many anchors are in this system? */
    public int allAnchorCnt() {
	if (anchors == null) return 0;
	int cnt=0;
	for(int i=0; i< anchors.size(); i++) {
	    cnt += ((Vector)anchors.elementAt(i)).size();
	}
	return cnt;
    }
    
    /** How many achors do Concept i of this system and Concept iother
	of the other system have  in common? This can be used
	in external similarity computations */
    public int anchorOverlap( ConceptSystem other, int iHere, int iOther) {
	if (anchors == null || other.anchors==null) return 0;

	Vector a = (Vector)anchors.elementAt(iHere);
	Vector b = (Vector)other.anchors.elementAt(iOther);
	int x=0;
	// The lists are usually very short, so it's OK to do "quadratic-cost"
	// searches	
	for(int ia=0; ia<a.size(); ia++) {
	    for(int ib=0; ib<b.size(); ib++) {
		if (a.elementAt(ia).equals(b.elementAt(ib))) x ++;
	    }
	}
	return x;
    }

    /** Creates and reads a concept system from a text file.
	@return The created concept system. */
    public static ConceptSystem readTxt( String filename ) throws Exception
    {
	ConceptSystem system = new ConceptSystem(); 
	boolean categorized = false, ranked = false, valued = false;
	boolean fv=false;
	boolean labeled = false, weighted = false;
	
	BufferedReader reader = new BufferedReader( new FileReader(filename));
	String line;
	int section = -1; // current section # read
	int lineCnt=0; // how many lines have been read so far

	while ( ( line = reader.readLine() ) != null ) {
	    lineCnt++;
	    
	    //// System.out.println("[sec="+section+"] input: "+line);

	    StringTokenizer st = new StringTokenizer( line, " :,-{}()\t" );
	    if ( !st.hasMoreTokens() || line.startsWith( "#" ) )
		continue;
	    String token = st.nextToken();
	    
	    switch ( section ) {

	    case -1: // expect header ConceptSystem
		if ( token.equals( "ConceptSystem" ) ) {
		    section = 0;
		    system.name = st.nextToken();
		}
		else 
		    throw new Exception( "Missing keyword ConceptSystem!" );
		break;	
	
	    case 0: // expect section ConceptCategories
		if ( token.equals( "ConceptCategories" ) ) {
		    section = 1;
		}
		else 
		    throw new Exception( "Missing section ConceptCategories!");
		break;		

	    case 1: // read section ConceptCategories
		if ( token.equals( "RelationModels" ) ) { // next section starts
		    if ( system.getCategoryNum() == 0 ) // add default category if not specified
			system.addCategory( "Default" );
		    section = 2;		
		}
		else if (token.equals("Concepts")|| token.equals("Relations")) 
		    throw new Exception( "Missing section RelationModels!" );
		else {
		    system.addCategory( token );
		    SimpleDriver.debugln( "!!!!! Got category " + token + " in line " + line );
		    while ( st.hasMoreTokens() ) {
			system.addCategory( st.nextToken() );
			SimpleDriver.debugln( "!!!!! Got category " );
		    }
		}
		break;		

	    case 2: // read section RelationModel
		if ( token.equals( "Concepts" ) ) {  // next section starts
		    if ( system.getModelNum() == 0 ) // add default model if not specified
			system.addModel( new RelationModel() );
		    section = 3;
		    while ( st.hasMoreTokens() ) {
			token = st.nextToken();
			if ( token.toLowerCase().equals( "categorized" ) ) 
			    categorized = true; 			    
			else if ( token.toLowerCase().equals( "ranked" ) )
			    ranked = true;
			else if ( token.toLowerCase().equals( "valued" ) )
			    valued = true;
			else if ( token.toLowerCase().equals( "fv" ) )
			    fv = true;
			else {
			    String msg = "Illegal modifier `"+token+
				"' in line"+lineCnt;
			    throw new Exception(msg);
			}
		    }
		    if (fv && valued) {
			// The two keywords are not compatible
			String msg = "Concepts can be valued, or feature-vectored, but not both! (line "+lineCnt+")";
			throw new Exception(msg);
			   
		    }
		}
		else if ( token.equals( "Relations" ) ) 
		    throw new Exception( "Missing section Concepts!" );
		else {
		    String label = token;		
		    String dirstr = st.nextToken().toLowerCase();
		    boolean directed = false;
		    if ( dirstr.equals( "directed" ) )
			directed = true;
		    else if ( dirstr.equals( "undirected" ) )
			directed = false;
		    else throw new Exception( "Illegal direction for RelationModel " + label );
		    RelationModel model = new RelationModel( label, directed );
		    system.addModel( model );
		}
		break;

	    case 3: // read section Concepts
		if ( token.equals( "Relations" ) ) { // next section starts
		    section = 4;		
		    while ( st.hasMoreTokens() ) {
			token = st.nextToken();
			if ( token.toLowerCase().equals( "labeled" ) )
			    labeled = true;
			else if ( token.toLowerCase().equals( "weighted" ) )
			    weighted = true;
		    }
		}
		else {
		    String name = token;
		    String category = system.getCategory(0);
		    int rank = Concept.RANK;

		    if ( categorized ) {
			category = st.nextToken();
			if ( system.getCategoryIndex( category ) == -1 )
			    throw new Exception( "Unknown category for concept " + name ); 
		    }

		    if ( ranked ) {
			rank = Integer.parseInt( st.nextToken() );
			if ( rank < 1 || rank > 10 )
			    throw new Exception( "Illegal rank for concept " + name +
						 ", must be an integer in the range of [1, 10]." );
		    }

		    Concept concept = null;
		    if (fv) {
			// Create a "feature vector" concept
			// Expect: a sequence of (index, value) pairs
			Vector v=new Vector();
			while ( st.hasMoreTokens() ) {
			    int idx =  Integer.parseInt( st.nextToken() );
			    if (!st.hasMoreTokens()) {
				String msg = "Feature index not followed by value (line "+lineCnt+")";
				throw new Exception(msg);
			    }
			    double val = Double.parseDouble( st.nextToken());
			    v.addElement(new 
				FeatureVectorConcept.IVPair(idx, val));
			}
			concept=new FeatureVectorConcept(name,category,rank,v);
		    } else if ( valued ) {
			// Expect: d1 d2
			double d1 = Double.parseDouble( st.nextToken() );
			double d2 = Double.parseDouble( st.nextToken() );
			if ( d1 < 0 || d1 > 1 || d2 < 0 || d2 > 1 )
			    throw new Exception( "Illegal values for concept " + name +
						 ",\n must be in the range of [0, 1]." ); 
			concept = new Concept( name, category, rank, d1, d2 );
		    } else {
			concept = new Concept( name, category, rank, 0, 0 );
		    }

		    system.addConcept( concept );
		}
		break;

	    case 4: // read section Relations
		if ( token.equals( "Anchors" ) ) {
		    section = 5;
		    continue;
		} else {

		    // model
		    RelationModel model = system.getModel(0);
		    if ( labeled ) {
			String label = token;
			model = system.getModel( label );
			if ( label == null )
			    throw new Exception( "Illegal relation label " + label );
		    }	
		    
		    // concepts
		    String name1 = st.nextToken();
		    String name2 = st.nextToken();
		    int idx1 = system.getConceptIndex( name1 );
		    int idx2 = system.getConceptIndex( name2 );
		    Concept c1 = system.getConcept( idx1 );
		    Concept c2 = system.getConcept( idx2 );
		    if ( c1 == null || c2 == null ) 
			throw new Exception( "Illegal concept " + name1 + " and/or " + name2 + " in relation." );
		    
		    // weight
		    double weight = 1;
		    if ( weighted ) {
			weight = Double.parseDouble( st.nextToken() );
			if ( weight < 0 || weight > 1 )
			    throw new Exception( "Illegal weight " + weight + " for relation,\n" +
						 "must be in the range of [0, 1]." );
		    }
		    
		    // relation
		    Relation relation = new Relation( model, c1, c2, weight );
		    if ( idx1 != idx2 )
			system.putRelation( idx1, idx2, relation );

		    /******************** fixed on 10/16/03 *********************/
		    // store undirected relation for the other direction
		    if ( !relation.isDirected() && idx1 != idx2 ) 
			system.putRelation( idx2, idx1, relation.symRelation() );

		    /*
		    // If the relation is undirected, it has to be
		    // stored for both (c1,c2) and (c2,c1) (V.M., Oct-17-2003)
		      if (!model.directed && idx1 != idx2) {
			Relation relRev = new Relation(model,c2,c1,weight );
			system.putRelation( idx2, idx1, relRev );
		    }
		    */

		}
		break;
		
	    case 5: // read section Anchors	    
		String innerName = token;
		int idx = system.getConceptIndex( innerName );
		
		while ( st.hasMoreTokens() ) {
		    token = st.nextToken();
		    system.addAnchor(idx, token);

		}			
		break;
	    } // end of switch
	} // end of while ...

	if ( section < 4 ) {
	    //// System.err.println( "Section = " + section);
	    throw new Exception( "ConceptSystem file incomplete, only has " + section + " sections." );
	}
	reader.close();
	return system;
    }	

    /** Creates and reads a concept system from a binary file.
	@return The created concept system. */
    public static ConceptSystem readBin( String fileName ) throws Exception
    {
	FileInputStream fis = new FileInputStream( fileName );
	ObjectInputStream ois = new ObjectInputStream( fis );	
	ConceptSystem system = ( ConceptSystem )ois.readObject();
	ois.close();	    
	return system;
    }

    /** Writes out the concept systems object into a binary file. */
    public void writeBin( String fileName ) throws Exception
    {
	FileOutputStream fos = new FileOutputStream( fileName );
	ObjectOutputStream oos = new ObjectOutputStream( fos );	
	oos.writeObject( this );
	oos.close();
    }

    /** Returns a string representation of this concept system. */
    public String toString()
    {
        String str = "ConceptSystem " + name + " ----------\n";
	int i;

        str += "\nConcept Categories ----- \n{ ";
        for ( i=0; i < getCategoryNum(); i++ ) {
            str += getCategory(i);
	    if ( i < getCategoryNum()-1 )
		str += ", ";
	    else
		str += " }\n";
	}

        str += "\nRelation Models ----- \n";
        for ( i=0; i < getModelNum(); i++ ) 
            str += getModel(i) + "\n";

        str += "\nConcepts -----\n";
	str += "Concept Name\tCategory\tRank\tValues\n"; 
        for ( i=0; i < getConceptNum(); i++ ) 
            str += getConcept(i) + "\n";

        str += "\nRelations -----\n";
	str += "Relation Label\tConcepts\tWeight\n";
        Enumeration enum = relations.elements();
        while ( enum.hasMoreElements() ) {
            Vector rellist = ( Vector )enum.nextElement();
	    for ( i=0; i < rellist.size(); i++ )
		str += ( Relation )rellist.get(i) + "\n";
	}

        return str;    
    }
}
