///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// ABSURDIST II
// Ying Feng, Vladimir Menkov
// Computer Science, Indiana University
// June 2004
//
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


package absurdist;
    
import java.io.*;
import java.text.*;
import java.util.*;

import common.*;  // Classes used both by ABSURDIST and by other our packages
import driver.*;
import gui.*;



/** Mapping quality measurements 
 */

class Measures {

    /** Return type for evalMap() */
    static class Quality {
	int misCon;
	double misRel;
	double accuracy;
	Quality() {
	    misCon=0;
	    misRel=0;
	    accuracy=0;
	}
    }

  /** Evaluates the quality of the mapping given in mapON[].

      The "mismapped edges", or "mismapped relations" measure is computed 
      for a permutation map[] as follows:

      <pre>
      misRel(map) = 
        0.5 * sum_{i,j} Diff(Link(A,i,j), Link(B,map[i],map[j]) -
        alpha/beta*(n-1)* ( sum_{q,x}ExSim[q][x] - sum_{i}ExSim[i][map[i]] ).
      </pre>

      In the case of matching non-labeled, non-weighted graphs with no
      extrenal similarity, misRel(map) is equal to the number of edges
      that are present in one graph but absent in the other.  When
      external similarity is present, misRel also includes the numeber
      of pairs (q,x) where external similarity is present, but q is
      not matched to x, with the weight of (alpha/beta)*(n-1)*ExSim(q,x).
      

      On the set of all permutation matrices, minimizing the misRel
      functional is equivalent to maximizing the energy functionals,
      because
      
      <pre>
      Energy(map) = (alpha/(n-1))*(n/2) + alpha*sum_{q,x}ExSim[q][x] -
          (beta/(n-1))*misRel(map).
      </pre>




        <p> The "mismapped concepts" measure measures the difference
        between the only applies to mappings between a concept system
        and its noisy copy, assuming that the noise is small enough so
        that the mapping should be identity, i.e. map[Ci]=Ci'.  </p>

	@param absurdist Is used to access the structures of the 
	concept systems, via Absurdist.calcLinkSym()
	@param mapON The mapping whose quality is to be evaluated
	@param correspond Used to compute results.accuracy
	@param results  Used to return the results (misCon,
	misRel, accuracy)
	
    */
    static public Quality evalMap(Absurdist absurdist, 
				  int mapON[], 
				  double correspond[][]) {


        SimpleDriver.debugln( "Evaluate mapping ..." );
    
        int no = mapON.length;
        int nn = correspond[0].length;
	Quality results=new Quality();
	results.misCon = 0;
	results.misRel = 0;
	results.accuracy = 0;

        for ( int io=0; io<no; io++ ) {    
	    int in = mapON[io];
	    if ( io != in )
		// misCon is the total number of mismapped concepts, assuming identity mapping is the correct one 
		results.misCon++;
            for ( int jo=0; jo<no; jo++ ) {
		if (jo == io) continue; // skip diagonal element
                int jn = mapON[jo];  
                if ( in != -1 && jn != -1 ) {
		    // misRel is the sum of weight differences of
		    // mapped relations (which is equivalent to 
		    // our energy functional -- June 2004).
		    double sim = absurdist.calcLinkSim( new Link( io, jo ), 
							new Link( in, jn ));

		    //System.out.println("EM: sim("+io+", "+jo+" ; "+
		    //		       in+", "+jn+")=" + sim + ", diff=" +
		    //		       Math.abs(1-sim)	       );
		    results.misRel += Math.abs(1-sim);
		    // "accuracy" is the sum of mapped relation
		    // similarities weighted by their correspondences
                    results.accuracy += 
			sim * correspond[io][in] * correspond[jo][jn];
		}
	    }
        }

	results.misRel /= 2; 
	//System.out.println("EM: misRel=" + results.misRel);

  
	if (absurdist.alpha !=0) {
	    double exSimFunc = 0;	    
	    for ( int io=0; io<no;  io++ ) {
		int jn = mapON[io];  
		for ( int in=0; in<nn; in++ ) {
		    if (in != jn) exSimFunc += absurdist.exSim[io][in];
		}
	    }
	    results.misRel +=(absurdist.alpha/absurdist.beta)*(no-1)*exSimFunc;
	    System.out.println("EM(ext): misRel=" + results.misRel);
	}


        results.accuracy /= no * nn;
	return results;
    }            
    

}
