///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// ABSURDIST++
// Ying Feng
// Computer Scicence, Indiana University
// Feb-Aug 2003
//
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////



package absurdist;

import java.text.*;
import java.util.*;
import java.io.Serializable;

import driver.*;



/** Class Relation
    Specifies the data structure for a relation, which contains its involved concepts and some features.
*/ 
public class Relation implements Serializable
{
    /** Relation model of this relation. 
     */
    public RelationModel model = null;
    /** Concepts involved in this relation. 
     */
    public Vector concepts = new Vector(2);
    /** Strength measurement of this relation ( 0 < weight (quantitative) <= 1 ). 
     */
    public double weight = 0;
    
    /*
    //  concepts involved 
    public Concept c1 = null, c2 = null;
    //  label (nominal) 
    public String label = "";
    Link link = null;		            // corresponding link
    */

    /** Creates an empty relation.
     */
    public Relation ()
    {
    }

    /** Creates a relation of the specified model with default weight between two concepts.
     */
    public Relation ( RelationModel model, Concept c1, Concept c2 )
    {
	this.model = model;
	concepts.add( c1 );
	concepts.add( c2 );
	weight = 1;
	/*
	label = "Default";
	this.directed = false;
	*/
    }

    /** Creates a relation of the specified model with random weight if weighted between two concepts.
     */
    public Relation ( RelationModel model, Concept c1, Concept c2, boolean weighted )
    {
	this.model = model;
	concepts.add( c1 );
	concepts.add( c2 );
	if ( weighted ) 
	    weight = Double.MIN_VALUE + Math.random();	
	else
	    weight = 1;
	if ( weight > 1 )
	    weight = 1;
	/*
	this.label = label;
	this.directed = false;
	*/
    }

    /** Creates a relation with full features between two concepts.
     */
    public Relation ( RelationModel model, Concept c1, Concept c2, double weight )
    {
	this.model = model;
	concepts.add( c1 );
	concepts.add( c2 );
	this.weight = weight;
    }

    /** Creates a copy of a relation sharing the same concepts.
     */
    public Relation ( Relation relation )
    {
	model = relation.model;
	concepts = relation.concepts;
	weight = relation.weight;
	/*
	label = relation.label;
	this.directed = directed;
	*/
    }

    /** Creates a copy of a relation, using the specified concepts.
     */
    public Relation ( Relation relation, Concept c1, Concept c2 )
    {
	model = relation.model;
	concepts.add( c1 );
	concepts.add( c2 );
	weight = relation.weight;
    }

    /** Creates a copy of a relation, using the specified concepts, with some noise.
     */
    public Relation ( Relation relation, Concept c1, Concept c2, double noise )
    {
	model = relation.model;
	concepts.add( c1 );
	concepts.add( c2 );
	weight = relation.weight + noise * ( 2 * Math.random() - 1 );
	if ( weight <= 0 )
	    weight = Double.MIN_VALUE;
	if ( weight > 1 )
	    weight = 1;    
    }

    /** Returns the symmetrical copy of this relation.
     */
    public Relation symRelation()
    {
	return new Relation( this, getConcept2(), getConcept1() );
    }


    /** Adds random noise to the relation, i.e. randomize its weight.
	Has no effect to unweighted relations.
     */
    public void addNoise ()
    {
	if ( isWeighted() ) 
	    weight = Math.random();    
    }

    /** Adds some noise to the relation's weight.
     */
    public void addNoise ( double noise )
    {
	weight += noise * ( 2 * Math.random() - 1 );
	if ( weight <= 0 )
	    weight = Double.MIN_VALUE;
	if ( weight > 1 )
	    weight = 1;    
    }

    /** Returns the 1st concept in the relation.
     */
    public Concept getConcept1()
    {
	return ( Concept )concepts.get(0);
    }

    /** Returns the 2nd concept in the relation.
     */
    public Concept getConcept2()
    {
	return ( Concept )concepts.get(1);
    }

    /** Returns the label of the relation.
     */
    public String getLabel()
    {
	return model.label;
    }

    /** Checks whether this relation is directed.
     */
    public boolean isDirected()
    {
	return model.directed;
    }

    /** Checks whether this relation is weighted.
     */
    public boolean isWeighted()
    {
	return ( weight != 0 && weight != 1 );
    }

    /** Compares the lexicographical order of this relation and the specified one, according to their label.
     */
    public int compareTo( Relation relation )
    {
	return model.compareTo( relation.model );
    }

    /** Return the difference between the weight of this relation and the specified one.
     */
    public double difWeight( Relation relation )
    {
	return ( Math.abs( weight - relation.weight ) );
    }

    /** Returns a string representation of this relation.
     */
    public String toString()
    {
	String str = "Relation " + model.label + "( ";
	str += getConcept1().name + ", " + getConcept2().name;
	DecimalFormat df = new DecimalFormat( "0.0000" );
	str += " )\t" + df.format( weight );

	/*
	int arity  = concepts.size();
	for ( int i=0; i < getArity(); i++ ) 
	    str += getConcept(i).name + ", ";
	str += getConcept( getArity() - 1 ).name;
	if ( directed )
	    str +=  "directed";
	else
	    str += "undirected";
	*/

	return str;
    }
}
