package common;

import java.util.*;
import absurdist.*;
import driver.*;



/** Concept System Graph Interface (CSGI) allows us to access the
information about a ConceptSystem as if it were a standard graph.
The main advantage of using this interface consists in enabling 
us to directly access the "neighbor list" for each node, without
having to query the hash table for each possible pair. This is 
analogous to using a sparse-matrix representation as compared
to a dense matrix.
*/
public class CSGI {

    /** nei[j][] is the sorted list of nodes connected to the node j 
     (not including j itself)*/
    private int nei[][];

    /** neiPlus[j] = nei[j] + {j}	 */
    private int neiPlus[][];

    /** Returns the sorted list of nodes connected to the node j */
    public int[] getNei(int j) {
	return nei[j];
    }

    /** Returns the sorted list of nodes connected to the node j, plus
     j itself */
    public int[] getNeiPlus(int j) {
	return neiPlus[j];
    }
	
    /** The degree of node j */
    public int degree(int j) {
	return nei[j].length;
    }


    /** The index of the node that has the largest degree.
	(If several nodes have this degree, the first one
	will be returned).
    */
    public int getMaxDgrConceptIdx() {
	int idx=0;
	int maxDeg = 0;
	for(int i=0; i<nei.length; i++) {
	    if (i==0 || degree(i)>maxDeg) {
		idx = i;
		maxDeg = degree(i);
	    }
	}
	return idx;
    }

    /** The index of the node that has the lowest degree.
	(If several nodes have this degree, the first one
	will be returned).
    */
    public int getMinDgrConceptIdx() {
	int idx=0;
	int minDeg = 0;
	for(int i=0; i<nei.length; i++) {
	    if (i==0 || degree(i)<minDeg) {
		idx = i;
		minDeg = degree(i);
	    }
	}
	return idx;
    }

    /** Creates a CSGI representation for  a ConceptSystem */
    public CSGI( ConceptSystem s) throws IllegalArgumentException {
	try {
	int n=s.getConceptNum(); 
	int eCnt=0;
	nei = new int[n][];	
	neiPlus = new int[n][];	
	for(int i=0; i<n; i++) {
	    int v[] = new int[n];
	    int m=0;
	    for(int j=0; j<n; j++) {
		if (i==j) continue;

		/*
		double w= s.getRelation(i, j, s.getModel(0).label ).weight;
		if (w==0) {
		} else if (w ==1) {
		    v[m++]=j;
		} else {
		    String msg= "Graph edge has weight=" + w +
			". We only can handle 0 and 1";
		    throw new IllegalArgumentException(msg);
		}
		*/

		Vector r1 = s.getRelations(i, j);
		Vector r2 = s.getRelations(j, i);
		if (r1!=null && r1.size() >0 || r2!=null && r2.size() >0 ) {
		    v[m++]=j;
		}

	    }
	    nei[i] = new int[m];
	    neiPlus[i] = new int[m+1];	    
	    for(int j=0; j<m; j++) {
		nei[i][j]=v[j];
	    }
	    for(int j=0; j<m; j++) {
		neiPlus[i][j]=v[j];
	    }
	    neiPlus[i][m] = i;
	    eCnt+=m;
	}
	SimpleDriver.debugln("CSGI: storing " + eCnt + "/2 edges in an " + n + "-node graph");
	} catch(NullPointerException e) {
	    e.printStackTrace();
	    throw e;
	}
    }
}
