package common;

/** An instance of this class can be used to store a correspondence
    matrix between the two systems, or any other object of the same
    dimension */
public class CorrespondenceVector implements Cloneable {
    


    public double[][] d;
    private double[] _rowSum=null;
    private double[] _colSum=null;
    public final int n, m;
    /** When this value is initialized, we set allSumSet, to indicate
	possibility of reuse */
    private double _allSum=0;
    private boolean allSumSet=false;


    /** Creates a CorrepspondenceVector structure of the specified sizes */
    public CorrespondenceVector(int n, int m) {
	this.n=n;
	this.m=m;
	d = new double[n][m];
	flush();
    }

    /** Creates a CorrepspondenceVector structure "wrapped around" an
	existing data array. Does *not* copy the data.

	@param arr a properly allocated rectangualr array.
    */
    public CorrespondenceVector(double [][] arr) {
	d = arr;
	this.n = arr.length;
	this.m = arr[0].length;
	flush();
    }

    /** Prepares a copy of this matrix */
    public Object clone() {
	CorrespondenceVector v = new CorrespondenceVector(n,m);
	for(int i=0;i<d.length;i++) {
	    for(int j=0; j<d[i].length; j++) {
		v.d[i][j] = d[i][j];
	    }
	}
	return v;
    }

    /** Flushes the "cached" values of row and column sums. You must
	call this method every time when you change the stored
	values. */
    public void flush() {
	_rowSum=_colSum=null;
	allSumSet = false;
    }

    public double[][] getMatrix() {
	return d;
    }
    
    /** Lazy evaluation for f(a)=C(a,All)=sum_b{C(a,b)}
     */
    public  double[] rowSum() {
	if (_rowSum==null) {
	    _rowSum = new double[d.length];
	    for(int i=0;i<d.length;i++) {
		double s=0;
		for(int j=0; j<d[i].length; j++) s+=d[i][j];
		_rowSum[i] = s;
	    }
	}
	return _rowSum;
    }
    /** Lazy evaluation for f(b)=C(All,b)=sum_a{C(a,b)}
     */
    public double[] colSum() {
	if (_colSum==null) {
	    _colSum = new double[d[0].length];
	    for(int i=0;i<d[0].length;i++) {
		double s=0;
		for(int j=0; j<d.length; j++) s+=d[j][i];
		_colSum[i] = s;
	    }
	}
	return _colSum;
     }
 
    /** @return sum_{a,b} C(a,b)
     */
     public double allSum() {
	 if (!allSumSet) {
	     double r[] = rowSum();
	     _allSum=0;
	     for(int j=0; j<r.length; j++) _allSum +=r[j];
	     allSumSet=true;
	 }
	 return _allSum;
    }

    
    /** @return sum_{y in B \({x}+Nei(x)) } C(a,x)
     */
    public double sum_1_far(final int a, final int x, final int neiPlusX[]) {
	double r[] = rowSum();
	double s= r[a];
	for(int j=0; j<neiPlusX.length; j++) s -= d[a][ neiPlusX[j]];
	return s;
    }

    /** @return sum_{b in A \({a}+Nei(a)) } C(b,x)
     */
    public double sum_far_1(final int a, final int x, final int neiPlusA[]) {
	double c[] = colSum();
	double s= c[x];
	for(int j=0; j<neiPlusA.length; j++) s -= d[neiPlusA[j]][x];
	return s;
    }
   
   /** @return sum_{b in A \({a}+Nei(a)) } sum_{y in B \({x}+Nei(x)) } C(b,y)
     */
    public double sum_far_far(final int a, final int x, 
			      final int neiPlusA[],  final int neiPlusX[]) {
	double r[] = rowSum();
	double c[] = colSum();
	double s= allSum();
	for(int j=0; j<neiPlusA.length; j++) s -= r[ neiPlusA[j]];
	for(int j=0; j<neiPlusX.length; j++) s -= c[ neiPlusX[j]];

	for(int j=0; j<neiPlusA.length; j++) {
	    final int b = neiPlusA[j];
	    for(int k=0; k<neiPlusX.length; k++) {
		int y = neiPlusX[k];
		s +=  d[b][y];
	    }
	}	
	return s;
    }
   

    /** Dot product */
    public double dp(CorrespondenceVector v) {
	double s=0;
	for(int i=0;i<d.length;i++) {
	    for(int j=0; j<d[i].length; j++) {
		s+=d[i][j] * v.d[i][j];
	    }
	}
	return s;
    }

    /** this += alpha  * v */
    public void daxpy(double alpha, CorrespondenceVector v) {
	flush();
	double s=0;
	for(int i=0;i<d.length;i++) {
	    for(int j=0; j<d[i].length; j++) {
		d[i][j] += alpha * v.d[i][j];
	    }
	}
    }

    /** this += v */
    void plus(CorrespondenceVector v) {
	daxpy(1.0, v);
    }

    /** this -= v */
    void minus(CorrespondenceVector v) {
	daxpy(-1.0, v);
    }


}
 
