package graph;

import java.io.*;
import java.lang.*;
import java.util.*;

import absurdist.Relation;
import absurdist.Concept;
import absurdist.ConceptSystem;

/** An un-directed graph */
public class Graph {

    static boolean debug=false;

    /** An instance of this class represents one node of the graph */
    class Node {
	/** Sorted list of nodes connected to this node. Each neigh[i]
	 is Integer. */
	public Vector neighbors = new Vector();

	/** Some invariants used for sorting nodes according to
            their topological properties */
	int inv, invNext;

	/** The k-th neighbor */
	public int neigh(int k) {
	    return ((Integer)(neighbors.elementAt(k))).intValue();
	}

	/** How many neighbors? */
	public int degree() {
	    return neighbors.size();
	}

	public String toString() {
	    StringBuffer s = new StringBuffer();
	    for(int i=0; i<neighbors.size(); i++) {
		s.append(" " + neigh(i));
	    }
	    s.append( "\n");
	    return s.toString();
	}

	/** Is this node linked to node x? */
	boolean linkedTo(int x) {
	    return find(x) >= 0;
	}
	/** At which position is x in the neighbor list?
	    @return The index of x, or -1 if not found */
	int find(int x) {
	    // Use binary search to find x in the list of neighbors
	    int k1=0, k2=neighbors.size();
	    while( k1 < k2) {
		int k=(k1+k2)/2;
		int stored = neigh(k);
		if (stored > x) { k2 = k; }
		else if (stored == x) { return k; }
		else { k1 = k+1; }
	    }
	    return -1;
	}


	/** Insertx x into the neighbor list, if it is not there yet 
	 @return true if the value has been inserted; false if it has not,
	because it already was in the list */
	public boolean insertIfNotThere(int x) {
	    if (linkedTo(x)) return false;
	    // Where to insert?
	    int i=0;
	    for(i=0; i<neighbors.size() && neigh(i) < x; i++) {}
	    neighbors.insertElementAt(new Integer(x), i);
	    return true;
	}

	Node permute(int [] p) {
	    int h[]=new int[degree()];
	    for(int i=0; i<degree(); i++) {
		h[i]=p[neigh(i)];
	    }
	    Arrays.sort(h);
	    Node g=new Node();
	    for(int i=0; i<degree(); i++) {
		g.neighbors.addElement( new Integer(h[i]));
	    }
	    return g;
	}
    };

    /** The nodes of the graph. Use the size of this array to find
     the number of nodes in the graph */
    Node[] node;

    // Return total number of nodes.
    public int getNodeNum()
    {
        return node.length;
    }
    
    /*
    // Return the ith node.
    public Node getNode( int i )
    {
        return node[i];
    }
    */
    
    // Get the degree of the ith node.
    public int getNodeDegree( int i )
    {
        return node[i].degree();
    }
    
    // Get the jth neighbour of the ith node.
    public int getNodeNeigh( int i, int j )
    {
        return node[i].neigh( j );
    }
    
    public String toString() {
	StringBuffer s = new StringBuffer("<<<=== Graph:\n");
	for(int i=0; i<node.length; i++) {
	    s.append("Node " + i + ":" + node[i]);
	}
	return s.toString();
    }

    /** Empty graph. */
    private Graph() {}
    
    /** Constructs a random graph with N nodes and (N*D)/2 edges */
    public Graph(int N, double D) throws Exception {
	if (D > (N-1)) {
	    String msg = "In a graph with " +N+ 
		" nodes, node cannot have average degree of "+D+"!";
	    throw new Exception(msg);
	}
	// How many edges?
	int e = (int)(N*D*0.5);
	node = new Node[N];
	for(int i=0; i<N; i++) node[i] = new Node();
	for(int i=0; i<e; i++) {
	    while(true) {
		int n1 = (int)( N * Math.random());
		int n2 = (int)( (N-1) * Math.random());
		if (n2 >= n1) n2++;
		// Try edge (n1,n2)
		boolean inserted = node[n1].insertIfNotThere(n2);
		if (inserted) {
		    node[n2].insertIfNotThere(n1);
		    break;
		}
	    }
	}	
    }


    /** Show how the nodes are divided into classes of equivalency */
    void displayEquiv(NodeInfo w[]) {
	int cnt=0;
	int lastVal=0;
	for(int i=0; i<node.length; i++) {
	    int id =w[i].id;
	    int inv = node[id].inv;
	    //System.out.print("(" + id + ": " + node[id].inv + ") ");
	    if (cnt==0) {
		System.out.print("{");
		lastVal=inv;
		cnt=1;
	    } else if (inv==lastVal) {
		cnt++;
	    } else {
		System.out.print("} {");			
		lastVal=inv;
		cnt=1;
	    }

	    if (cnt>1) System.out.print(" ");
	    System.out.print(id);
	}
	if (cnt>0) System.out.print("}");
	System.out.println();
    }

    /** Tests whether this graph may have some symmetry, by means
	of looking at N topological invariants of the vertices.
	@return false is we have proven that the graph is asymmetric,
	by studying the topological invariants of its nodes; true if
	we have failed to prove asymmetry, which means that the
	graph may (or may not) be symmetric.
    */
    public boolean mayBeSymmetric() {
	return mayBeSymmetric(-1);
    }
	
    /** Tests whether this graph may have some symmetry, by means
	of looking at topological invariants of the vertices.
	@param maxNI How many invariants to test. If -1 is passed, 
	the method will test N invariants. In practice, 5 or 10 probably
	would be enough for most random graphs.
	@return false is we have proven that the graph is asymmetric,
	by studying the topological invariants of its nodes; true if
	we have failed to prove asymmetry, which means that the
	graph may (or may not) be symmetric.
    */
    public boolean mayBeSymmetric(int maxNI) {
	if (maxNI < 0) maxNI = node.length;
	if (debug) System.out.println("Starting symmetry check...");
	NodeInfo w[]=new NodeInfo[node.length];
	/** Start with the first invariant: node degree */
	for(int i=0; i<node.length; i++) {
	    w[i] = new NodeInfo(i);
	    node[i].invNext = node[i].neighbors.size();
	}
	/** Keep checking invariants, until you either see that
	    all nodes are topologically different, or you get
	    tired... */
	for(int nChecked = 0; nChecked < maxNI; nChecked ++) {
	    // Concatenate the current invariant to the key, as 
	    // the string. (FIXME: this is rather inefficient;
	    // a byte array would make more sense)
	    for(int i=0; i<w.length; i++) {
		int id = w[i].id;
		node[id].inv = node[id].invNext;
		w[i].key += ":" + Integer.toHexString(node[id].inv);
	    }
	    // Sort with respect to the current key
	    Arrays.sort(w);
	    //if (debug) displayEquiv(w);
	    // Are they all different now?
	    if (NodeInfo.allKeysDifferent(w)) {
		if (debug) displayEquiv(w);
		if (debug) System.out.println("Detected topological uniqueness of all nodes after processing " + (nChecked+1) + " invariants");
		return false;
	    }

	    // Let's prepare one more invariant, then...
	    // (invNext(X) based on inv(Y) of all neighbors Y of X)
	    for(int i=0; i<node.length; i++) {
		Node a = node[i];
		Integer r[]=new Integer[a.neighbors.size()];
		for(int k=0; k< a.neighbors.size(); k++) {
		    int j = a.neigh(k);
		    r[k] = new Integer(node[j].inv);
		}
		Arrays.sort(r);
		// Now, the array r[] contains a new topological
		// invariant. For convenience, lets convert it
		// into a single number
		node[i].invNext = NodeInfo.hashArray(r);
	    }
	}
	if (debug) {
	    displayEquiv(w);
	    System.out.println("Even after checking " + node.length + " invariants, some nodes are not distinguished...");
	}
	// We still have failed to resolve some nodes. Have to try
	// some permutations now...
	return true;
    }

    /** Returns an array of integers, from 0 to n-1, randomly permuted.
     */
    static public int[] randomPermutation(int n) {
	Hashtable listed=new Hashtable();
	int p[] = new int[n];
	for(int i=0; i<n; i++) {
	    int v;
	    do {
		v = (int)(Math.random() * n);
	    } while (listed.get(new Integer(v))!=null);
	    p[i] = v;
	    listed.put(new Integer(v), "");
	}
	return p;
    }

    /** Permutes the graph. In the new graph, the p[0]-th node is
	the image of the 0-th node in the graph, etc. 
    @param p Permute according to this table. */
    public Graph permute(int p[]) {
	Graph g=new Graph();
	int n=node.length;
	if (n != p.length) throw new IllegalArgumentException("Array length mismatch");
	g.node = new Node[ n];
	for(int i=0; i<n; i++) {
	    g.node[p[i]] = node[i].permute(p);
	}	
	return g;
    }

    /** Converts a Graph to a ConceptSystem. */
    /*
    public absurdist.ConceptSystem toConceptSystem(String name) {
	Vector concepts = new Vector();          // all concepts in the system
	for (int i=0; i<node.length; i++ ) {
	    concepts.add( new Concept( "C" + i ) );
	}

	Hashtable relations = new Hashtable();   // all relations in the system

	for (int i=0; i<node.length; i++ ) {
	    for(int k=0; k<node[i].degree(); k++) {
		int j=node[i].neigh(k);
		Relation relation =  new Relation( (Concept)concepts.get(i), 
						   (Concept)concepts.get(j) );
		relations.put( ConceptSystem.getKey( i, j ), relation );
	    }
	}
	
	return new ConceptSystem(name, concepts, relations );
    }
    */

    public static void main(String[] arg) throws Exception {
	if (arg.length != 3) {
	    System.out.println("Usage: Graph N D graph-count\n");
	    System.exit(1);
	}
	int N = Integer.parseInt(arg[0]);
	double D = Double.valueOf(arg[1]).doubleValue();
	int cnt = Integer.parseInt(arg[2]);
	System.out.println("Producing " + cnt + " random graphs with " + N +
			   " vertices, avg degree = " + D);
	for(int i=0; i<cnt; i++) {
	    Graph g = new Graph(N, D);
	    //System.out.println(g);
	    boolean sym = g.mayBeSymmetric();
	    if (sym) {
		System.out.println("May be symmetric");	    
	    } else {
		System.out.println("Proven asymmetric by studiying invariants");
	    }
	}
    }

}
