///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// ABSURDIST II
// Ying Feng
// Computer Science, Indiana University
// Feb-Aug 2003
//
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


package gui;

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import javax.swing.*;
import javax.swing.event.*;

import absurdist.*;
import graph.*;
import driver.*;


/**
    Class AbsPanel
    is the main panel in the ABSURDIST II interactive GUI window.
*/
public class AbsPanel extends JPanel implements ActionListener, MouseInputListener
{
    // refresh the screen every 100 milliseconds
    static final int DT = 100;	
    
    // controls for source system    
    JLabel sourceLabel;	        // label for source system name
    JButton importSrcButton;    // button to import source system
    JLabel weightSrcLabel;	// label for weight noise to source system
    JSlider weightSrcSlider;	// slider to adjust weight noise for source system
    JLabel rewireSrcLabel;	// label for rewire noise to source system
    JSlider rewireSrcSlider;	// slider to adjust rewire noise for source system
    JButton noiseSrcButton;     // button to add noise to source system

    // controls for target system    
    JLabel targetLabel;	        // label for target system name
    JButton importTgtButton;    // button to import target system
    JLabel weightTgtLabel;	// label for weight noise to target system
    JSlider weightTgtSlider;	// slider to adjust weight noise for target system
    JLabel rewireTgtLabel;	// label for rewire noise to target system
    JSlider rewireTgtSlider;	// slider to adjust rewire noise for target system
    JButton noiseTgtButton;     // button to add noise to target system

    // controls for mapping and export
    JLabel exSimLabel;	        // label for external similarity coefficient
    JSlider exSimSlider;	// slider to adjust external similarity coefficient
    JButton mapButton;          // button to map the systems
    ////JButton viewButton;         // button to view the mapping matrix results
    ////JButton cancelButton;   // button to cancel the current mapping iteration
    JButton exportButton;       // button to export the mapping results

    // display panels
    GraphPanel graphPanel;	// panel to display the two graphs being mapped
    MatrixPanel matrixPanel;	// panel to display important mapping matrices and their plots    

    // grid bag layout/constraints used by sub-panels in control box
    GridBagLayout gbLayout; 
    GridBagConstraints gbConstr;

    /** absurdist instance for the whole system */
    Absurdist absurdist = new Absurdist();

    /** Directory from/to which the user last read/wrote files. */
    File filedir = new File(".");    


    /** Constructor for AbsPanel. */
    public AbsPanel( int width, int height ) 
    {
	setPreferredSize( new Dimension( width, height ) );            
	setLayout( new BorderLayout() );

	// make sure sliders are not too big
	Dimension sliderDim = new Dimension( 150, 30 );
        
	// box containing control components 	
	Box controlBox = Box.createHorizontalBox();
        controlBox.setPreferredSize( new Dimension( width, 150 ) );
	controlBox.add( Box.createHorizontalStrut(1) );
	add( controlBox, BorderLayout.NORTH );

	// grid bag layout/constraints used by sub-panels in control box
	gbLayout = new GridBagLayout(); 
	gbConstr = new GridBagConstraints();
	gbConstr.fill = GridBagConstraints.BOTH;
	gbConstr.weighty = 1;

	// controls for source system 
	JPanel srcPanel = new JPanel();
	srcPanel.setLayout( gbLayout );
	controlBox.add( srcPanel );	

        sourceLabel = new JLabel( "Source System" );
        addGridComp( sourceLabel, srcPanel, 0, 0, 1, 1, 0 );        
        importSrcButton = new JButton( "Import Source" );
	importSrcButton.addActionListener( this );
	addGridComp( importSrcButton, srcPanel, 2, 0, 1, 1, 0 );

        weightSrcLabel = new JLabel( "Weight Noise" );
        addGridComp( weightSrcLabel, srcPanel, 0, 1, 1, 1, 0 );        
        weightSrcSlider = new JSlider( 0, Absurdist.maxRCNoise );
	weightSrcSlider.setMajorTickSpacing(5);
	weightSrcSlider.setMinorTickSpacing(1);
	weightSrcSlider.setPaintTicks(true);
	weightSrcSlider.setPaintLabels(true);
	weightSrcSlider.setPreferredSize( sliderDim );
        addGridComp( weightSrcSlider, srcPanel, 0, 2, 1, 1, 0 );
        
        rewireSrcLabel = new JLabel( "Rewire Noise" );
        addGridComp( rewireSrcLabel, srcPanel, 1, 1, 1, 1, 0 );        
        rewireSrcSlider = new JSlider( 0, Absurdist.maxRCNoise );
	rewireSrcSlider.setMajorTickSpacing(5);
	rewireSrcSlider.setMinorTickSpacing(1);
	rewireSrcSlider.setPaintTicks(true);
	rewireSrcSlider.setPaintLabels(true);
	rewireSrcSlider.setPreferredSize( sliderDim );
        addGridComp( rewireSrcSlider, srcPanel, 1, 2, 1, 1, 0 );
        
        noiseSrcButton = new JButton( "Add Noise to Source" );
        noiseSrcButton.setEnabled( false ); // disabled until source is loaded
        noiseSrcButton.addActionListener( this );
	addGridComp( noiseSrcButton, srcPanel, 2, 1, 2, 1, 0 );

	// controls for mapping systems    
	JPanel mapPanel = new JPanel();
	mapPanel.setLayout( gbLayout );
	controlBox.add( Box.createHorizontalStrut(1) );
	controlBox.add( mapPanel );	

        exSimLabel = new JLabel( "External Similarity" );
        addGridComp( exSimLabel, mapPanel, 0, 0, 1, 1, 0 );        
        exSimSlider = new JSlider();
	exSimSlider.setMajorTickSpacing(50);
	exSimSlider.setMinorTickSpacing(10);
	exSimSlider.setPaintTicks(true);
	exSimSlider.setPaintLabels(true);
	exSimSlider.setPreferredSize( sliderDim );
        addGridComp( exSimSlider, mapPanel, 0, 1, 1, 1, 0 );

        mapButton = new JButton( "Map Systems" );
        mapButton.setEnabled( false ); // disabled until both systems are loaded
        mapButton.addActionListener( this );
        addGridComp( mapButton, mapPanel, 1, 0, 2, 1, 0 ); 	        

	/*
        viewButton = new JButton( "View Matrices" );
        viewButton.addActionListener( this );
        addGridComp( viewButton, mapPanel, 1, 0, 1, 1, 0 ); 	        
	*/

        exportButton = new JButton( "Export Results" );
        exportButton.setEnabled( false ); // disabled until systems are mapped
        exportButton.addActionListener( this );
        addGridComp( exportButton, mapPanel, 2, 0, 2, 1, 0 ); 	        

	// controls for target system 
	JPanel tgtPanel = new JPanel();
	tgtPanel.setLayout( gbLayout );
	controlBox.add( tgtPanel );	

        targetLabel = new JLabel( "Target System" );
        addGridComp( targetLabel, tgtPanel, 0, 0, 1, 1, 0 );        
        importTgtButton = new JButton( "Import Target" );
        importTgtButton.addActionListener( this );
	addGridComp( importTgtButton, tgtPanel, 2, 0, 1, 1, 0 );

        weightTgtLabel = new JLabel( "Weight Noise" );
        addGridComp( weightTgtLabel, tgtPanel, 0, 1, 1, 1, 0 );        
        weightTgtSlider = new JSlider( 0, Absurdist.maxRCNoise );
	weightTgtSlider.setMajorTickSpacing(5);
	weightTgtSlider.setMinorTickSpacing(1);
	weightTgtSlider.setPaintTicks(true);
	weightTgtSlider.setPaintLabels(true);
	weightTgtSlider.setPreferredSize( sliderDim );
        addGridComp( weightTgtSlider, tgtPanel, 0, 2, 1, 1, 0 );
        
        rewireTgtLabel = new JLabel( "Rewire Noise" );
        addGridComp( rewireTgtLabel, tgtPanel, 1, 1, 1, 1, 0 );        
        rewireTgtSlider = new JSlider( 0, Absurdist.maxRCNoise );
	rewireTgtSlider.setMajorTickSpacing(5);
	rewireTgtSlider.setMinorTickSpacing(1);
	rewireTgtSlider.setPaintTicks(true);
	rewireTgtSlider.setPaintLabels(true);
 	rewireTgtSlider.setPreferredSize( sliderDim );
        addGridComp( rewireTgtSlider, tgtPanel, 1, 2, 1, 1, 0 );
        
        noiseTgtButton = new JButton( "Add Noise to Target" );
        noiseTgtButton.setEnabled( false ); // disabled until target is loaded
        noiseTgtButton.addActionListener( this );
	addGridComp( noiseTgtButton, tgtPanel, 2, 1, 2, 1, 0 );       

        // display panel for graphs 
	graphPanel = new GraphPanel( absurdist );
	graphPanel.setBackground( Color.white );
	graphPanel.setPreferredSize( new Dimension( width, (height-150)*2/3 ) );
        graphPanel.addMouseListener( this );
        add( graphPanel, BorderLayout.CENTER );                

        // display panel for matrices
	matrixPanel = new MatrixPanel( absurdist );
        matrixPanel.setBackground( Color.white );
	matrixPanel.setPreferredSize( new Dimension( width, (height-150)/3 ) );
	matrixPanel.addMouseListener( this );
        add( matrixPanel, BorderLayout.SOUTH );                
    }

    // Add components to a container using grid layout.
    private void addGridComp( Component component, Container container, 
			      int row, int column, int width, int height, double weightx )
    {
	gbConstr.gridx = column;
	gbConstr.gridy = row;
	gbConstr.gridwidth = width;
	gbConstr.gridheight = height;
	gbConstr.weightx = weightx;
	gbLayout.setConstraints( component, gbConstr );
	container.add( component );
    }       


    /** Open an Absurdist session from a session file, and update the GUI. */
    public void openSession()
    {
	JFileChooser fileChooser = new JFileChooser( filedir );     
	int returnVal = fileChooser.showOpenDialog( this );
	String filepath = "";

	if ( returnVal == JFileChooser.APPROVE_OPTION ) {
	    ApplicDriver.debugln( "Open Absurdist session from a session file " + filepath );	
	    File file = fileChooser.getSelectedFile();
	    filepath = file.getPath();
	    filedir = file.getParentFile();

	    try {
		// read two systems and generate their graphs, reset status
		absurdist.readSession( filepath );
		int width = graphPanel.getWidth() / 2;
		int height = graphPanel.getHeight();
		ConceptGraph graphO = new ConceptGraph( absurdist.systemO, 0, 0, width, height );
		ConceptGraph graphN = new ConceptGraph( absurdist.systemN, width, 0, width, height );
		absurdist.setSystemO ( graphO );
		absurdist.setSystemN ( graphN );
		sourceLabel.setText( graphO.name );
		targetLabel.setText( graphN.name );

		// enable buttons and update displays
		noiseSrcButton.setEnabled( true );
		noiseTgtButton.setEnabled( true );
		mapButton.setEnabled( true );
		graphPanel.repaint();
		matrixPanel.invalidate();
	    }
	    catch ( Exception e ) {
		ApplicDriver.debugln( "OpenSession: serialization from file " + filepath +
				      " failed: " + e.getMessage() );
		e.printStackTrace( System.err );
	    }
	}	
    }

    /** Save an Absurdist session to a session file. */
    public void saveSession()
    {
	JFileChooser fileChooser = new JFileChooser( filedir );     
	int returnVal = fileChooser.showOpenDialog( this );
	String filepath = "";

	if ( returnVal == JFileChooser.APPROVE_OPTION ) {
	    ApplicDriver.debugln( "Save Absurdist session into a session file " + filepath );	
	    File file = fileChooser.getSelectedFile();
	    filepath = file.getPath();
	    filedir = file.getParentFile();

	    try {
		absurdist.writeSession( filepath );
	    }
	    catch ( Exception e ) {
		ApplicDriver.debugln( "SaveSession: serialization into file " + filepath +
				      " failed: " + e.getMessage() );
		e.printStackTrace( System.err );
	    }
	}
    }

    /** Imports the source/target system from a text data file. 
	@param src true if this is for source system
    */
    public void importSystem( boolean src )
    {
	JFileChooser fileChooser = new JFileChooser( filedir );     
	int returnVal = fileChooser.showOpenDialog( this );
	String filepath = "";
	String filename = "";

	if ( returnVal == JFileChooser.APPROVE_OPTION ) {
	    ApplicDriver.debugln( "Import concept system from a text data file " + filepath );	
	    File file = fileChooser.getSelectedFile();
	    filepath = file.getPath();
	    filedir = file.getParentFile();
	    filename = file.getName();

	    try {
		// read the system and generate its graph
		ConceptSystem system = ConceptSystem.readTxt( filepath );
		int width = graphPanel.getWidth() / 2;
		int height = graphPanel.getHeight();		
		int x = src ? 0 : width;	    
		ConceptGraph graph = new ConceptGraph( system, x, 0, width, height );

		// set Absurdist source/target system
		// update its name and enable corresponding noise button 
		if ( src ) {
		    absurdist.setSystemO( graph );		    
		    absurdist.printSystemO();
		    sourceLabel.setText( filename );
		    noiseSrcButton.setEnabled( true );
		}
		else {
		    absurdist.setSystemN( graph );
		    absurdist.printSystemN();		    
		    targetLabel.setText( filename );
		    noiseTgtButton.setEnabled( true );
		}

		// enable map button if both systems are set, update displays
		if ( absurdist.systemO != null && absurdist.systemN != null )
		    mapButton.setEnabled( true );
		graphPanel.repaint();
		matrixPanel.invalidate();
	    }
	    catch ( Exception e ) {
		ApplicDriver.debugln( "Open source/target system from file " + filepath +
				      " failed: " + e.getMessage() );
		e.printStackTrace( System.err );
	    }	    
	}
    }  

    /** Add noise to the source/target system. 
	@param src true if this is for source system
    */
    public void noiseSystem( boolean src )
    {
	JSlider weightSlider = src ? weightSrcSlider : weightTgtSlider;
	JSlider rewireSlider = src ? rewireSrcSlider : rewireTgtSlider;
	double weightNoise = (double)weightSlider.getValue() / 100; 
	double rewireNoise = (double)rewireSlider.getValue() / 100; 
	ConceptSystem system = src ? absurdist.systemO : absurdist.systemN;

	try {
	    // add weight & rewire noise to the system and generate its graph
	    system.addNoise( weightNoise, rewireNoise );
	    int width = graphPanel.getWidth() / 2;
	    int height = graphPanel.getHeight();		
	    int x = src ? 0 : width;	    
	    ConceptGraph graph = new ConceptGraph( system, x, 0, width, height );

	    // set Absurdist source/target system
	    // enable corresponding noise button
	    if ( src ) {
		absurdist.setSystemO( graph );		    
		absurdist.printSystemO();		    
		noiseSrcButton.setEnabled( true );
	    }
	    else {
		absurdist.setSystemN( graph );
		absurdist.printSystemN();		    
		noiseTgtButton.setEnabled( true );
	    }
	    
	    // enable map button if both systems are set, update displays
	    if ( absurdist.systemO != null && absurdist.systemN != null )
		mapButton.setEnabled( true );
	    graphPanel.repaint();	    
	    matrixPanel.invalidate();
	}
	catch ( Exception e ) {
	    ApplicDriver.debugln( "Add noise to source/target system failed: " + e.getMessage() );
	    e.printStackTrace( System.err );
	}	        
    }  

    /** Maps the two systems and updates the display. */
    public void mapSystems()
    {
	// disable these buttons until the computation thread finishes
	importSrcButton.setEnabled( false );
	importTgtButton.setEnabled( false );
	noiseSrcButton.setEnabled( false );
	noiseTgtButton.setEnabled( false );
	mapButton.setEnabled( false );
	exportButton.setEnabled( false );

	// set external similarity coefficient
	Absurdist.alphaShare = (double)exSimSlider.getValue() / 100; 

	// start animation
	AnimationThread animThread = new AnimationThread();
	animThread.start();  
	ApplicDriver.debugln( "Main thread: Animation thread has been spawned off." );	

	// start computation
	ComputationThread compThread = new ComputationThread( animThread );
	compThread.start();
	ApplicDriver.debugln( "Main thread: Computation thread has been spawned off." );
    }    

    /** Exports the mapping results into corresponding files. */
    public void exportResults()
    {
	ConceptSystem systemO = absurdist.systemO;
	ConceptSystem systemN = absurdist.systemN;
	String corName = systemO.name + systemN.name + ".cor";
	String mapName = systemO.name + systemN.name + ".map";
	File corFile = new File( filedir, corName );
	File mapFile = new File( filedir, mapName );
	try {
	    absurdist.writeCorrespond( corFile.toString() );
	    absurdist.writeMap( mapFile.toString() );
	}
	catch ( Exception e ) {
	    ApplicDriver.debugln( "Export mapping results to file failed: " + e.getMessage() );
	    e.printStackTrace( System.err );
	}	    	
    }

	
    // Action listener for text fields and buttons.
    public void actionPerformed( ActionEvent e )
    {     
        if ( e.getSource() == importSrcButton )  // import source system 
	    importSystem( true );
	else if ( e.getSource() == importTgtButton )  // import target system 
	    importSystem( false ); 
	else if ( e.getSource() == noiseSrcButton )  // add noise to source system 
	    noiseSystem( true ); 
	else if ( e.getSource() == noiseTgtButton )  // add noise to target system 
	    noiseSystem( false );  
	else if ( e.getSource() == mapButton )  // map concept systems 
	    mapSystems();
	else if ( e.getSource() == exportButton )  // export mapping results 
	    exportResults();
    }	
        
    public void mouseClicked( MouseEvent e )
    {    
    }
        
    public void mouseEntered( MouseEvent e )
    {    
    }        

    public void mouseExited( MouseEvent e )
    {    
    }        

    public void mousePressed( MouseEvent e )
    {    
    }        

    public void mouseReleased( MouseEvent e )
    {    
    }  

    public void mouseDragged( MouseEvent e )
    {    
    }  

    public void mouseMoved( MouseEvent e )
    {    
    } 


    // Thread to monitor the state of the absurdist object and updates various plots every once in a while.
    class AnimationThread extends Thread 
    {
	// flag to be set after the computation thread completes, and the animation thread is no longer needed
	public boolean finished = false;

	// Run the animation.
	public void run()
	{
	    ApplicDriver.debugln( "Animation thread: started running" );

	    while( !finished ) {
		if ( absurdist.isMapping() ) {
		    // initializes the maps/plots for the correspondence and net input matrices
		    // this actually gets done only once at the beginning of mapping
		    matrixPanel.setMatrices();

		    // update graph panel and matrix panel during animation
		    matrixPanel.setStep( absurdist.nStep );	    
		    		    
		    // ensure that painting is triggered within DT msec
		    graphPanel.repaint();
		    matrixPanel.repaint();	    
		}

		// try to sleep for a bit
		try { 
		    Thread.sleep( DT ); 
		}
		catch ( Exception ex ) {		    
		    ApplicDriver.debugln( "Exception caught: " + ex.getMessage());
		    ex.printStackTrace();
		    return; // this terminates the thread
		}		
	    }
		 	    	    		
	    // make sure the final results are displayed/highlighted before quitting
	    graphPanel.repaint();
	    matrixPanel.repaint(); 

	    ApplicDriver.debugln( "Animation thread: finished running" );	    
	}     
    }    


    // Thread to run the mapping algorithms.
    class ComputationThread extends Thread 
    {
	AnimationThread animThread;
	public ComputationThread( AnimationThread animThread ) 
	{
	    this.animThread = animThread;
	}

	// Runs the computation
	public void run() 
	{
	    ApplicDriver.debugln( "Computation thread: started running" );

	    try {
		absurdist.mapSystem();
		absurdist.printCorrespond();
		absurdist.printMap();
	    } 
	    catch ( Exception ex ) {
		ApplicDriver.debugln( "Exception caught: " + ex.getMessage());
		ex.printStackTrace();
		return;
	    } 
	    finally {
		// tell the animation thread to finish
		animThread.finished = true;

		// bring buttons back to life
		importSrcButton.setEnabled( true );
		importTgtButton.setEnabled( true );
		noiseSrcButton.setEnabled( true );
		noiseTgtButton.setEnabled( true );
		mapButton.setEnabled( true );
		exportButton.setEnabled( true );

		ApplicDriver.debugln( "Computation thread: finished running" );
	    }
	}
    }

}
