///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// ABSURDIST++
// Ying Feng
// Computer Scicence, Indiana Universit
// Feb-Aug 2003
//
//
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


 
package gui;

import java.awt.*;

import absurdist.*;
import driver.*;
import java.text.*;



/** 
    Class MatrixPlot
    Visual representation of an Absurdist matrix using line plot.
*/
public class MatrixPlot
{
    /** Offest of coordinate origin, to leave space for scales. */
    static final int OX = 40, OY = 20;  
 
    /** Number of pixels each iteration step (on x-axis) occupies */
    static final int DX = 1;

    /** Number of pixels each matrix unit value (on y-axis) occupies */
    int dy = 0;  // will compute to fit the graph
    
    /** Number of pixels each of x/y axis scale occupies */
    final int DSX = 50;
    ////int DSY = 20;  // will recompute

    double[][][] matrix;         // matrix the plot is for
    int nStep = 0;               // the actual # of steps recorded in matrix
    
    int x = 0, y = 0;            // top-left corner of its bound
    int width = 0, height = 0;   // width and height of its bound
    

    /** Creates the plot for the specified matrix and current step, 
	with the specified position and dimension. */
    public MatrixPlot( double[][][] matrix, int nStep, int x, int y, int width, int height )
    {
	this.matrix = matrix;
	this.nStep = nStep;
	this.x = x;
	this.y = y;
	this.width = width;
	this.height = height;	
    }


    /** Sets the matrix. */
    public void setMatrix( double[][][] matrix )
    {
        this.matrix = matrix;
    }
    
    /** Sets the number of steps. */
    public void setStep( int nStep )
    {
        this.nStep = nStep;
    }

    /** Sets the boundary of the plot. */
    public void setBound( int x, int y, int width, int height )
    {
        this.x = x;
	this.y = y;
	this.width = width;
	this.height = height;		
    }


    /** Returns a pair numbers that encloses the entire range of matrix[][][] */
    public int[] getRange() 
    {
	double minval = 0, maxval = 0;
	int no = matrix[0].length;
	int nn = matrix[0][0].length;

	if (nStep < 1) 
	    maxval = 1;
	else {
	    for ( int step = 1; step < nStep; step++ ) {
		for ( int io=0; io<no; io++ ) {
		    for ( int in=0; in<nn; in++ ) {
			double v = matrix[step-1][io][in];
			if (v < minval) minval=v;
			if (v > maxval) maxval=v;		
		    }
		}
	    }	
	}
	maxval = Math.ceil(maxval); // round up 
	minval = Math.floor(minval);

	int [] bounds = { (int)minval, (int)maxval };
	return bounds;
    }

        
    /** Paints the matrix plot.
	The formula for mapping v to y:
	<pre>
	y = oy - dy * (v - minval);
	</pre>
	where 
	<pre>
	dy = H/(maxval-minval);
	H=(height - OY) 
	</pre>
    */
    public void paint( Graphics2D g )
    {
	// draw frame
        g.setColor( Color.lightGray );
	g.setStroke( new BasicStroke(5) );
	g.drawRect( x, y, width, height );

	int ox = x + OX;            // x-axis goes right from origin
	int yh = height - OY;       // y-axis total height
	int oy = y + yh;            // y-axis goes up from origin
	int right = x + width;      // right end of x-axis
	int dxs = Absurdist.intLoop * DSX / DX; // interval of x-scale (step scale)

	// Adjustable scale (V.M., Jul-2-2003)
	final int ranges[] = getRange();
	final int minval = ranges[0];
	final int maxval = (ranges[1] > 0 ? ranges[1] : 1); // to avoid zero divide
	final int range = maxval - minval;	
	dy = yh / range;

	Font oldFont = g.getFont();
	Font font = oldFont.deriveFont(10);
	g.setFont( font );
	FontMetrics fm = g.getFontMetrics();
	int fa = fm.getAscent();

	// draw x-axis (iteration steps)
        g.setColor( Color.black );	
        g.drawLine( ox, oy, right, oy );

	// draw scales on x-axis, starting from first one after origin
	int sx = ox + DSX;     // current x-scale x-pos on right end
	int sxy = oy + 5 + fa; // current x-scale y-pos
	int xs = dxs;          // current x-scale
	while ( sx < right ) {
	    String xsstr = "" + xs; 
	    int sw = fm.stringWidth( xsstr );
	    g.drawString( xsstr, sx-sw, sxy ); // align right
	    sx += DSX;
	    xs += dxs;
	}	    

	// draw y-axis (matrix values)	
        g.setColor( Color.black );	
        g.drawLine( ox, oy, ox, y );

	// draw scales on y-axis
	DecimalFormat df = new DecimalFormat( "0.00" );	
	int NSV = 4;
	int jmin = NSV * minval / maxval;
	for ( int j=jmin ; j<=NSV; j++ ) {
	    double v = j * maxval / (double)NSV;
	    String ysstr = df.format( v );
	    int sy  = oy - (int)( (v - minval) * dy ) + 5 + fa;
	    g.drawString( ysstr, x+5, sy ); // align top
	}

	g.setFont( oldFont );
	
	// draw line in plot
        g.setColor( Color.blue );	
	g.setStroke( new BasicStroke(1) );
	int no = matrix[0].length;
	int nn = matrix[0][0].length;
	for ( int step = 1; step < nStep; step++ ) {
	    int x1 = ox + ( step - 1 ) * DX;
	    int x2 = ox + step * DX;
	    for ( int io=0; io<no; io++ ) 
		for ( int in=0; in<nn; in++ ) {
		    int y1 = oy - ( int )( (matrix[step-1][io][in]-minval) * dy );
		    int y2 = oy - ( int )( (matrix[step][io][in]-minval) * dy );
		    g.drawLine( x1, y1, x2, y2 );
		}
	}
    }
}
