package penalty;

import common.*;
import absurdist.*;

    
import java.io.*;
import java.util.*;
import java.text.*;


/** This algorithm tries to compute the correspondence between two
concept systems that minimizes a certain penalty function. The penalty
function is constructed so that there is a penalty imposed whenever
there is a correspondence between node a in the first concept system
and node x in the second concept system, and a correspondence between
node b and node y, while the relationship between a and b in the
first system is not identical to the relationship between x and y in the 
second system.

<P>
The penalty function is defined as a function of a correspondence matrix
C as follows:

<pre>
Penalty(C) = MatchingPenalty(C) + kappa * NormalizationPenalty(C)
</pre>

Here, MatchingPenalty(C) is designed as to minimize the degree of
mismatch between the structures of the graphs (i.e., for two graphs
that are isomorphic, the absolute minimum of MatchingPenalty(C)=0 will
be reached on a correspondence matrix C where C(a,x)!=0 only
on pairs (a,x)=(a,f(a)) for some isomorphism f). NormalizationPenalty(C)
is designed to be zero on a matrix C for which the values in 
each row and each column sum to 1. (Or to sqrt(m/n) and sqrt(n/m),
respectively, if the two systems have a different number of nodes).
Namely,

<pre>
MatchingPenalty(C) = 
  = K1 * sum_{a,b,x,y: linked(a,b) & !linked(x,y)}{ C(a,x)*C(b,y)} +
  + K1 * sum_{a,b,x,y: !linked(a,b) & linked(x,y)}{ C(a,x)*C(b,y)} +
  + K2 * sum_{a,x,y:  !linked(x,y)}{ C(a,x)*C(a,y)} +
  + K2 * sum_{a,b,x:  !linked(a,b)}{ C(a,x)*C(b,x)} +
  + K3 * sum_{a,x,y:  linked(x,y)}{ C(a,x)*C(a,y)} +
  + K3 * sum_{a,b,x:  linked(a,b)}{ C(a,x)*C(b,x)}.
</pre>

Thus, K1 is the penalty for matching a pair of linked nodes to a pair
of non-linked nodes, or vice versa; K2 is the penalty for matching a
single node to two non-linked nodes, or vice versa; and K3 is the
penalty for matching a single node to two linked nodes, or vice versa.

<pre>
NormalizationPenalty(C) =
   = sum_{a}{ (sum_{x}{C(a,x)} - sqrt(m/n))^2 } +
   + sum_{x}{ (sum_{a}{C(a,x)} - sqrt(n/m))^2 }.
</pre>

<p>
The approach used in this method, that is, minimizing
MatchingPenalty(C)+kappa*NormalizationPenalty(C) on the set of all
matrices with non-negative coeffciients, is, really, a cludge, and may
only produce an approximate solution. To obtain an exact solution, one
would want to minimize MatchingPenalty(C) over the set of all
matrices for which NormalizationPenalty(C)=0. But the latter would
be a somewhat more complex quadratic-programming problem.

<p>
This program uses a very traditional fastest descent method, and works
as follows:

<pre>
1. We take an initial point C (a random matrix with positive
coefficients);

2. We compute the gradient of the penalty functional g; we
know that C is increases in the direction of vector g, and decreases
in the direction  of -g;

3. We obtain the "restricted" ascent vector v. v is based on g,
but if the current point C alread has some zero coefficients we
may replace the corresponding component of v with 0 to ensure that
going along the direction of -v won't make that component of C negative.

4. We solve the one-dimension minimization problem, finding the
real number xi that minimzes Penalty( C - xi * v), while C'=C-xi*v
still stays in the non-negative domain.

5. Assign C := C' = C - xi*v, and continue to step 2.
</pre>

<p>
This program termiates after a specified number of iterations,
or when it is detected that the penalty function cannot be
noticeably reduced any more. (That is, when Penalty(C)-Penalty(C')
is very small, or when v is almost zero).

<p>
The general outline of the code is based on that of 
{@link absurdist.Absurdist Absurdist} class, as it stood in April 2003.

@see Absurdist

*/

public class PenaltyMinimizer implements SystemMatcher
{

    static final boolean debug=false;

    /** 
	K1 and K3 are the mapping penalty parameters.
In general, the elements of the penalty matrix are as follows:
<pre>
	    C(a,a)  C(a,b)  C(a,c)
   C(x,x)   0       K3      K2
   C(x,y)   K3      0       K1
   C(x,z)   K2      K1      0
</pre>   
Here, a is a node in the first graph; b is a node in the first graph
linked to a; c is a node in the first graph NOT linked to a.
Similarly, x is a node in the second graph; y is a node in the second graph
linked to x; z is a node in the second graph NOT linked to x.
<p>

Thus, K1 is a penalty for mapping a pair of linked nodes to a pair
of distinct but not lined nodes; K2 is a penalty for mapping a single
node to two distinct, not linked nodes; and K3 is a penalty for mapping
a single node to two distinct linked nodes.
<p>

However, in order to be able to construct an efficient implementation,
we always have K1=K2. Thus, K1=K2 is a penalty for a mapping two
<p>
    */
    public double K1, K2, K3;

    /** Kappa is the normzlization penalty parameter. The larger it
	is, the closer the solution will be to the "normalized-matrix
	plane", where the values in each row and each column of the
	matrix sums to 1.  */
    public double kappa;
    
    ConceptSystem systemO, systemN; // original system and system added with noise
    /** Graph-style interface for systemO and systemN */
    CSGI sysO, sysN;

    double[][] correspond;          // correspondence matrix between the two systems    
    int[] mapON, mapNO;             // mappings from systemO to systemN, and vice versa
    double accuracy;		    // mapping accuracy

    /** Maximum number of iterations. It would be wise to set this
	to, at least, systemO.n * systemN.n
    */
    int nLoop = 1000;

    double mapT=0;

    public PenaltyMinimizer()
    {
	/* // Option 1: this will get you a "permutation-like" matrix,
	   // but it will be just a local minimum 
	K1=1;
	K2=10;
	K3=10;
	kappa = 100;
	*/

	K1=10;
	K2=1;
	K3=1;
	kappa = 100;


	/*
	K1=K2=K3=0;
	kappa=1;
	*/
    }


    // Set the two concepts systems for translation.
    public void setSystem( ConceptSystem systemO, ConceptSystem systemN )
    {
	this.systemO = systemO;
	this.systemN = systemN;
	// Prepare a graph-style interface, to exploit sparseness
	sysO = new CSGI( systemO);
	sysN = new CSGI( systemN);
	//// correspond = new double[ systemO.getConceptNum() ][ systemN.getConceptNum() ];	
    }
          

    // Adjust beta & chi coefficeint according to the node number and edge fullness of the graph. 
    public void adjustCoeff()
    {
	int no = systemO.getConceptNum();
	int nn = systemN.getConceptNum();
        int n = Math.min( no, nn );
    }


    /** Computes the normalization penalty
	|B*c-z*b|^2
    */
    double normPenalty(CorrespondenceVector c, double zu) {
	return normPenalty(c, zu, c, zu);
    }

    /** Computes the dot product of two vector combinations.

	@return
	<pre>
	(B*u-z_u * b, B*v-z_v * b) =
	sum_a { (sum_x{u(a,x)} - sqrt(m/n))*(sum_x{v(a,x)} - sqrt(m/n)) }
	sum_x { (sum_a{u(a,x)} - sqrt(m/n))*(sum_a{u(a,x)} - sqrt(m/n)) }
	</pre>
    */
    double normPenalty(CorrespondenceVector u, double zu,
		       CorrespondenceVector v, double zv) {

	double [] ru=u.rowSum(), rv=v.rowSum();
	double [] cu=u.colSum(), cv=v.colSum();
	double q = Math.sqrt( cu.length/(double)ru.length);
	double sum1=0, sum2=0;
	for(int i=0; i<ru.length; i++) {
	    sum1 += (ru[i]-zu*q)  * (rv[i]-zv*q);
	}
	q = 1/q;
	for(int i=0; i<cu.length; i++) {
	    sum2 += (cu[i]-zu*q)  * (cv[i]-zv*q);
	}	
	return sum1 + sum2;
    }


    double matchingPenalty(CorrespondenceVector u,
			   CorrespondenceVector v) {
	if (u.n != systemO.getConceptNum() || u.m != systemN.getConceptNum())
	    throw new IllegalArgumentException("u");
	if (v.n != systemO.getConceptNum() || v.m != systemN.getConceptNum())
	    throw new IllegalArgumentException("v");
	double total=0;
	double s=0;
	double [] ru=u.rowSum(), rv=v.rowSum();
	double [] cu=u.colSum(), cv=v.colSum();

	// (1): K1*...
	s = 0;
	for(int a=0; a<u.n; a++) {
	    int bList[] = sysO.getNei(a);
	    for(int j=0; j<bList.length; j++) {
		int b = bList[j];
		s += ru[a] * rv[b];
	    }
	}
	for(int x=0; x<u.m; x++) {
	    int yList[] = sysN.getNei(x);
	    for(int j=0; j<yList.length; j++) {
		int y = yList[j];
		s += cu[x] * cv[y];
	    }
	}
	for(int a=0; a<u.n; a++) {
	    int bList[] = sysO.getNei(a);
	    for(int j=0; j<bList.length; j++) {
		int b = bList[j];
		for(int x=0; x<u.m; x++) {
		    int yList[] = sysN.getNei(x);
		    for(int k=0; k<yList.length; k++) {
			int y = yList[k];
			s -= u.d[a][x] * v.d[b][y];
		    }
		}
	    }
	}
	total += K1 * s;

	//2. K2*...
	s=0;
	for(int a=0; a<u.n; a++) {
	    s += ru[a]*rv[a];
	}
	for(int x=0; x<u.m; x++) {
	    s += cu[x]*cv[x];
	}
	s -= u.dp(v);
	total += K2 * s;

	//3. (K3-K1-K2)*...
	s=0;
	for(int a=0; a<u.n; a++) {
	    for(int x=0; x<u.m; x++) {
		int yList[] = sysN.getNei(x);
		for(int j=0; j<yList.length; j++) {
		    int y = yList[j];
		    s +=  u.d[a][x]*v.d[a][y];
		}
	    }
	}
	for(int x=0; x<u.m; x++) {
	    for(int a=0; a<u.n; a++) {
		int bList[] = sysO.getNei(a);
		for(int j=0; j<bList.length; j++) {
		    int b = bList[j];
		    s +=  u.d[a][x]*v.d[b][x];
		}
	    }
	}
	total += (K3-K1-K2)*s;
	return total;
    }

    double totalPenalty(CorrespondenceVector u, double zu,
			CorrespondenceVector v, double zv) {
	return matchingPenalty(u, v) + kappa*normPenalty(u,zu, v, zv);
    }

    double totalPenalty(CorrespondenceVector u) {
	return totalPenalty(u,1.0, u, 1.0);
    }

    
    CorrespondenceVector gradMatchingPenalty( CorrespondenceVector u) {
	CorrespondenceVector grad=new CorrespondenceVector( u.n, u.m);

	double [] ru=u.rowSum(), cu=u.colSum();
	
	for(int a=0; a<u.n; a++) {
	    for(int x=0; x<u.m; x++) {
		double total =0;
		double s=0;

		// 1.
		s=0;		
		int bList[] = sysO.getNei(a);
		for(int j=0; j<bList.length; j++) {
		    int b = bList[j];
		    s+= ru[b];
		}
		int yList[] = sysN.getNei(x);
		for(int j=0; j<yList.length; j++) {
		    int y = yList[j];
		    s +=  cu[y];
		}
		for(int j=0; j<bList.length; j++) {
		    int b = bList[j];
		    for(int k=0; k<yList.length; k++) {
			int y = yList[k];
			s -=  u.d[b][y];
		    }
		}
		total += K1 * s;

		// 2.
		s = ru[a] + cu[x] - u.d[a][x];
		total += K2 * s;

		// 3.
		s = 0;
		for(int j=0; j<bList.length; j++) {
		    int b = bList[j];
		    s += u.d[b][x];
		}
		for(int j=0; j<yList.length; j++) {
		    int y = yList[j];
		    s += u.d[a][y];
		}
		total += (K3-K1-K2)*s;
		
		grad.d[a][x]=total;
	    }
	}
	return grad;
    }
  
    CorrespondenceVector gradNormPenalty( CorrespondenceVector u) {
	CorrespondenceVector grad=new CorrespondenceVector( u.n, u.m);

	double [] ru=u.rowSum(), cu=u.colSum();
	double q = Math.sqrt( cu.length/(double)ru.length);

	for(int a=0; a<u.n; a++) {
	    for(int x=0; x<u.m; x++) {
		grad.d[a][x] = ru[a] + cu[x] - (q+1.0/q);
	    }
	}
	return grad;
    }


    CorrespondenceVector gradTotalPenalty( CorrespondenceVector u)     {
	CorrespondenceVector g = gradMatchingPenalty(u);
	CorrespondenceVector g2 = gradNormPenalty(u);
	g.daxpy( kappa, g2);
	return g;
    }

    /** Creates a ranomly-initialized vector. It is in the positive
	octant, but not 1-normalized */
    CorrespondenceVector initRandom() {
	int n = systemO.getConceptNum(), m = systemN.getConceptNum();
	CorrespondenceVector v = new CorrespondenceVector(n,m);
	final double mid=0.5, range = 0.6;
	final double base = mid- range*0.5;
	for(int a=0; a<v.n; a++) {
	    for(int b=0; b<v.m; b++) {
		v.d[a][b] = base + range * Math.random();
	    }
	}
	v.flush();
	return v;
    }    


    void restrictGrad(CorrespondenceVector g, 
		      final CorrespondenceVector c) {
	for(int a=0; a<g.n; a++) {
	    for(int x=0; x<g.m; x++) {
		if (g.d[a][x]>0 && c.d[a][x]<=0) {
		    g.d[a][x] = 0;
		}
	    }
	}
	g.flush();
    }

    /** Ensures that the vector  c - xi*v stays in the positive octant */
    double restrictXi(double xi, 
		      final CorrespondenceVector v, 
		      final CorrespondenceVector c) 
    throws Exception {
	final double eps=1e-9;
	if (xi <=0) {
	    throw new
		IllegalArgumentException("xi is expected to be positive!");
	}
	for(int a=0; a<v.n; a++) {
	    for(int x=0; x<v.m; x++) {
		double c1=c.d[a][x] - xi * v.d[a][x];
		if (c1 < 0) {
		    if (c.d[a][x] <0) {
			String msg="C is expected to be non-negative!";
			throw new IllegalArgumentException(msg);
		    }
		    double xi1 = c.d[a][x] / v.d[a][x];
		    if (xi1 >= xi) {
			if (xi < eps) {
			    // Well, we're as goood as converged anyway
			    return 0.0;
			}
			String msg=
			    "Internal error? (xi="+xi+", xi'="+xi1+ ")";
			throw new Exception(msg);
		    }
		    xi = xi1;
		}
	    }
	}
	return xi;
    }

   /** Replaces near-zero coordinate values of the correpsondence
       matrix c with exact zeros. This has two purposes:
       <ol>
       <li> it ensures that the vector  c  stays in the non-negative octant;
       <li> it allows proper component removal in vector v in restrictGrad.
       </ol>
       <p>

       This method expects that all components of c are already almost
       non-negative */
    void smooth( CorrespondenceVector c) {
	final double eps = 1e-6;
	for(int a=0; a<c.n; a++) {
	    for(int x=0; x<c.m; x++) {
		if (Math.abs(c.d[a][x]) <eps) {
		    // Making a "near-zero" into 
		    c.d[a][x] = 0;
		} else if (c.d[a][x]<0) {
		    String msg= "c is expected to be almost non-negative!";
		    throw new IllegalArgumentException(msg);
		} 
	    }
	}
	c.flush();
    }
	


    /** Iterates, using the steepest descent method, to 
	find a correspondence between two concept systems
	minimizing the penalty functional.
	<p>

	In practice, this methods often converges to a local
	minimum, rather than the global one, thus providing
	a poor quality match.
    */
    public void iterateNet() throws Exception {
	final boolean doRestrict=true;
	final double eps=1e-9;

	PrintWriter cout = new PrintWriter(System.out);

	int iter=0;
	CorrespondenceVector c=initRandom();
	CorrespondenceVector v=null,g=null;
	double f = totalPenalty(c);
	correspond = c.getMatrix();
	for(iter=0; iter<nLoop; iter++) {
	    if (debug) cout.println("Iter " + iter +"; F(c)=" + f);
	    g=gradTotalPenalty(c);
	    if (debug) cout.println("|c|^2=" + c.dp(c) +", |grad F(c)|^2=" + g.dp(g));

	    v = (CorrespondenceVector)g.clone();

	    if (doRestrict) restrictGrad(v,c);
	    // F(c - xi*v) = F(c) - 2*lambda*xi + mu*xi^2, 
	    // where
	    // lambda = (v,grad)F(c) = F(c,1, v,0)
	    // mu = F(v,0, v,0)
	    double lambda = totalPenalty(c,1, v,0);
	    double mu = totalPenalty(v,0, v,0);
	    if (debug) cout.println("|v|^2="+v.dp(v)+ "; lambda=(Gc,v)=" + lambda+
			 "; mu=(Gv,v)="+ mu);

	    double xi;
	    if (Math.abs(lambda) < eps) break;
	    if (lambda>=0) {
		if (mu>0) xi = lambda/mu;
		else xi = 1000;
	    } else {
		throw new Exception("Optimization error: lambda=(Ac,v)<0");
	    }
	    double xi0=xi;
	    if (xi < 0) {
		throw new Exception("Optimization error: xi<0; function non-convex!\n"); 
	    }
	    if (doRestrict) xi = restrictXi(xi, v, c);
	    c.daxpy(-xi, v);
	    if (doRestrict) smooth(c);
	    correspond = c.getMatrix();
	    double f0 = f;
	    f = totalPenalty(c);
	    if (debug) cout.println("xi_orig="+xi0+", xi=" + xi +"; delta F=" + (f-f0));
	    if (xi==0.0) break; // No point to continue
	}

	cout.println(" After " + iter + " iterations:");
	cout.println(" C=");
	printMatrix(c.getMatrix(), cout);
	cout.println(" grad C=");
	printMatrix(g.getMatrix(), cout);
	cout.println(" V=");
	printMatrix(v.getMatrix(), cout);

	cout.flush();

    }

    /** Calculates the similarity between R(Cio, Cjo) in systemO and
     R(Cin, Cjn) in systemN, counting all types of relations in both
     directions.  */
    private double calcLinkSim( int io, int jo, int in, int jn )
    {
	int ko, kn;
	double dw = 0, sim = 0;
	Relation ro, rn;
	Vector rso, rsn;

	// add relation weight differences in direction i->j
	rso = ( Vector )systemO.getRelations( io, jo );
	rsn = ( Vector )systemN.getRelations( in, jn );
	if ( rso == null && rsn == null ) {} // don't do anything
	else if ( rsn == null ) {
	    for ( ko=0; ko < rso.size(); ko++ ) {
		ro = ( Relation ) rso.get(ko);
		dw += ro.weight;
	    }
	}
	else if ( rso == null ) {
	    for ( kn=0; kn < rsn.size(); kn++ ) {
		rn = ( Relation ) rsn.get(kn);
		dw += rn.weight;
	    }
	}
	else {
	    ko = 0;
	    kn = 0;
	    while ( ko < rso.size() || kn < rsn.size() ) {
		ro = ( Relation ) rso.get(ko);	
		rn = ( Relation ) rsn.get(kn);
		if ( kn == rsn.size() || ro.compareTo( rn ) < 0 ) { // rn missing
		    dw += ro.weight;
		    ko++;
		}
		else if ( ko == rso.size() || ro.compareTo( rn ) > 0 ) { // ro missing
		    dw += rn.weight;
		    kn++;	
		}	
		else { // both exist
		    dw += ro.difWeight( rn );
		    ko++;
		    kn++;
		}
	    }
	}

	// add relation weight differences in direction j->i
	rso = ( Vector )systemO.getRelations( jo, io );
	rsn = ( Vector )systemN.getRelations( jn, in );
	if ( rso == null && rsn == null ) {} // don't do anything
	else if ( rsn == null ) {
	    for ( ko=0; ko < rso.size(); ko++ ) {
		ro = ( Relation ) rso.get(ko);
		dw += ro.weight;
	    }
	}
	else if ( rso == null ) {
	    for ( kn=0; kn < rsn.size(); kn++ ) {
		rn = ( Relation ) rsn.get(kn);
		dw += rn.weight;
	    }
	}
	else {
	    ko = 0;
	    kn = 0;
	    while ( ko < rso.size() || kn < rsn.size() ) {
		ro = ( Relation ) rso.get(ko);	
		rn = ( Relation ) rsn.get(kn);
		if ( kn == rsn.size() || ro.compareTo( rn ) < 0 ) { // rn missing
		    dw += ro.weight;
		    ko++;
		}
		else if ( ko == rso.size() || ro.compareTo( rn ) > 0 ) { // ro missing
		    dw += rn.weight;
		    kn++;	
		}	
		else { // both exist
		    dw += ro.difWeight( rn );
		    ko++;
		    kn++;
		}
	    }
	}
	
	dw /= 2 * systemO.getModelNum(); // the two systems have the same z# of models
	if ( dw < 1 ) {
	    sim = 1 - dw;
	} else {
	    sim = 0;
	} 
	return sim;

	/*
	double lo = systemO.getRelWeight( io, jo);
	double ln = systemN.getRelWeight( in, jn );
	double sim = 0;
	if ( lo-ln == 0 ) 
	    sim = 1;
	else	    
	    sim = 0;
	return sim;
	*/
    }

    /** Finds the strongest valid correspondence as successful mappings.
     */
    public void findMapping()
    {
        //Test.debug( "Finding mapping ..." );

        int no = systemO.getConceptNum();
        int nn = systemN.getConceptNum();
        int io, in;
	mapON = new int[no];
	mapNO = new int[nn];
        for ( io=0; io<no; io++ ) 
            mapON[io] = -1;
        for ( in=0; in<nn; in++ ) 
            mapNO[in] = -1;
        
        for ( int k=0; k<no; k++ ) {
	    double max = mapT;
            int mo = -1, mn = -1;
            for ( io=0; io<no; io++ ) {
                if ( mapON[io] != -1 ) 
                    continue;
                for ( in=0; in<nn; in++ ) {
                    if ( mapNO[in] != -1 ) 
                        continue;
                    if ( mo<0 || correspond[io][in] > max ) {
                        mo = io;
                        mn = in;
                        max = correspond[io][in];
                    }
                }
            }
            if ( mo != -1 && mn != -1 ) {
                mapON[mo] = mn;
                mapNO[mn] = mo;
                /*
                mapON[mo] = systemN.getConcept(mn);
                mapNO[mn] = systemO.getConcept(mo);
                */
            }
            else
                break;
        }
    }	       
    
    /*
    // Find the strongest valid correspondence as successful mappings.
    public void findMapping()
    {
        int no = systemO.getConceptNum();
        int nn = systemN.getConceptNum();
	mapping = new Concept[no];
	for ( int io=0; io<no; io++ ) {	    
	    double max = thresh;
            mapping[io] = null;
	    for ( int in=0; in<nn; in++ ) {
		if ( correspond[io][in] > max ) {
                    max = correspond[io][in];
		    mapping[io] = systemN.getConcept(in);
                }
	    }
	}
    }	       
    */
    
    /** Computes the accuracy of the mapping. */
    public void evalMap()
    {
        //Test.debug( "Evaluate mapping ..." );
    
        int no = systemO.getConceptNum();
        int nn = systemN.getConceptNum();
        double sim = 0;

        for ( int io=0; io<no; io++ ) {    
            for ( int jo=0; jo<no; jo++ ) {
                int in = mapON[io];
                int jn = mapON[jo];  
                if ( in != -1 && jn != -1 )              
                    accuracy += calcLinkSim( io, jo, in, jn ) * correspond[io][in] * correspond[jo][jn];
            }
        }
        accuracy /= no * no;
        //// return sim;
    }                

    /** An alternative way to evaluate the accuracy of the mapping.
	@return The number that counts edges present in one graph in
	absent in its "image", and vice versa. This should be 0 is
	the mapping is a homomorphism, i.e. a perfect matching
	
	date: April-10-2003
    */
    public double evalMap2()     {
        //Test.debugln( "Evaluate mapping ..." );
    
        int no = systemO.getConceptNum();
        int nn = systemN.getConceptNum();
        double sim = 0;
        //// int nStep = nLoop / DLOOP;

	double  errCnt =0;
        for ( int io=0; io<no; io++ ) {    
            for ( int jo=0; jo<no; jo++ ) {
                int in = mapON[io];
                int jn = mapON[jo];  
                if ( in != -1 && jn != -1 ) {
		    errCnt += Math.abs(1 - calcLinkSim( io, jo, in, jn ));
		}
            }
        }
	return errCnt;
        //// return sim;
    }        

    
    /** Maps orginal concept system to the one with added noise.  */
    public void mapSystem( ConceptSystem systemO, ConceptSystem systemN )
    throws Exception
    {
        setSystem( systemO, systemN );
        adjustCoeff();
        iterateNet();
        findMapping();
        evalMap();
    }

    public void printSystem( PrintWriter out ) throws Exception
    {
        //// PrintWriter out = new PrintWriter( new FileOutputStream( "system.dat" ) );
        System.out.println( "\n" + systemO );
        System.out.println( "\n" + systemN );
	if (out != null) {
	    out.println( "\n" + systemO );
	    out.println( "\n" + systemN );
	}
    }
    
    public void printCorrespond( PrintWriter out ) throws Exception {
        System.out.println( "\nCorrespondence matrix -----------" );
	PrintWriter cout = new PrintWriter(System.out);
	printMatrix( correspond, cout);
	cout.flush();
	if (out != null) {
	    out.println( "\nCorrespondence matrix -----------" );
	    printMatrix( correspond, out);
	}
    }

    public void printMatrix( double [][]c, PrintWriter out ) throws Exception
    {
	if (out == null) return;
	DecimalFormat df=new DecimalFormat("0.000");
        for ( int io=0; io<c.length; io++ ) {
            for ( int in=0; in<c[io].length; in++ ) {
                out.print( df.format(c[io][in]) + " " );
            }
            out.println();
        }
        //// out.close();
    }
    
    public void printMapping( PrintWriter out ) throws Exception
    {
	/// How well does the mapping match edges in the two graphs?
	double errCnt=evalMap2();
	String msg = "Mismatched edge count: " + errCnt;
	if (errCnt==0) msg += "; perfect matching!";
	System.out.println(msg);
	if (out!=null) out.println(msg);


        int no = systemO.getConceptNum();
        //// PrintWriter out = new PrintWriter( new FileOutputStream( "mapping.dat" ) );
	msg = "\nMapping obtained using K1="+K1+", K2="+K2+", K3="+K3+
	    ", kappa="+kappa;
	System.out.println(msg);
	if (out!=null) out.println(msg);
        System.out.println( "Mapping with accuracy " + accuracy + "-----------" );
        if (out!=null) out.println( "Mapping with accuracy " + accuracy + " -----------" );
        for ( int io=0; io<no; io++ ) {
            int in = mapON[io];
            if ( in != -1 ) {
		msg = systemO.getConcept(io).name + " --> " + 
		    systemN.getConcept(in).name;		
            }
            else {
		msg = systemO.getConcept(io).name + " --> null";
            }
	    System.out.println( msg);
	    if (out!=null) out.println( msg);
        }
    } 

    
    /** Prints the original and noisy concept systems, the
        correspondence matrix, and the final mappings. */
    public void print(boolean toFile) throws Exception
    {
	PrintWriter out=null;
	if (toFile) {
	    out = new PrintWriter( new FileOutputStream( systemO.name +systemN.name + ".dat" ) );
	}
	printSystem( out );
	printCorrespond( out );
	printMapping( out );
	if (toFile) out.close();
    }
}
