#!/usr/bin/perl

#Usage: splitAnchors.pl input.txt > out.txt

# This program reads a file describing a graph, and produces an output 
# based on the input file, but with the following transformations:
#
# 1) Removes all references to relation R1
#
# 2) Assumes that every node with a name starting with a '?' is an
# "internal node", and all other nodes are leaves. Keeps internal
# nodes and removes leaves.
#
# 3) Creates one more section in the file, names "Anchors:", in which
# after the name of an internal node A it will list all leaves L
# to which the node was connected by R1(A,L).


my $section=undefined;
my $s;

# Maps an internal node to the least of leaves attached to it
my %internal2leaves=();

# Keys are all leaves
my %leaves = ();

# Read the file line by line
while(defined($s=<>)) {

    my $text = $s;
    # Strip leading and trailing spaces
    $text =~ s/^\s+//;
    $text =~ s/\s+$//;
    if ($text =~ "^ConceptCategories:" ||
	$text =~ "^RelationModels:" ||
	$text =~ "^Concepts:" ||
	$text =~ "^Relations:" ) {

	# Section header found
	$section = $text;
	# Strip ':' and everything that follows
	$section =~ s/:.*//;

    } elsif ($section eq "RelationModels") {

	if ($text =~ /^R1:/) {
	    # Remove the definition of R1
	    $s = "";
	} 

    } elsif ($section eq "Concepts") {

	if ($text =~ /(\S+)\s*:/) {
	    # Was a concept...
	    my $concept = $1;

	    if ($text =~ /^\?/) {
		# This was an internal node; keep it
		$internal2leaves{$concept} = [];
	    } else {
		# This was a leaf; remove it
		$leaves{$concept} = 1;
		$s = "";
	    }
	}

    } elsif ($section eq "Relations") {
	
	# Strip all spaces, for easy parsing
	$text =~ s/\s//g;

	if ($text =~ /R1\((\S+),(\S+)\)/ ) {
	    my ($a,$b) = ($1,$2);
	    # Check  if 'a' and 'b' are indeed known as an internal node
	    # and a leaf
	    if (!exists $internal2leaves{$a}) {
		die "Don't know about internal node `$a'!\n";
	    } elsif (!exists $leaves{$b}) {
		die "Don't know about leaf `$b'!\n";
	    } else {
		# Add leaf 'b' to the list of anchors for 'a'
		push( @{$internal2leaves{$a}}, $b);
	    }
	    # Remove the link from the file
	    $s = "";
	}

	


    }

    print $s;

}

# Now, print the anchor info

print "\n";
print "Anchors:\n\n";
foreach my $node (sort keys %internal2leaves) {
    my @vals = @{$internal2leaves{$node}};
    if (@vals > 0) {
	print "$node: " . join( ', ', @vals) . "\n";
    }

}
