package test;

import java.io.*; 
import java.text.*; 

import absurdist.*;

/** The base class for all Test classes that were prepared originally
    for the FLAIRS paper and then updated for the IJAIT paper (July
    2004). Contains utilities used for reporting test results. */
class TestBase {
   
    public static boolean debugMode = true;
    /**  If the derived class sets this to true, absurdist.print()
	 will be triggered, printing both systems etc */
    public static boolean debugAll = false;

    public static  int NRuns = 100;

    /** Optional: name of the old results file (from an earlier
	interrupted run) to recover.  */
    public static String recover=Absurdist.prop.getOption("recover", null);

    public static void debug( String s )
    {
        if ( debugMode )
            System.out.print( s );
    }
        
    public static void debugln( String s )
    {
        if ( debugMode )
            System.out.println( s );
    }
        
    public static void error( String s )
    {
        //// if ( errorMode )
        System.err.println( s );
    }
      
 
    static DecimalFormat dfa = new DecimalFormat( "0.00000" );

    /** Local copies of important variables. They are maintained as
	instance members so that getSysO(), getSysN() etc. could
	access them .*/
    int n;
    double fullness, noise;
    ConceptSystem systemO;
    ConceptSystem systemN;
    PrintStream out;

    TestBase() {}

    /** Creates the base system. Derived classes may redefine
	this method according to how they need to create it.
    */
      
    ConceptSystem getSysO() throws Exception {
	return new ConceptSystem( n, fullness );
    }

    /** Creates the noisy system. Derived classes may redefine
	this method according to how they need to create it.
    */
    ConceptSystem getSysN( ConceptSystem systemO)  throws Exception{
	return new ConceptSystem( systemO, 0, 0, noise, 0 );
    }


    /** Fills the external similarity matrix and sets alpha. By default, it is
	filled with all zeros, but derived classes implementing 
	experiments with external similarity may set some non-zero
	values.
     */
    void  seedExSim(Absurdist absurdist) {
 	// initialize the external similarity matrix to all 0's
	absurdist.initExSim();
	absurdist.alpha=0;
   }

    private double nextToken(  StreamTokenizer token, String msg, int pos) 
	throws IOException {
	if (token.nextToken()==token.TT_EOF || token.ttype!=token.TT_NUMBER) 
	    throw new IOException(msg + ": expected a number in pos " + pos);
	return token.nval;
    }


    /** Reads data saved to a file during an earlier incomplete run,
	and adds them to the results accumulator

	@param avg The results accumulator array to add recovered data to
	@param out The new results file
	@return number of runs recovered
	
    */
    int recoverOldFile(Absurdist.Results avg, int NRuns, PrintStream out) 
	throws IOException {
	if (recover == null) {
	    System.out.println("No file recovery requested\n");
	    return 0;
	} else {
	    System.out.println("Recovering data from file " + recover);
	}
	LineNumberReader in = new LineNumberReader(new FileReader(recover));
	String line=null;
	int cnt = 0;
	while( (line = in.readLine())!=null && cnt<NRuns) {
	    StreamTokenizer token =
		new StreamTokenizer(new StringReader(line));

	    String msg = "File " + recover +", line no. " + in.getLineNumber();

	    // Data format:
	    // index misCon misRel misRelAdjId misRelAdjOp accuracyiterations
	    // 0       8.0     14.0    2.0     4.0     0.72222 5341.0
	    
	    // Skip lines not beginning with numbers
	    if (token.nextToken() == token.TT_EOF ||
		token.ttype != token.TT_NUMBER) continue;

	    double j = token.nval;
	    if (j != cnt) 
		throw new IOException(msg + ": index mismatch: found "+ j +
				      ", expected " + cnt);

	    Absurdist.Results r = new Absurdist.Results();

	    r.misCon = nextToken(token, msg, 2);
	    r.misRel = nextToken(token, msg, 3);
	    r.misRelAdjId = nextToken(token, msg, 4);
	    r.misRelAdjOp = nextToken(token, msg, 5);
	    r.accuracy = nextToken(token, msg, 6);
	    r.iterCnt =  nextToken(token, msg, 7);

	    out.println(cnt + 
			"\t" + r.misCon +
			"\t" + r.misRel +
			"\t" + r.misRelAdjId +
			"\t" + r.misRelAdjOp +
			"\t" + dfa.format(r.accuracy) +
			"\t" + r.iterCnt);
	    avg.add(r);
	    cnt ++;
	}	
	in.close();

	System.out.println("Recovered " +cnt+ " entries from file " + recover);
	return cnt;
    }

    /** Create a system with a specified number of nodes and
	"relation fulness", and its noisy copy with a specified noise
	level; match them; repeat NRuns times.
     */   
    void mainLoop(int n, double fullness, double noise, 
		  int NRuns, PrintStream out ) throws Exception {
	this.n = n;
	this.fullness = fullness;
	this.noise = noise;
	this.out = out;

	out.println( "\nindex\tmisCon\tmisRel\tmisRelAdjId\tmisRelAdjOp\taccuracy \titerations" );
	out.println( "----------------------------------------" );
	
	Absurdist.turnOffHistory(); // avoid wasting memory

	// test on NRuns random systems
	Absurdist.Results avg = new Absurdist.Results();

	int i=0;

	// Recover an old results file, if requested
	i=recoverOldFile( avg, NRuns, out);

	// New runs
	for ( ; i < NRuns; i++ ) {
	    systemO = getSysO();
	    systemN = getSysN(systemO);
	    Absurdist absurdist = new Absurdist( systemO, systemN );
	    // Fill the external similarity matrix (with zeros, or
	    // otherwise, as required for this experiment)
	    seedExSim(absurdist);
	    absurdist.mapSystem();
	    if (debugAll) absurdist.print( );
    
	    out.println( i + 
			 "\t" + absurdist.results.misCon +
			 "\t" + absurdist.results.misRel +
			 "\t" + absurdist.results.misRelAdjId +
			 "\t" + absurdist.results.misRelAdjOp +
			 "\t" + dfa.format(absurdist.results.accuracy) +
			 "\t" + absurdist.results.iterCnt);
	    avg.add(absurdist.results);
	
	}
	
	avg.averages();
	
	printStats( n, fullness, noise,  NRuns, avg, out   );
	
    }
 
    
    static void printStats(int n, double fullness, double noise, int NRuns,
		    Absurdist.Results avg, PrintStream out   ) {

	// adjust average mismapped edges depending on noise level, 
	// i.e. minus the supposed mismapped edges under even perfect map 

	double adjAvgMisRel = avg.misRel - noise * n * (n-1) / 2;
	
	double percent = ( double )avg.perfectCnt/NRuns;      // percentage of optimal mappings
	double perMisCon = avg.misCon / n;                    // percentage of mismapped concepts 
	double maxDiff = fullness*n*(n-1);

	double perMisRel = avg.misRel / maxDiff; // percentage of mismapped relations 	
	double adjPerMisRel = adjAvgMisRel / maxDiff; // percentage of adjusted mismapped relations 
	double perMisRelAdjId = avg.misRelAdjId / maxDiff; // percentage of mismapped relations 
	double perMisRelAdjOp = avg.misRelAdjOp / maxDiff; // percentage of mismapped relations 

	
	DecimalFormat df = new DecimalFormat( "0.000" );
	

	out.println("\nAverage iteration count             : " +
		    df.format(avg.iterCnt)+" +- "+
		    df.format(avg.iterCntDev));
	

	out.println( "Total perfect mappings              : " + 
		     avg.perfectCnt + 
		     ", percentage: " + percent );
	out.println( "Average mismapped concepts          : " +
		     df.format(avg.misCon)+
		     " +- "+  df.format(avg.misConDev) +
		     ", percentage: " + df.format(perMisCon) );

	out.println( "Average mismapped relations         : "+
		     df.format(avg.misRel)+
		     " +- "+ df.format(avg.misRelDev)+ 
		     ", percentage: " + df.format(perMisRel) );
	out.println( "Average identity-adj mismapped relations    : "+
		     df.format(avg.misRelAdjId)+
		     " +- "+ df.format(avg.misRelAdjIdDev)+ 
		     ", percentage: " + df.format(perMisRelAdjId) );

	if (Absurdist.adjBest) {
	out.println( "Average optimum-adj mismapped relations     : "+
		     df.format(avg.misRelAdjOp)+
		     " +- "+ df.format(avg.misRelAdjOpDev)+ 
		     ", percentage: " + df.format(perMisRelAdjOp) );
	}

	out.println( "[Theor] Adjusted average mismapped relations: " + 
		     df.format(adjAvgMisRel) +
		     ", percentage: " + df.format(adjPerMisRel) );
	out.println( "Average mapping accuracy: " +df.format(avg.accuracy)+
		     " +- "+ df.format(avg.accuracyDev));
	

    }

}
